/*
 * FreeMarker: a tool that allows Java programs to generate HTML
 * output using templates.
 * Copyright (C) 1998-2004 Benjamin Geer
 * Email: beroul@users.sourceforge.net
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA  02111-1307, USA.
 */

package freemarker.doc;

import java.io.*;
import java.util.Map;
import java.util.Properties;

import freemarker.engine.*;
import freemarker.template.*;
import freemarker.ext.misc.ExtendedList;
import freemarker.ext.misc.WriterExceptionListener;
import freemarker.ext.misc.Transforms;
import freemarker.ext.jdom.NodeListModel;
import org.jdom.*;
import org.jdom.input.SAXBuilder;

/**
 * Creates the FM-Classic web site navigation bars.
 *
 * @version $Id: WebSite.java 1140 2013-04-14 12:01:41Z run2000 $
 */
public final class WebSite {
    private Map m_cProperties;

    /** Default constructor. */
    public WebSite() {
    }

    /**
     * Creates the navigation bar for the manual. The parsing happens
     * courtesy of the {@link freemarker.ext.jdom.NodeListModel}
     * class, and most of the output logic happens in the template.
     */
    public void makeTableOfContents() throws Exception {
        FMEngine engine = FMEngine.singleton();
        String path = getRootPath();
        engine.setProperty(PropertyConstants.TEMPLATE_PATH, path);
        engine.setProperty(PropertyConstants.TEMPLATE_RUNTIME_ESCAPE,
                "freemarker.ext.misc.HtmlEscape");

        SAXBuilder builder = new SAXBuilder();
        FileInputStream cInputStream;
        Document document;

        Writer errWriter = new OutputStreamWriter( System.err );
        WriterExceptionListener errListener = new WriterExceptionListener( errWriter );
        TemplateRuntimeHandler runtimeHandler = new SimpleEventAdapter( errListener );

        engine.setObject(PropertyObjectConstants.TEMPLATE_RUNTIME_HANDLER, runtimeHandler);
        engine.init();

        System.out.println( "Making manual table of contents..." );

        cInputStream = new FileInputStream( path +
                "/src/freemarker/doc/xml/website.xml" );
        document = builder.build( cInputStream );

        FMModel modelRoot = new FMModel();
        modelRoot.putModel( "document", new NodeListModel( document ));
        modelRoot.putModel( "transforms", Transforms.getInstance() );
        modelRoot.putModel( "currentTab", new FastScalar( "Main" ));
        modelRoot.putModel( "relativePath", new FastScalar( "/" ));

        ExtendedList tabs = new ExtendedList();
        TabItem tab = new TabItem( "Main", "nav-main.html" );
        tabs.add( tab );
        tab = new TabItem( "Manual", "nav-manual.html" );
        tabs.add( tab );
        modelRoot.putModel( "tabList", tabs );

        Template template = engine.getTemplate(
                "src/freemarker/doc/template/web-manual.html.template");

        Writer fileOut = new FileWriter(
                new File( path, "nav-main.html" ));

        engine.apply(template, modelRoot, fileOut);

        fileOut.flush();
        fileOut.close();

        cInputStream = new FileInputStream( path +
                "/src/freemarker/doc/xml/contents.xml" );
        document = builder.build( cInputStream );
        modelRoot.putModel( "currentTab", new FastScalar( "Manual" ));
        modelRoot.putModel( "relativePath", new FastScalar( "/docs/manual/" ));

        modelRoot.putModel( "document", new NodeListModel( document ));

        fileOut = new FileWriter(
                new File( path, "nav-manual.html" ));

        engine.apply(template, modelRoot, fileOut);

        errWriter.flush();
        fileOut.flush();
        fileOut.close();

        System.out.println( "Finished manual table of contents." );
    }

    /**
     * Retrieve the root path of the FreeMarker distribution from a
     * properties file. This is used to set the root of the FreeMarker
     * cache, which in turn is used to find the example source code.
     *
     * @return the root path of the FreeMarker distribution
     */
    private String getRootPath() throws IOException {
        if( m_cProperties == null ) {
            loadProperties();
        }
        return (String)m_cProperties.get( "rootPath" );
    }

    /**
     * Loads properties from the <code>doc.properties</code> file.
     */
    private void loadProperties() throws IOException {
        InputStream input = getClass().getResourceAsStream( "doc.properties" );
        m_cProperties = new Properties();

        if( input != null ) {
            ((Properties)m_cProperties).load( input );
            input.close();
        } else {
            System.err.println( "Unable to find doc.properties file!" );
        }
    }

   /**
    * Main method for creating the website navigation bar.
    *
    * @param args the command line arguments
    */
    public static void main (String args[]) {
        WebSite cWebSite = new WebSite();

        try {
            cWebSite.makeTableOfContents();
        } catch( Exception e ) {
            System.err.println( "Unable to create FreeMarker web site:" );
            System.err.println( e.toString() );
        }
    }

}
