/*
 * FreeMarker: a tool that allows Java programs to generate HTML
 * output using templates.
 * Copyright (C) 1998-2013 Benjamin Geer
 * Email: beroul@users.sourceforge.net
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA  02111-1307, USA.
 */

package freemarker.engine;

import freemarker.ext.misc.TransformPipeline;
import freemarker.template.*;
import freemarker.template.cache.*;

import java.io.*;
import java.lang.reflect.*;
import java.util.*;

/**
 * This class provides services to an application developer, such as:
 * <ul>
 *     <li>Initialization methods for initializing the cache,
 *     the default runtime handler, and default escaping model</li>
 *     <li>Methods to apply the template engine to streams
 *     to allow easy embedding</li>
 *     <li>Other utility methods</li>
 * </ul>
 * <p>
 *  The methods in this class were developed to allow easy access to FM-Classic
 *  facilities without elaborate object construction.
 * </p>
 * <p>
 *   This class provides can be used either as separate, new-able instances
 *   of this engine, or as a singleton using the <code>singleton()</code>
 *   static method.
 * </p>
 * <p>
 *  Please ensure that you call one of the init() variants. This is critical
 *  for proper behavior.
 * </p>
 *
 * @author Nicholas Cull
 * @version $Id: FMEngine.java 1172 2013-04-21 10:00:22Z run2000 $
 * @since 1.9
 */
public final class FMEngine {
    private Properties properties;
    private boolean isInitialized;
    private CachingStrategy cache;
    private TemplateRuntimeHandler runtimeHandler;
    private TemplateModel escapeModel;
    private TemplateModel globalFilter;
    private String instanceMethods = "instance;getinstance;newinstance;singleton;getsingleton";
    private String instanceFields = "instance;singleton";

    /** The singleton instance, if one is required */
    private static FMEngine instance;

    /**
     * Return a singleton instance of a FM-Classic engine. If this is the
     * first time this method has been called, a new singleton instance
     * is created.
     *
     * @return a singleton FMEngine instance
     */
    public static FMEngine singleton() {
        synchronized (FMEngine.class) {
            if(instance == null) {
                instance = new FMEngine();
            }
            return instance;
        }
    }

    /**
     * Create a new FM-Classic engine.
     */
    public FMEngine() {
        properties = PropertyConstants.getDefaultProperties();
        isInitialized = false;
    }

    /**
     * Set a property using one of the property keys from the
     * {@link PropertyConstants} class.
     *
     * @param key the property name from {@link PropertyConstants}
     * @param value the string value corresponding to the given key
     * @throws IllegalStateException the engine's init() method has
     * already been called
     */
    public FMEngine setProperty(String key, String value) {
        if(isInitialized) {
            throw new IllegalStateException("Engine is already initialized");
        }
        properties.setProperty(key, value);
        return this;
    }

    /**
     * Get a property using one of the property keys from the
     * {@link PropertyConstants} class.
     *
     * @param key the property name from {@link PropertyConstants}
     * @return a string value corresponding to the given key
     */
    public String getProperty(String key) {
        return properties.getProperty(key);
    }

    /**
     * Initialize this engine using properties supplied from the given
     * properties object. Only these properties will be used to initialize
     * the engine.
     *
     * @param properties the properties to initialize this engine, using
     * property keys from the {@link PropertyConstants} class
     * @throws IllegalStateException the engine's init() method has
     * already been called
     */
    public FMEngine init(Properties properties) {
        if(isInitialized) {
            throw new IllegalStateException("Engine is already initialized");
        }
        this.properties = properties;
        init();
        return this;
    }

    /**
     * Initialize this engine using properties set using the
     * {@link #setProperty(String, String)} method. Default properties
     * are supplied from the {@link PropertyConstants#getDefaultProperties()}
     * method.
     *
     * @throws IllegalStateException the engine's init() method has
     * already been called
     */
    public FMEngine init() {
        if(isInitialized) {
            throw new IllegalStateException("Engine is already initialized");
        }

        // Set up reflection parameters to get a new singleton instance
        instanceMethods = properties.getProperty(PropertyConstants.TEMPLATE_ENGINE_INSTANCE_METHODS);
        instanceFields = properties.getProperty(PropertyConstants.TEMPLATE_ENGINE_INSTANCE_FIELDS);

        if(instanceMethods == null) {
            instanceMethods = "getinstance;newinstance";
        }
        if(instanceFields == null) {
            instanceFields = "instance";
        }

        if(cache == null) {
            // Build up the cache
            FileRetrieverParameters frp = FileRetrieverParameters.newInstance();
            String root = properties.getProperty(PropertyConstants.TEMPLATE_PATH);
            if(root != null) {
                frp = frp.withDirectoryRoot(root);
            }
            String encoding = properties.getProperty(PropertyConstants.TEMPLATE_ENCODING);
            if(encoding != null) {
                frp = frp.withEncoding(encoding);
            }
            String suffix = properties.getProperty(PropertyConstants.TEMPLATE_CACHE_SUFFIX);
            if(suffix != null) {
                frp = frp.withFilenameSuffix(suffix);
            }

            CacheRetrieverFactory crf = null;
            String tr = properties.getProperty(PropertyConstants.TEMPLATE_RETRIEVER);
            if(tr != null) {
                Object obj = newInstanceFromClassName(tr);
                if(obj instanceof CacheRetrieverFactory) {
                    crf = (CacheRetrieverFactory) obj;
                }
            }
            if(crf == null) {
                crf = FileRetrieverFactory.INSTANCE;
            }
            CacheRetriever cr = crf.create(frp);

            StrategyParameters sp = StrategyParameters.newInstance(cr);
            String defaultType = properties.getProperty(PropertyConstants.TEMPLATE_TYPE_DEFAULT);
            if(defaultType != null) {
                sp = sp.withDefaultTemplateType(defaultType);
            }
            String delay = properties.getProperty(PropertyConstants.TEMPLATE_CACHE_UPDATE_PERIOD);
            if(delay != null) {
                long delayLong = Long.parseLong(delay);
                sp = sp.withDelay(delayLong);
            }

            String maxAge = properties.getProperty(PropertyConstants.TEMPLATE_CACHE_UPDATE_MAXAGE);
            if(maxAge != null) {
                int maxAgeInt = Integer.parseInt(maxAge);
                sp = sp.withMaximumAge(maxAgeInt);
            }

            String strategy = properties.getProperty(PropertyConstants.TEMPLATE_CACHE_STRATEGY);
            if( strategy != null ) {
                Object obj = newInstanceFromClassName(strategy);
                if(obj instanceof CachingStrategyFactory) {
                    CachingStrategyFactory factory = (CachingStrategyFactory) obj;
                    cache = factory.create( sp );
                }
            }
            if(cache == null) {
                cache = new LoDWithRefreshCachingStrategy( sp );
            }
        }
        cache.startAutoUpdate();

        // Set up the runtime parameters
        if(runtimeHandler == null) {
            String rh = properties.getProperty(PropertyConstants.TEMPLATE_RUNTIME_HANDLER);
            if(rh != null) {
                Object obj = newInstanceFromClassName(rh);
                runtimeHandler =  (TemplateRuntimeHandler) obj;
            }
            if(runtimeHandler == null) {
                runtimeHandler = TemplateEventAdapter.DefaultEventAdapter;
            }
        }

        // Set up the escape model, which may be a pipeline
        if(escapeModel == null) {
            String escape = properties.getProperty(PropertyConstants.TEMPLATE_RUNTIME_ESCAPE);
            if((escape != null) && (escape.length() > 0)) {
                String[] classNames = splitBySeparator(escape);
                if(classNames.length == 1) {
                    Object obj = newInstanceFromClassName(escape);
                    if(obj instanceof TemplateModel) {
                        escapeModel = (TemplateModel) obj;
                    }
                } else {
                    List<TemplateModel> models = new ArrayList<TemplateModel>();
                    for (String c : classNames) {
                        Object obj = newInstanceFromClassName(c);
                        if (obj instanceof TemplateModel) {
                            models.add((TemplateModel) obj);
                        }
                    }
                    if(!models.isEmpty()) {
                        escapeModel = new TransformPipeline(models);
                    }
                }
            }
        }

        // Set up the global filter, which may be a pipeline
        if(globalFilter == null) {
            String filter = properties.getProperty(PropertyConstants.TEMPLATE_RUNTIME_GLOBAL_FILTER);
            if((filter != null) && (filter.length() > 0)) {
                String[] classNames = splitBySeparator(filter);
                if(classNames.length == 1) {
                    Object obj = newInstanceFromClassName(filter);
                    if(obj instanceof TemplateModel) {
                        globalFilter = (TemplateModel) obj;
                    }
                } else {
                    List<TemplateModel> models = new ArrayList<TemplateModel>();
                    for (String c : classNames) {
                        Object obj = newInstanceFromClassName(c);
                        if (obj instanceof TemplateModel) {
                            models.add((TemplateModel) obj);
                        }
                    }
                    if(!models.isEmpty()) {
                        globalFilter = new TransformPipeline(models);
                    }
                }
            }
        }

        isInitialized = true;
        return this;
    }

    /**
     * Sets an object property using constants from the
     * {@link PropertyObjectConstants} class.
     *
     * @param propertyName a property name from the {@link PropertyObjectConstants}
     * class
     * @param objValue the object value to be set
     */
    public void setObject(String propertyName, Object objValue) {
        if(isInitialized) {
            throw new IllegalStateException("Engine is already initialized");
        }

        if(PropertyObjectConstants.TEMPLATE_CACHING_STRATEGTY.equals(propertyName)) {
            cache = (CachingStrategy) objValue;
        } else if(PropertyObjectConstants.TEMPLATE_RUNTIME_HANDLER.equals(propertyName)) {
            runtimeHandler = (TemplateRuntimeHandler) objValue;
        } else if(PropertyObjectConstants.TEMPLATE_RUNTIME_ESCAPE.equals(propertyName)) {
            escapeModel = (TemplateModel) objValue;
        } else if(PropertyObjectConstants.TEMPLATE_RUNTIME_GLOBAL_FILTER.equals(propertyName)) {
            globalFilter = (TemplateModel) objValue;
        }
    }

    /**
     * Get the template cache initialized by this engine.
     *
     * @return the cache initialized for this engine
     */
    public Cache getCache() {
        return cache;
    }

    /**
     * Return the named template from the cache.
     *
     * @param name the name of the template
     * @return the named Template, or <code>null</code> if it is not found
     */
    public Template getTemplate(String name) {
        return (Template) cache.getItem(name, "template");
    }

    /**
     * Return the named cacheable item from the cache.
     *
     * @param name the name of the cacheable item
     * @return the named item, or <code>null</code> if it is not found
     */
    public Cacheable getItem(String name) {
        return cache.getItem(name);
    }

    /**
     * Return the named, typed cacheable item from the cache.
     *
     * @param name the name of the cacheable item
     * @param type the type of the cacheable item
     * @return the named Template, or <code>null</code> if it is not found
     */
    public Cacheable getItem(String name, String type) {
        return cache.getItem(name, type);
    }

    /**
     * Apply the given model to the named template, writing the result to
     * the given writer.
     *
     * @param templateName the name of the template to be loaded from the cache
     * @param model the model to apply to the template
     * @param writer the result of applying the model to the template
     * @return the exit status
     * @throws IOException there was a problem writing the result to the writer
     */
    public TemplateProcessor.ExitStatus apply(String templateName, FMModel model, Writer writer) throws IOException {
        Template template = (Template) cache.getItem(templateName, "template");
        return apply(template, model.getModel(), writer);
    }

    /**
     * Apply the given model to the given template, writing the result to
     * the given writer.
     *
     * @param template the template applied against the model
     * @param model the model to apply to the template
     * @param writer the result of applying the model to the template
     * @return the exit status
     * @throws IOException there was a problem writing the result to the writer
     */
    public TemplateProcessor.ExitStatus apply(Template template, FMModel model, Writer writer) throws IOException {
        return apply(template, model.getModel(), writer);
    }

    /**
     * Apply the given model to the named template, writing the result to
     * the given writer.
     *
     * @param templateName the name of the template to be loaded from the cache
     * @param model the model to apply to the template
     * @param writer the result of applying the model to the template
     * @return the exit status
     * @throws IOException there was a problem writing the result to the writer
     */
    public TemplateProcessor.ExitStatus apply(String templateName, TemplateWriteableHashModel model, Writer writer)
            throws IOException {
        Template template = (Template) cache.getItem(templateName, "template");
        return apply(template, model, writer);
    }

    /**
     * Apply the given model to the given template, writing the result to
     * the given writer.
     *
     * @param template the template applied against the model
     * @param model the model to apply to the template
     * @param writer the result of applying the model to the template
     * @return the exit status
     * @throws IOException there was a problem writing the result to the writer
     */
    public TemplateProcessor.ExitStatus apply(Template template, TemplateWriteableHashModel model, Writer writer)
            throws IOException {

        TemplateModel filterModel = globalFilter;
        TemplateProcessor.ExitStatus result = TemplateProcessor.ExitStatus.OK;

        try {
            if((filterModel != null) && (!filterModel.isEmpty())) {
                StringWriter sw = new StringWriter();
                TemplateProcessorParameters tpp = TemplateProcessorParameters.newInstance(sw)
                        .withEventHandler(runtimeHandler)
                        .withEscape(escapeModel)
                        .withModel(model);

                result = template.process(tpp);

                if(filterModel instanceof TemplateTransformModel) {
                    TemplateTransformModel transform = (TemplateTransformModel) filterModel;
                    StringReader sr = new StringReader(sw.toString());
                    PrintWriter pw = new PrintWriter(writer);
                    transform.transform(sr, pw);
                } else if(filterModel instanceof TemplateTransformModel2) {
                    TemplateTransformModel2 transform = (TemplateTransformModel2) filterModel;
                    StringReader sr = new StringReader(sw.toString());
                    transform.transform(sr, writer);
                } else {
                    writer.write(sw.toString());
                }
            } else {
                TemplateProcessorParameters tpp = TemplateProcessorParameters.newInstance(writer)
                        .withEventHandler(runtimeHandler)
                        .withEscape(escapeModel)
                        .withModel(model);

                result = template.process(tpp);
            }
        } catch (TemplateModelException e) {
            runtimeHandler.fireExceptionThrown(this, new TemplateException(
                    "Couldn't perform global filter", e),
                    writer, "freemarker.engine.FMEngine.apply", TemplateRuntimeHandler.Severity.WARNING);
        }

        return result;
    }

    /**
     * A simple way of creating objects and adding them to a new template
     * model, using class named supplied from a properties object.
     * Reflection is used to create the objects.
     *
     * @param properties a key-value pairing of names to class names to be
     * loaded and added to the new template model
     * @return the new template model
     */
    public FMModel createModel(Properties properties) {
        Enumeration e = properties.propertyNames();
        FMModel model = new FMModel();
        while(e.hasMoreElements()) {
            String modelName = (String) e.nextElement();
            String className = (String) properties.get(modelName);
            Object obj = newInstanceFromClassName(className);
            if(obj != null) {
                model.putObject(modelName, obj);
            }
        }
        return model;
    }

    /**
     * Create a new instance of the named class using reflection. See the
     * {@link PropertyConstants} class for details on how the reflection
     * code creates the object.
     *
     * @param c the named class to be loaded
     * @return the instantiated class, or <tt>null</tt> if it cannot be loaded
     */
    public Object newInstanceFromClassName(String c) {
        Class clazz = null;
        int idx = c.lastIndexOf('#');

        if(idx < 0) {
            try {
                clazz = Class.forName(c);
                return newObjectFromClass(clazz);
            } catch(ClassNotFoundException e) {
                //  Do nothing here
            }
        }

        if(idx < 0) {
            idx = c.lastIndexOf('.');
        }

        // Maybe the parent is the class name instead?
        if(idx < 0) {
            return null;
        }

        String c2 = c.substring(0, idx);
        String f2 = c.substring(idx + 1);

        try {
            clazz = Class.forName(c2);
            return fieldObjectFromClass(clazz, f2);
        } catch(ClassNotFoundException e) {
            // Do nothing here
        }
        return null;
    }

    private Object newObjectFromClass(Class clazz) {

        if(clazz.isInterface() || clazz.isArray() || clazz.isPrimitive()) {
            return null;
        }

        // Try no-argument constructor first
        Constructor[] constructors = clazz.getConstructors();
        for (Constructor cs : constructors) {
            try {
                if (Modifier.isPublic(cs.getModifiers())) {
                    Class[] params = cs.getParameterTypes();
                    if (params.length == 0) {
                        // Accessible, no argument constructor. Let's call it.
                        return cs.newInstance(new Object[0]);
                    }
                }
            } catch (InstantiationException e) {
                // Is abstract or interface. Will not succeed.
                break;
            } catch (IllegalAccessException e) {
                // Maybe another constructor would work?
            } catch (InvocationTargetException e) {
                // Some other exception in the constructor, try another?
            }
        }

        // Try static methods next
        if(instanceMethods != null) {
            Method[] methods = clazz.getDeclaredMethods();
            String[] methodNames = splitBySeparator(instanceMethods);
            for (Method m : methods) {
                try {
                    int modifiers = m.getModifiers();
                    if (Modifier.isPublic(modifiers) && Modifier.isStatic(modifiers)) {
                        Class[] params = m.getParameterTypes();
                        if (params.length == 0) {
                            String name = m.getName();
                            for (String methodName : methodNames) {
                                if (methodName.equalsIgnoreCase(name)) {
                                    return m.invoke(null, new Object[0]);
                                }
                            }
                        }
                    }
                } catch (IllegalAccessException e) {
                    // Maybe a different method will work?
                } catch (InvocationTargetException e) {
                    // Some other exception in the method, try another?
                }
            }
        }

        // Finally, try static final fields
        if(instanceFields != null) {
            Field[] fields = clazz.getDeclaredFields();
            String[] fieldNames = splitBySeparator(instanceFields);
            for (Field f : fields) {
                try {
                    int modifiers = f.getModifiers();
                    if (Modifier.isPublic(modifiers) && Modifier.isStatic(modifiers) && Modifier.isFinal(modifiers)) {
                        String name = f.getName();
                        for (String fieldName : fieldNames) {
                            if (fieldName.equalsIgnoreCase(name)) {
                                return f.get(null);
                            }
                        }
                    }
                } catch (IllegalAccessException e) {
                    // Maybe a different field will work?
                }
            }
        }

        return null;
    }

    private Object fieldObjectFromClass(Class clazz, String fieldName) {

        // Try static final named fields first
        Field[] fields = clazz.getDeclaredFields();
        for (Field f : fields) {
            try {
                int modifiers = f.getModifiers();
                if (Modifier.isPublic(modifiers) && Modifier.isStatic(modifiers) && Modifier.isFinal(modifiers)) {
                    String name = f.getName();
                    if (fieldName.equalsIgnoreCase(name)) {
                        return f.get(null);
                    }
                }
            } catch (IllegalAccessException e) {
                // Maybe a different field will work?
            }
        }

        // Try static methods
        Method[] methods = clazz.getDeclaredMethods();
        for (Method m : methods) {
            try {
                int modifiers = m.getModifiers();
                if (Modifier.isPublic(modifiers) && Modifier.isStatic(modifiers)) {
                    Class[] params = m.getParameterTypes();
                    if (params.length == 0) {
                        String name = m.getName();
                        if (fieldName.equalsIgnoreCase(name)) {
                            return m.invoke(null, new Object[0]);
                        }
                    }
                }
            } catch (IllegalAccessException e) {
                // Maybe a different method will work?
            } catch (InvocationTargetException e) {
                // Some other exception in the method, try another?
            }
        }

        return null;
    }

    private static String[] splitBySeparator(String property) {
        if(property == null) {
            return null;
        }
        List<String> l = new ArrayList<String>();
        StringTokenizer t = new StringTokenizer(property, ";: ");
        while(t.hasMoreTokens()) {
            l.add(t.nextToken());
        }
        String[] result = new String[l.size()];
        l.toArray(result);
        return result;
    }

    public static void main(String[] args) throws Exception {
        Properties properties = PropertyConstants.getDefaultProperties();
        properties.setProperty(PropertyConstants.TEMPLATE_RUNTIME_ESCAPE, "freemarker.ext.misc.HtmlEscape");
        properties.setProperty(PropertyConstants.TEMPLATE_RUNTIME_GLOBAL_FILTER, "freemarker.ext.misc.CompressLines");
        FMEngine engine = new FMEngine().init(properties);
        Template t = engine.getTemplate("src/freemarker/testcase/template/test-listliteral.html");
        StringWriter sw = new StringWriter();
        FMModel model = new FMModel();
        model.putObject("message", "Hello &copy world");
        engine.apply(t, model, sw);
        System.out.println(sw.toString());
    }
}
