/*
 * FreeMarker: a tool that allows Java programs to generate HTML
 * output using templates.
 * Copyright (C) 1998-2013 Benjamin Geer
 * Email: beroul@users.sourceforge.net
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA  02111-1307, USA.
 */

package freemarker.engine;

import freemarker.ext.beans2.Wrapper;
import freemarker.template.*;

import java.util.HashMap;

/**
 * General purpose class for passing object models to a FM-Classic
 * template. These can be passed in as objects or classes using the
 * {@link #putObject(String, Object)} and {@link #putClass(String, Class)}
 * methods.
 * <p>
 * For occasions where a custom <code>TemplateModel</code> is required,
 * these can be added using the {@link #putModel(String, freemarker.template.TemplateModel)}
 * method.
 * </p>
 * <p>Once the model is complete, it can be applied using one of the
 * {@link FMEngine#apply(freemarker.template.Template, FMModel, java.io.Writer)
 * FMEngine.apply()} methods.</p>
 *
 * @author Nicholas Cull
 * @version $Id: FMModel.java 1149 2013-04-15 09:57:11Z run2000 $
 * @since 1.9
 */
public class FMModel {
    protected final TemplateHashModel m_cInnerModel;
    protected final HashMap<String, TemplateModel> m_cModel;

    /**
     * Create a new empty FMModel. The model can be applied against a
     * template using the {@link FMEngine} object.
     */
    public FMModel() {
        this.m_cInnerModel = null;
        this.m_cModel = new HashMap<String, TemplateModel>();
    }

    /**
     * Create a new FMModel that wraps the given inner model. The inner
     * model can be read, but not altered from this model.
     *
     * @param innerModel the inner model to be wrapped
     */
    public FMModel(TemplateHashModel innerModel) {
        this.m_cInnerModel = innerModel;
        this.m_cModel = new HashMap<String, TemplateModel>();
    }

    /**
     * Create a new FMModel that wraps the given inner model. The inner
     * model can be read, but not altered from this model.
     *
     * @param innerModel the inner model to be wrapped
     */
    public FMModel(FMModel innerModel) {
        this.m_cInnerModel = innerModel.getModel();
        this.m_cModel = new HashMap<String, TemplateModel>();
    }

    /**
     * Puts the given template model into this FMModel, with the given name.
     *
     * @param name the name of the object to be added
     * @param model the template model to be added
     */
    public final FMModel putModel(String name, TemplateModel model) {
        if(model == null) {
            throw new NullPointerException("Model cannot be null");
        }
        m_cModel.put(name, model);
        return this;
    }

    /**
     * Puts the given object into this FMModel, with the given name.
     *
     * @param name the name of the object to be added
     * @param obj the object to be added
     */
    public final FMModel putObject(String name, Object obj) {
        if(obj == null) {
            throw new NullPointerException("Object cannot be null");
        }
        m_cModel.put(name, Wrapper.wrap(obj));
        return this;
    }

    /**
     * Puts the given class into this FMModel, with the given name.
     *
     * @param name the name of the object to be added
     * @param clz the class to be added
     */
    public final FMModel putClass(String name, Class clz) {
        if(clz == null) {
            throw new NullPointerException("Class cannot be null");
        }
        m_cModel.put(name, Wrapper.wrap(clz));
        return this;
    }

    /**
     * Puts the named class into this FMModel, with the given name.
     *
     * @param name the name of the object to be added
     * @param className the name of the class to be added
     */
    public final FMModel putClass(String name, String className) throws ClassNotFoundException {
        if(className == null) {
            throw new NullPointerException("Class name cannot be null");
        }
        Class clz = Class.forName(className);
        m_cModel.put(name, Wrapper.wrap(clz));
        return this;
    }

    /**
     * Retrieve this model to apply using the {@link FMEngine} class.
     * This clones the underlying map, meaning FMModel can be re-used to
     * build models across multiple template calls without (necessarily)
     * side effects.
     */
    TemplateModelRoot getModel() {
        HashMap<String, TemplateModel> outerMap = (HashMap<String, TemplateModel>) m_cModel.clone();
        if(m_cInnerModel == null) {
            return new FastHash<TemplateModel>(outerMap);
        }
        return new RootModelWrapper( m_cInnerModel, outerMap );
    }
}
