/*
 * FreeMarker: a tool that allows Java programs to generate HTML
 * output using templates.
 * Copyright (C) 1998-2013 Benjamin Geer
 * Email: beroul@users.sourceforge.net
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA  02111-1307, USA.
 */

package freemarker.engine;

import java.util.Properties;

/**
 * Constants to be supplied to {@link FMEngine} to configure the engine's
 * runtime behaviour. These are supplied using the {@link FMEngine#setProperty(String, String)
 * setProperty()} method or the {@link freemarker.engine.FMEngine#init(Properties)
 * init(Properties)} methods of the <code>FMEngine</code> class.
 * <p>
 * Many of the property values are supplied as class names to be loaded using
 * reflection.
 * </p>
 * <p>Class names may be given as:</p>
 * <pre>
 * package.name.classname
 * </pre>
 * <p>Static fields may be addressed within a class as:</p>
 * <pre>
 * package.name.classname.field
 * </pre>
 * <p>or:</p>
 * <pre>
 * package.name.classname#field
 * </pre>
 *
 * <p>When reflection is performed, it is handled the following way:
 * <ul>
 * <li>First of all, the name of the class is determined. There may
 * be a field name as the last part of the name</li>
 * <li>If the name consists only of a class name,
 *  <ul>
 *      <li>Try and call a public no-argument constructor</li>
 *      <li>Otherwise, look for a public, static, no-argument method that
 *      returns a singleton object</li>
 *      <li>Otherwise, look for a public, static, final field that
 *      contains a singleton object</li>
 *  </ul>
 *  Singleton method and field names are determined using the
 *  {@link #TEMPLATE_ENGINE_INSTANCE_METHODS} and
 *  {@link #TEMPLATE_ENGINE_INSTANCE_FIELDS} properties described below.
 * </li>
 * <li>Otherwise, look for a public, static, final field with the given
 * name that contains a singleton object</li>
 * <li>Otherwise, look for a public, static, no-argument method that
 * returns a singleton object</li>
 * </ul>
 * </p>
 *
 * @author Nicholas Cull
 * @version $Id: PropertyConstants.java 1172 2013-04-21 10:00:22Z run2000 $
 * @since 1.9
 */
public final class PropertyConstants {

    private PropertyConstants() {
    }

    /**
     * The fully-qualified path to a directory containing template files.
     */
    public static final String TEMPLATE_PATH = "template.path";
    /**
     * The character encoding to be used when reading templates from
     * the file system.
     */
    public static final String TEMPLATE_ENCODING = "template.encoding";
    /**
     * The default template type. The names are registered with the
     * TemplateRegistry.
     */
    public static final String TEMPLATE_TYPE_DEFAULT = "template.type.default";
    /**
     * The class name of the <code>CacheRetrieverFactory</code> for retrieving
     * items to be cached.
     * @see freemarker.template.cache.CacheRetrieverFactory
     */
    public static final String TEMPLATE_RETRIEVER = "template.retriever";
    /**
     * The class name of the <code>CachingStrategyFactory</code> for caching items.
     * @see freemarker.template.cache.CachingStrategyFactory
     */
    public static final String TEMPLATE_CACHE_STRATEGY = "template.cache.strategy";
    /**
     * Determines when items in the cache should be expired, as an integer value.
     */
    public static final String TEMPLATE_CACHE_UPDATE_MAXAGE = "template.cache.update.maxage";
    /**
     * Determines how often templates in the cache should be checked for
     * changes, as an integer number of seconds.
     */
    public static final String TEMPLATE_CACHE_UPDATE_PERIOD = "template.cache.update.period";
    /**
     * The filename suffix for items to be stored in the cache. Only items
     * with the given suffix will be stored.
     */
    public static final String TEMPLATE_CACHE_SUFFIX = "template.cache.suffix";

    /**
     * The class name or names of the <code>TemplateTransformModel</code> or
     * <code>TemplateTransformModel2</code> objects used whenever a variable
     * is written to the writer. If multiple classes are specified, the names
     * are separated by semicolons, colons, or spaces.
     * @see freemarker.template.TemplateTransformModel
     * @see freemarker.template.TemplateTransformModel2
     */
    public static final String TEMPLATE_RUNTIME_ESCAPE = "template.runtime.escape";
    /**
     * The class name or names of the <code>TemplateTransformModel</code> or
     * <code>TemplateTransformModel2</code> that filters all output from the
     * template. If multiple classes are specified, the names are separated by
     * semicolons, colons, or spaces.
     * @see freemarker.template.TemplateTransformModel
     * @see freemarker.template.TemplateTransformModel2
     */
    public static final String TEMPLATE_RUNTIME_GLOBAL_FILTER = "template.runtime.global.filter";
    /**
     * The class name of the <code>TemplateRuntimeHandler</code> used whenever
     * an error is reported by the runtime engine.
     * @see freemarker.template.TemplateRuntimeHandler
     */
    public static final String TEMPLATE_RUNTIME_HANDLER = "template.runtime.handler";

    /**
     * A list of method names for looking up a singleton method of a class.
     * The names are separated by semicolons, colons, or spaces.
     */
    public static final String TEMPLATE_ENGINE_INSTANCE_METHODS = "template.engine.instance.methods";
    /**
     * A list of field names for looking up a singleton field of a class.
     * The names are separated by semicolons, colons, or spaces.
     */
    public static final String TEMPLATE_ENGINE_INSTANCE_FIELDS = "template.engine.instance.fields";

    /**
     * Retrieve a default set of properties sufficient to bootstrap the
     * <code>FMEngine</code> object. This is the default set used when no other
     * properties are supplied.
     *
     * @return a Properties object containing a default set of properties
     * sufficient to bootstrap the <code>FMEngine</code> object
     */
    public static Properties getDefaultProperties() {
        Properties p = new Properties();

        p.setProperty(TEMPLATE_PATH, System.getProperty("user.dir"));
        p.setProperty(TEMPLATE_ENCODING, "UTF-8");
        p.setProperty(TEMPLATE_TYPE_DEFAULT, "template");
        p.setProperty(TEMPLATE_RETRIEVER, "freemarker.template.cache.FileRetrieverFactory");
        p.setProperty(TEMPLATE_CACHE_STRATEGY, "freemarker.template.cache.LoadOnDemandWithRefreshCacheFactory");
        p.setProperty(TEMPLATE_CACHE_UPDATE_MAXAGE, "0");
        p.setProperty(TEMPLATE_CACHE_UPDATE_PERIOD, "30");
        p.setProperty(TEMPLATE_CACHE_SUFFIX, "");
        p.setProperty(TEMPLATE_RUNTIME_HANDLER, "freemarker.template.TemplateEventAdapter#DefaultEventAdapter");
        p.setProperty(TEMPLATE_ENGINE_INSTANCE_METHODS, "instance;getinstance;newinstance;singleton;getsingleton");
        p.setProperty(TEMPLATE_ENGINE_INSTANCE_FIELDS, "instance;singleton");

        return p;
    }
}
