/*
 * ====================================================================
 *
 * Copyright (c) 2000 Attila Szegedi.  All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution, if
 *    any, must include the following acknowlegement:
 *       "This product includes the Expose library
 *        (http://www.szegedi.org/expose)."
 *    Alternately, this acknowlegement may appear in the software itself,
 *    if and wherever such third-party acknowlegements normally appear.
 *
 * 4. The names "Expose" and "Attila Szegedi" must not be used to endorse
 *    or promote products derived from this software without prior written
 *    permission. For written permission, please contact expose@szegedi.org.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR ITS CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
 * OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT
 * OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
 * BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
 * OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN
 * IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * ====================================================================
 *
 */

package freemarker.ext.beans;

import freemarker.template.*;
import freemarker.template.expression.ExpressionUtils;

import java.lang.reflect.Array;
import java.io.ObjectInputStream;
import java.io.IOException;
import java.io.InvalidObjectException;

/**
 * <p>A class that will wrap an arbitrary array into {@link TemplateHashModel},
 * {@link TemplateMethodModel2}, {@link TemplateIndexedModel} and
 * {@link TemplateListModel2} interfaces.</p>
 *
 * <p>The models are cached (meaning requesting a model for same object
 * twice will return the same model instance) unless the system property
 * <kbd>expose.reflection.nocache</kbd> is set to true.</p>
 *
 * <p>Apart from supporting retrieval through <code>array[index]</code> syntax
 * and <code>array("index")</code> syntax, it also supports the <code>array.length</code>
 * syntax for retrieving the length of the array. A template can iterate over
 * the array directly using the <code>&lt;list&gt;</code> or
 * <code>&lt;foreach&gt;</code> syntax.</p>
 *
 * <p>The <code>array.iterator</code> syntax for retrieving a
 * {@link freemarker.template.TemplateListModel2} interface backed by this array
 * is supported for backward compatibility with earlier releases.</p>
 *
 * <p>Using the model as a list model is thread-safe, as it maintains the
 * list position on a per-iterator basis using the new {@link TemplateListModel2}
 * interface.</p>
 *
 * @author Attila Szegedi, attila@szegedi.org
 * @version $Id: ArrayModel.java 1101 2013-04-01 04:17:32Z run2000 $
 * @deprecated this class is deprecated in favour of the classes in the
 * {@link freemarker.ext.beans2} package
 */
public final class ArrayModel extends ObjectModelBase implements
        TemplateHashModel, TemplateMethodModel2, TemplateIndexedModel,
        TemplateListModel2
{
	private static final String ITERATOR = "iterator";
	private static final String LENGTH = "length";

	// True if the array contains scalar values
	private transient boolean scalar;

	// Cached length of the array
	private transient int length;

    /** Serialization UUID for this class. */
    private static final long serialVersionUID = -5227864619177422669L;

	/**
	 * Creates a new model that wraps the specified array object.
	 * @param array the array object to wrap into a model.
	 * @throws IllegalArgumentException if the passed object is not a Java array.
	 */
	public ArrayModel(Object array)
	{
		super(array);
		Class clazz = array.getClass();
		if(!clazz.isArray())
			throw new IllegalArgumentException("Object is not an array, it is " + array.getClass().getName());
		scalar = BeansWrapper.isScalar(clazz.getComponentType());
		length = Array.getLength(array);
	}

	/**
	 * Returns a model wrapping the specified array object. If there is already
	 * a cached model instance for this array, returns the cached model instance.
	 * Models are cached using <code>WeakReference</code> objects. The caching can be turned
	 * off by setting the <kbd>expose.reflection.nocache</kbd> system property to
	 * true. In this case calling this method is equivalent to constructing a new model.
	 * @param object the array to wrap into a model.
	 * @throws IllegalArgumentException if the passed object is not a Java array.
	 * @return the model for the array
	 */
	public static final ArrayModel getInstance(Object object)
	{
		if(noCache)
			return new ArrayModel(object);

		ObjectModelBase modelbase = lookup(object);
		if(modelbase instanceof ArrayModel)
			return (ArrayModel)modelbase;

		ArrayModel model = new ArrayModel(object);
		register(model, object);
		return model;
	}

	/**
	 * Returns the type of this object (which is TYPE_ARRAY)
	 */
	public @Override ModelType getType()
	{
		return ModelType.TYPE_ARRAY;
	}

	/**
	 * If the key can be parsed by the
     * <code>java.lang.Integer#parseInt(java.lang.String)</code> method into an
	 * array index, the array element at the parsed index is returned. If the
	 * key is named "length", the length of the array is returned.
	 */
	public TemplateModel get(String key) throws TemplateModelException
	{
		if(ITERATOR.equals(key))
			return new Iterator();
		if(LENGTH.equals(key))
			return new FastNumber(length);

        try {
            return getAtIndex(Integer.parseInt(key));
        } catch( NumberFormatException e ) {
            throw new TemplateModelException( "Couldn't parse key as an integer", e );
        }
    }

	/**
	 * The first argument of the list is interpreted as an array index (it can be
	 * either a Number, or a String containing parseable integer). The element at
	 * the specified index is returned.
	 */
    public TemplateModel exec(java.util.List<TemplateModel> arguments)
    throws
        TemplateModelException
	{
        if( arguments.size() != 1 ) {
            throw new TemplateModelException( "ArrayModel needs exactly one argument" );
        }
        TemplateModel argument = (TemplateModel)arguments.get(0);
        if(argument instanceof TemplateNumberModel) {
            try {
                return getAtIndex( ExpressionUtils.getAsNumber( argument ));
            } catch( TemplateException e ) {
                throw new TemplateModelException( e.getMessage(), e );
            }
        }
        Object index = BeansWrapper.unwrap(argument);
        try {
            if(index instanceof Number)
                return getAtIndex(((Number)index).longValue());
            else
                return getAtIndex(Long.parseLong(index.toString()));
        } catch( NumberFormatException e ) {
            throw new TemplateModelException( "Could not parse as long " + index.toString(), e );
        }
    }

    /**
     * Retrive the object at the specified index.
     *
     * @param index the index of the object to be retrieved
     * @return a TemplateModel wrapping the underlying value of the object
     */
	public TemplateModel getAtIndex(long index) throws TemplateModelException
	{
		return BeansWrapper.wrap(Array.get(object, (int)index), scalar);
	}

    /**
     * Retrieve an iterator over this array.
     *
     * @return an iterator over the array, or <code>null</code> if the array
     * is empty
     * @throws TemplateModelException
     */
    public TemplateIteratorModel templateIterator() throws TemplateModelException {
        if( length == 0 )
            return null;

        return new FastIndexedIterator( ArrayModel.this, 0, length - 1 );
    }

    /**
     * Return the iterator to the object pool, if any.
     *
     * @param iterator the iterator to be returned
     */
    public void releaseIterator( TemplateIteratorModel iterator ) {
        // We don't cache iterators, so just return immediately
    }

	/**
	 * Returns true if the wrapped array is null, or its length is 0.
	 */
	public @Override boolean isEmpty()
	{
		return length == 0;
	}

    /**
     * Serialized form only contains the wrapped object itself. We need to
     * recreate the transient values from scratch.
     */
    private void readObject(ObjectInputStream stream) throws IOException, ClassNotFoundException {
        stream.defaultReadObject();

        Class clazz = object.getClass();
        if(!clazz.isArray())
            throw new InvalidObjectException("Object is not an array, it is " +
                    object.getClass().getName());
        scalar = BeansWrapper.isScalar(clazz.getComponentType());
        length = Array.getLength(object);
    }

    /**
     * Iterator class for backward compatibility with earlier releases.
     */
    private class Iterator implements TemplateListModel2
    {
        public TemplateIteratorModel templateIterator() throws TemplateModelException {
            if( length == 0 )
                return null;

            return new FastIndexedIterator( ArrayModel.this, 0, length - 1 );
        }

        public void releaseIterator( TemplateIteratorModel iterator ) {
            // Do nothing
        }

        public boolean isEmpty() throws TemplateModelException {
            return length == 0;
        }
    }
}
