/*
 * ====================================================================
 *
 * Copyright (c) 2000 Attila Szegedi.  All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution, if
 *    any, must include the following acknowlegement:
 *       "This product includes the Expose library
 *        (http://www.szegedi.org/expose)."
 *    Alternately, this acknowlegement may appear in the software itself,
 *    if and wherever such third-party acknowlegements normally appear.
 *
 * 4. The names "Expose" and "Attila Szegedi" must not be used to endorse
 *    or promote products derived from this software without prior written
 *    permission. For written permission, please contact expose@szegedi.org.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR ITS CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
 * OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT
 * OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
 * BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
 * OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN
 * IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * ====================================================================
 *
 */

package freemarker.ext.beans;

import freemarker.template.*;
import java.lang.ref.*;
import java.util.*;
import java.io.Serializable;

/**
 * Base class for all classes that wrap objects into template models.
 * Instances of this class and descendant classes can be created directly,
 * or utilizing a model cache through <code>getInstance()</code> calls, or
 * polymorphically through the {@link BeansWrapper#wrap(Object)} method.
 *
 * @author Attila Szegedi, attila@szegedi.org
 * @version $Id: ObjectModelBase.java 1101 2013-04-01 04:17:32Z run2000 $
 * @deprecated this class is deprecated in favour of the classes in the
 * {@link freemarker.ext.beans2} package
 */
public class ObjectModelBase implements TemplateScalarModel,
        TemplateObjectModel, Serializable
{
    enum ModelType {
        /** Returned from {@link ObjectModelBase#getType()} if the object is a ObjectModelBase */
        TYPE_OBJECT_BASE,
        /** Returned from {@link ObjectModelBase#getType()} if the object is a ObjectModel */
        TYPE_OBJECT,
        /** Returned from {@link ObjectModelBase#getType()} if the object is a ArrayModel */
        TYPE_ARRAY,
        /** Returned from {@link ObjectModelBase#getType()} if the object is a CollectionModel */
        TYPE_COLLECTION,
        /** Returned from {@link ObjectModelBase#getType()} if the object is a MapModel */
        TYPE_MAP,
        /** Returned from {@link ObjectModelBase#getType()} if the object is a IteratorModel */
        TYPE_ITERATOR,
        /** Returned from {@link ObjectModelBase#getType()} if the object is a EnumerationModel */
        TYPE_ENUMERATION,
        /** Returned from {@link ObjectModelBase#getType()} if the object is a ResourceBundleModel */
        TYPE_RESOURCE_BUNDLE,
        /** Returned from {@link ObjectModelBase#getType()} if the object is a List */
        TYPE_LIST
    }

	static final boolean noCache = Boolean.getBoolean("expose.reflection.nocache");
    // This is a DEBUG flag, that should be set to true only in builds that
    // are used to tune performance
    private static final boolean trackCaching = false;
    static int cacheHits = 0;
    static int cacheMisses = 0;

    private static final Map<Object, ModelReference> modelCache = noCache ? null : new IdentityHashMap<Object, ModelReference>();
	private static final ReferenceQueue refQueue = noCache ? null : new ReferenceQueue();

    /** Serialization UUID for this class. */
    private static final long serialVersionUID = 3158510158883866641L;

	protected Object object;

	ObjectModelBase(Object object)
	{
		this.object = object;
	}

	/**
	 * Creates a new model that wraps the specified String object.
	 * Note that since the only FM-Classic-specific method in this class is
	 * <code>getAsString</code> there is no sense in creating an object of this
	 * class directly with anything except a String. If you need to wrap an object
	 * of another class, use one of specialized subclasses, or delegate the choice
	 * to {@link BeansWrapper#wrap(Object)}.
	 * @param object the String to wrap into a model.
	 */
    ObjectModelBase(String object)
	{
		this.object = object;
	}

	/**
	 * <p>Returns a model wrapping the specified String object. If there is already
	 * a cached model instance for this String, returns the cached model instance.
	 * Models are cached using <code>WeakReference</code> objects. The caching can be turned
	 * off by setting the <kbd>expose.reflection.nocache</kbd> system property to
	 * true. In this case calling this method is equivalent to constructing a new model.</p>
     *
	 * <p>Note that since the only FM-Classic-specific method in this class is
	 * {@link #getAsString} there is no sense in creating an object of this
	 * class directly with anything except a <code>String</code>. If you need to wrap an object
	 * of another class, use one of specialized subclasses, or delegate the choice
	 * to {@link BeansWrapper#wrap(Object)}.</p>
     *
	 * @param object the String to wrap into a model.
	 * @return the model for the String
	 */
	public static final ObjectModelBase getInstance(String object)
	{
		if(noCache)
        {
            if(trackCaching)
                doTrackCaching(false);
		    return new ObjectModelBase(object);
        }

		ObjectModelBase model = lookup(object);
		if(model == null)
		{
            if(trackCaching)
                doTrackCaching(false);
		    model = new ObjectModelBase(object);
		    register(model, object);
		}
        else if(trackCaching)
            doTrackCaching(true);

		return model;
	}

        private static synchronized void doTrackCaching(boolean cacheHit)
        {
            if(cacheHit)
                cacheHits++;
            else
                cacheMisses++;
            if((cacheHits + cacheMisses) % 100 == 0)
                System.err.println("Expose cache: hits=" + cacheHits + " misses=" + cacheMisses);
        }

        /**
         * Forces clearing of model cache.
         */
        static void clearModelCache()
        {
            synchronized(modelCache)
            {
                modelCache.clear();
            }
        }

	/**
	 * Returns the wrapped object.
	 */
	public Object getAsObject()
	{
		return object;
	}

	final static ObjectModelBase lookup(Object object)
	{
		ModelReference ref;
		// NOTE: we're doing minimal synchronizations -- which can lead to
		// duplicate wrapper creation. However, this has no harmful side-effects and
		// is a lesser performance hit.
		synchronized(modelCache)
		{
			ref = modelCache.get(object);
		}

		if(ref != null)
			return ref.get();

		return null;
	}

	final static void register(ObjectModelBase model, Object object)
	{
		synchronized(modelCache)
		{
			// Remove cleared references
			for(;;)
			{
				ModelReference queuedRef = (ModelReference)refQueue.poll();
				if(queuedRef == null)
					break;
				modelCache.remove(queuedRef.object);
			}
			// Register new reference
			modelCache.put(object, new ModelReference(model));
		}
	}

	/**
	 * Returns the type of this object (which is TYPE_OBJECT_BASE)
	 */
	public ModelType getType()
	{
		return ModelType.TYPE_OBJECT_BASE;
	}

	/**
	 * Returns either the toString() of the wrapped object or the string "null",
	 * if the wrapped object is null.
	 */
	public String getAsString()
	{
		return object == null ? "null" : object.toString();
	}

	/**
	 * Tells whether the model is empty. It is empty if either the wrapped object is
	 * null, or it is a Boolean with false value.
	 */
	public boolean isEmpty()
	{
		return object == null || (object instanceof Boolean && !((Boolean)object).booleanValue());
	}

	/**
	 * A special soft reference that is registered in the modelCache.
	 * When it gets cleared (that is, the model became unreachable)
	 * it will remove itself from the model cache.
	 */
	private static final class ModelReference extends SoftReference<ObjectModelBase>
	{
		Object object;

		ModelReference(ObjectModelBase ref)
		{
			super(ref, refQueue);
			object = ref.object;
		}
	}

	private static String print(Object object)
	{
		if(object == null)
			return null;
		return object.getClass().getName() + '@' + Integer.toHexString(System.identityHashCode(object));
	}
}
