/*
 * FreeMarker: a tool that allows Java programs to generate HTML
 * output using templates.
 * Copyright (C) 1998-2005 Benjamin Geer
 * Email: beroul@users.sourceforge.net
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA  02111-1307, USA.
 */
package freemarker.ext.beans2;

import java.beans.IntrospectionException;
import java.beans.PropertyDescriptor;
import java.beans.IndexedPropertyDescriptor;
import java.lang.reflect.Field;
import java.lang.reflect.Method;
import java.lang.reflect.InvocationTargetException;
import java.util.List;
import java.io.Serializable;

import freemarker.template.*;

/**
 * Wrapper for a JavaBean or a Plain Old Java Object. Provides access to
 * fields, properties and methods of the object. Also a base class for other
 * Wrapper objects.
 *
 * @author Nicholas Cull
 * @version $Id: BeanWrapper.java 1067 2011-06-06 10:55:28Z run2000 $
 * @since 1.9
 */
public class BeanWrapper<T> implements TemplateWriteableHashModel,
        ObjectWrapper, TemplateObjectModel, TemplateScalarModel,
        Serializable {

    /** Serialization UUID for this class */
    private static final long serialVersionUID = -7037399358517604046L;

    /** The object being wrapped by this wrapper */
    protected T object;

    /**
     * For the benefit of property read methods, pre-construct an array
     * of zero argument to pass to the method
     */
    protected static final Object[] emptyParams = new Object[] { };

    /**
     * Default constructor where an object is not provided initially. This allows
     * the wrapper to be used as a prototype.
     */
    public BeanWrapper() {
    }

    /**
     * Constructor that takes a POJO or JavaBean object to be wrapped.
     *
     * @param object the object to be wrapped
     */
    public BeanWrapper( T object ) {
        this.object = object;
    }

    public void setObject( Object object ) {
        this.object = (T)object;
    }

    public boolean isEmpty() throws TemplateModelException {
        return object == null;
    }

    public TemplateModel get( String key ) throws TemplateModelException {

        if( key == null ) {
            return null;
        }

        Features<?> features;

        try {
            features = Wrapper.getObjectFeatures( object );
        } catch( IntrospectionException e ) {
            throw new TemplateModelException( "Cannot introspect object: " + e.getMessage(), e );
        }

        Field field = features.getField( key );
        if( field != null ) {
            try {
                return Wrapper.wrap( field.get( object ));
            } catch( IllegalAccessException e ) {
                throw new TemplateModelException( "Cannot access field " + key +
                        ": " + e.getMessage(), e );
            }
        }

        PropertyDescriptor pd = features.getProperty( key );

        if(( pd != null ) && ( pd.getReadMethod() != null )) {
            if( pd instanceof IndexedPropertyDescriptor ) {
                return new IndexedPropertyWrapper(  object, (IndexedPropertyDescriptor)pd );
            }

            Method method = pd.getReadMethod();
            try {
                Object result = Wrapper.invokeMethod( method, object, emptyParams );
                return Wrapper.wrap( result );
            } catch( IllegalAccessException e ) {
                throw new TemplateModelException( "Cannot invoke property getter on property: " +
                        e.getMessage(), e );
            } catch( InvocationTargetException e ) {
                throw new TemplateModelException( "Cannot invoke property getter on property: " +
                        e.getTargetException().getMessage(), (Exception)e.getTargetException() );
            }
        }

        List<Method> methods = features.getMethods( key );

        if(( methods != null ) && ( ! methods.isEmpty() )) {
            return new MethodWrapper( object, key );
        }

        throw new TemplateModelException( "Could not find object method, property, or field " + key );
    }

    /**
     * Sets a value in the hash model.
     *
     * @param key   the hash key
     * @param model the value to be added to the hash model
     */
    public void put( String key, TemplateModel model )
            throws TemplateModelException {

        if( key == null ) {
            return;
        }

        Features<?> features;

        try {
            features = Wrapper.getObjectFeatures( object );
        } catch( IntrospectionException e ) {
            throw new TemplateModelException( "Cannot introspect object: " + e.getMessage(), e );
        }

        Field field = features.getField( key );
        if( field != null ) {
            try {
                Object value = Unwrapper.unwrap( model, field.getType() );
                field.set( object, value );
                return;
            } catch( IllegalAccessException e ) {
                throw new TemplateModelException( "Cannot write field " + key +
                        ": " + e.getMessage(), e );
            }
        }

        PropertyDescriptor pd = features.getProperty( key );

        if(( pd != null ) && ( pd.getWriteMethod() != null )) {

            Method method = pd.getWriteMethod();
            try {
                Class<?> propertyType = pd.getPropertyType();
                if( propertyType != null ) {
                    Object value = Unwrapper.unwrap( model, propertyType );
                    Wrapper.invokeMethod( method, object, new Object[] { value });
                    return;
                }
            } catch( IllegalAccessException e ) {
                throw new TemplateModelException( "Cannot invoke property setter on property: " +
                        e.getMessage(), e );
            } catch( InvocationTargetException e ) {
                throw new TemplateModelException( "Cannot invoke property setter on property: " +
                        e.getTargetException().getMessage(), (Exception)e.getTargetException() );
            }
        }

        throw new TemplateModelException( "Could not set field or property " + key );
    }

    /**
     * Return the underlying object to the reflection mechanism. Any variables,
     * methods or properties can be called directly via reflection.
     *
     * @return the underlying object for this template model
     * @throws freemarker.template.TemplateModelException
     *          the object could not be returned
     */
    public Object getAsObject() throws TemplateModelException {
        return object;
    }

    /**
     * Returns the bean's value as a <code>String</code> using the object's
     * toString method.
     *
     * @return the <code>String</code> value of this object.
     */
    public String getAsString() throws TemplateModelException {
        return object.toString();
    }

    /**
     * Creates and returns a copy of this object.
     *
     * @return a clone of this instance.
     * @throws RuntimeException the clone operation was not successful
     * @throws OutOfMemoryError if there is not enough memory.
     * @see Cloneable
     */
    public Object clone() {
        try {
            return super.clone();
        } catch( CloneNotSupportedException e ) {
            throw new RuntimeException( "Clone method threw an exception: " + e.getMessage() );
        }
    }
}
