/*
 * FreeMarker: a tool that allows Java programs to generate HTML
 * output using templates.
 * Copyright (C) 1998-2005 Benjamin Geer
 * Email: beroul@users.sourceforge.net
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA  02111-1307, USA.
 */
package freemarker.ext.beans2;

import freemarker.template.*;

import java.util.Collection;
import java.util.List;
import java.io.Serializable;

/**
 * Wraps Collection objects by providing additional methods to support list
 * models and method support for getting and setting values. Also a base class
 * for ListWrapper.
 *
 * @author Nicholas Cull
 * @version $Id: CollectionWrapper.java 1067 2011-06-06 10:55:28Z run2000 $
 * @since 1.9
 */
public class CollectionWrapper<T extends Collection> extends IterableWrapper<T>
        implements TemplateListModel2, TemplateMethodModel2, Serializable {

    /** Serialization UUID for this class */
    private static final long serialVersionUID = 8727042208424695411L;

    /**
     * Default constructor where an object is not provided initially. This allows
     * the wrapper to be used as a prototype.
     */
    public CollectionWrapper() {
    }

    /**
     * Constructor that takes a Collection object to be wrapped.
     *
     * @param object the collection object to be wrapped
     */
    public CollectionWrapper( T object ) {
        super( object );
    }

    /**
     * Tests or sets an object in a Collection. If the number of arguments is
     * one, then a test for object existance is assumed, and we return a
     * FastBoolean for the default unwrapped value of the first argument.
     * <p>
     * Otherwise, if the second argument is either empty or the boolean value
     * "false", then the object is removed from the collection.
     * <p>
     * Finally, if the second argument is any other value, then add the object
     * to the collection.
     *
     * @param arguments a <code>List</code> of <code>TemplateModel</code> objects
     *                  containing the values of the arguments passed to the method.
     * @return the <code>TemplateModel</code> produced by the method, or <code>null</code>.
     */
    public TemplateModel exec( List<TemplateModel> arguments ) throws TemplateModelException {
        int size = arguments.size();

        if( size < 1 || size > 2 ) {
            throw new TemplateModelException( "Collection wrapper method call must contain exactly one or two arguments" );
        }

        Collection coll = object;
        Object item = Unwrapper.defaultUnwrap( arguments.get( 0 ));

        if( size == 1 ) {
            return FastBoolean.getInstance( coll.contains( item ));
        }

        TemplateModel value = arguments.get( 1 );
        if(( value == null ) || ( value  == FastBoolean.FALSE )) {
            coll.remove( item );
            return null;
        } else {
            coll.add( item );
            return null;
        }
    }
}
