/*
 * FreeMarker: a tool that allows Java programs to generate HTML
 * output using templates.
 * Copyright (C) 1998-2005 Benjamin Geer
 * Email: beroul@users.sourceforge.net
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA  02111-1307, USA.
 */
package freemarker.ext.beans2;

import freemarker.template.TemplateModel;
import freemarker.template.TemplateModelException;

/**
 * An interface for classes implementing a conversion from a given
 * TemplateModel to a known Java object. Conversions should be context free,
 * and implementing classes should be flywheel classes, ie. not carrying
 * any internal state between calls. They may also be singletons.
 *
 * @author Nicholas Cull
 * @version $Id: Converter.java 1067 2011-06-06 10:55:28Z run2000 $
 * @since 1.9
 */
public interface Converter {

    /**
     * Convert the given template model into the target object type represented
     * by this converter. Callers of this method should first call {@link #converts}
     * to determine whether this converter can perform the requested conversion.
     *
     * @param model the template model to be converted
     * @return the converted template model to the target object type represented
     * by this converter
     * @throws IllegalArgumentException the template model was of the wrong
     * type for this converter
     * @throws NullPointerException the template model is null
     * @throws TemplateModelException there was an error while retrieving a
     * value from the template model
     */
    Object convert( TemplateModel model ) throws IllegalArgumentException,
            NullPointerException, TemplateModelException;

    /**
     * Determines whether this converter can convert from the given template
     * model into an instance of the given class target. The specified target
     * may be a superclass or superinterface of the object ultimately returned.
     *
     * @param model the template model to be converted
     * @param target the target class for the conversion
     * @return <code>true</code> if this converter can convert the given
     * template model to an instance of the target class, otherwise
     * <code>false</code>
     */
    boolean converts( TemplateModel model, Class<?> target );
}
