/*
 * FreeMarker: a tool that allows Java programs to generate HTML
 * output using templates.
 * Copyright (C) 1998-2005 Benjamin Geer
 * Email: beroul@users.sourceforge.net
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA  02111-1307, USA.
 */
package freemarker.ext.beans2;

import java.util.List;
import java.lang.reflect.Field;
import java.lang.reflect.Method;
import java.beans.PropertyDescriptor;

/**
 * Features encapsulate the methods, properties, and fields of a class or
 * object. This provides quick access to all features indexed by name.
 *
 * @author Nicholas Cull
 * @version $Id: Features.java 1067 2011-06-06 10:55:28Z run2000 $
 * @since 1.9
 */
public interface Features<T> {

    /**
     * Get a list of all methods for the given name. There may be several
     * overloaded methods for a given name.
     *
     * @param name the name of the method to be retrieved
     * @return a List of Method objects for the given name
     */
    List<Method> getMethods( String name );

    /**
     * Get a PropertyDescriptor object for the given name. Property names are
     * unique for a given class, so no overloading is required.
     *
     * @param name the name of the property to be retrieved
     * @return a PropertyDescriptor for the given name
     */
    PropertyDescriptor getProperty( String name );

    /**
     * Get a Field object for the given name. Field names are unique for a given
     * class, so no overloading is required.
     *
     * @param name the name of the field to be retrieved
     * @return a Field object for the given name
     */
    Field getField( String name );

    /**
     * Get an array of all enumeration values for the enum class.
     *
     * @return an array of all enumerated values for this class, or <code>null</code>
     * if the class is not an enum
     */
    T[] getEnumValues();
}
