/*
 * FreeMarker: a tool that allows Java programs to generate HTML
 * output using templates.
 * Copyright (C) 1998-2005 Benjamin Geer
 * Email: beroul@users.sourceforge.net
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA  02111-1307, USA.
 */
package freemarker.ext.beans2;

import java.beans.IndexedPropertyDescriptor;
import java.lang.reflect.Array;
import java.lang.reflect.Method;
import java.lang.reflect.InvocationTargetException;

import freemarker.template.*;

/**
 * Wrapper for an indexed property of a JavaBean.
 *
 * @author Nicholas Cull
 * @version $Id: IndexedPropertyWrapper.java 1067 2011-06-06 10:55:28Z run2000 $
 * @since 1.9
 */
public class IndexedPropertyWrapper implements TemplateListModel2,
        TemplateWriteableIndexedModel, TemplateHashModel,
        TemplateObjectModel {

    /** The indexed property descriptor to be invoked by this wrapper */
    private IndexedPropertyDescriptor ipd;
    /** The object on which the indexed property descriptor will be invoked */
    private Object object;
    /** Convenience value for methods that take zero arguments */
    private static final Object[] emptyParams = new Object[] { };

    /**
     * Constructor that takes an indexed property descriptor and a target object
     * on which it will be invoked.
     *
     * @param object the object on which the indexed property will be called
     * @param ipd the descriptor for the indexed property
     */
    public IndexedPropertyWrapper( Object object, IndexedPropertyDescriptor ipd ) {
        this.object = object;
        this.ipd = ipd;
    }

    /**
     * Is the object empty?
     *
     * @return <code>true</code> if this object is empty, otherwise <code>false</code>
     */
    public boolean isEmpty() throws TemplateModelException {
        return object == null;
    }

    public TemplateModel getAtIndex( long index ) throws TemplateModelException {

        Method method = ipd.getIndexedReadMethod();
        if( method == null ) {
            throw new TemplateModelException( "No indexed property read method for " + ipd.getName() );
        }

        try {
            Object result = Wrapper.invokeMethod( method, object, new Object[] { new Integer( (int) index )});
            return Wrapper.wrap( result );
        } catch( IllegalAccessException e ) {
            throw new TemplateModelException( "Cannot invoke property getter on property: " +
                    e.getMessage(), e );
        } catch( InvocationTargetException e ) {
            throw new TemplateModelException( "Cannot invoke property getter on property: " +
                    e.getTargetException().getMessage(), (Exception)e.getTargetException() );
        }
    }

    public void putAtIndex( long index, TemplateModel model )
            throws TemplateModelException {

        Method method = ipd.getIndexedWriteMethod();
        if( method == null ) {
            throw new TemplateModelException( "No indexed property write method for " + ipd.getName() );
        }

        try {
            Class<?> type = ipd.getIndexedPropertyType();
            if( type != null ) {
                Object value = Unwrapper.unwrap( model, type );
                Wrapper.invokeMethod( method, object, new Object[] { new Integer( (int) index ), value });
                return;
            }
        } catch( IllegalAccessException e ) {
            throw new TemplateModelException( "Cannot invoke property getter on property: " +
                    e.getMessage(), e );
        } catch( InvocationTargetException e ) {
            throw new TemplateModelException( "Cannot invoke property getter on property: " +
                    e.getTargetException().getMessage(), (Exception)e.getTargetException() );
        }

        throw new TemplateModelException( "Cannot invoke indexed property write method for " + ipd.getName() );
    }

    public TemplateIteratorModel templateIterator()
            throws TemplateModelException {

        Method method = ipd.getReadMethod();
        int length;

        try {
            Object result = Wrapper.invokeMethod( method, object, emptyParams );
            length = Array.getLength( result );
        } catch( IllegalAccessException e ) {
            throw new TemplateModelException( "Cannot get invoke property getter on indexed property: " +
                    e.getMessage(), e );
        } catch( InvocationTargetException e ) {
            throw new TemplateModelException( "Cannot get invoke property getter on indexed property: " +
                    e.getTargetException().getMessage(), e );
        }

        return new FastIndexedWriteableIterator( this, 0, length - 1 );
    }

    public void releaseIterator( TemplateIteratorModel iterator ) {
        // Do nothing
    }

    /**
     * Gets a {@link freemarker.template.TemplateModel} from the hash.
     *
     * @param key the name by which the <code>TemplateModel</code>
     *            is identified in the template.
     * @return the <code>TemplateModel</code> referred to by the key,
     *         or <code>null</code> if not found.
     */
    public TemplateModel get( String key ) throws TemplateModelException {
        if( "length".equals( key )) {
            Method method = ipd.getReadMethod();
            try {
                Object result = Wrapper.invokeMethod( method, object, emptyParams );
                return new FastNumber( Array.getLength( result ));
            } catch( IllegalAccessException e ) {
                throw new TemplateModelException( "Cannot get invoke property getter on indexed property: " +
                        e.getMessage(), e );
            } catch( InvocationTargetException e ) {
                throw new TemplateModelException( "Cannot get invoke property getter on indexed property: " +
                        e.getTargetException().getMessage(), e );
            }
        }
        return null;
    }

    /**
     * Return the underlying object to the reflection mechanism. Any variables,
     * methods or properties can be called directly via reflection.
     *
     * @return the underlying object for this template model
     * @throws freemarker.template.TemplateModelException
     *          the object could not be returned
     */
    public Object getAsObject() throws TemplateModelException {
        Method method = ipd.getReadMethod();

        try {
            return Wrapper.invokeMethod( method, object, emptyParams );
        } catch( IllegalAccessException e ) {
            throw new TemplateModelException( "Cannot get invoke property getter on indexed property: " +
                    e.getMessage(), e );
        } catch( InvocationTargetException e ) {
            throw new TemplateModelException( "Cannot get invoke property getter on indexed property: " +
                    e.getTargetException().getMessage(), e );
        }
    }
}
