/*
 * FreeMarker: a tool that allows Java programs to generate HTML
 * output using templates.
 * Copyright (C) 1998-2005 Benjamin Geer
 * Email: beroul@users.sourceforge.net
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA  02111-1307, USA.
 */
package freemarker.ext.beans2;

import java.lang.reflect.Method;
import java.lang.reflect.InvocationTargetException;
import java.util.Iterator;
import java.util.ListIterator;

import freemarker.template.TemplateModelException;
import freemarker.template.TemplateListModel2;
import freemarker.template.TemplateIteratorModel;

/**
 * Wraps a method that returns an iterator. The iterator is then wrapped in an
 * IteratorWrapper or ListIteratorWrapper.
 *
 * @author Nicholas Cull
 * @version $Id: IteratorMethodWrapper.java 1067 2011-06-06 10:55:28Z run2000 $
 * @since 1.9
 */
public class IteratorMethodWrapper implements TemplateListModel2 {
    /** The target object for the Method */
    protected Object object;
    /** The method returning the Iterator */
    protected Method method;
    /** Any parameters to be passed to the Method */
    protected Object[] paramValues;

    /**
     * Default constructor where an iterator method is not provided initially.
     */
    public IteratorMethodWrapper() {
    }

    /**
     * Constructor that takes a method and object parameters to be provided to
     * the method call.
     *
     * @param method the Method to be called
     * @param paramValues parameters to be supplied to the method
     */
    public IteratorMethodWrapper( Method method, Object[] paramValues ) {
        this.method = method;
        this.paramValues = paramValues;
    }

    /**
     * Constructor that takes a method, the object target for the method,
     * and object parameters to be provided to the method call.
     *
     * @param method the Method to be called
     * @param paramValues parameters to be supplied to the method
     * @param object the object that is the target of the method
     */
    public IteratorMethodWrapper( Method method, Object[] paramValues,
                                  Object object ) {
        this.method = method;
        this.paramValues = paramValues;
        this.object = object;
    }

    /**
     * Sets the object target for the method to be called.
     *
     * @param object the object target for the method
     */
    public void setObject( Object object ) {
        this.object = object;
    }

    /**
     * Sets the method to be called by this iterator method wrapper.
     *
     * @param method the method to be called
     */
    public void setMethod( Method method ) {
        this.method = method;
    }

    /**
     * Sets the object values to be passed as parameters to the method call.
     *
     * @param paramValues the values to be passed to the method
     */
    public void setParamValues( Object[] paramValues ) {
        this.paramValues = paramValues;
    }

    public boolean isEmpty() throws TemplateModelException {
        return method == null;
    }

    /**
     * Retrieves an iterator to iterate over this list. Can choose to return
     * <code>null</code> if there are no elements in the list.
     *
     * @return an iterator to iterate over the current list.
     * @throws freemarker.template.TemplateModelException
     *          the next item in the list can't be
     *          retrieved, or no next item exists.
     */
    public TemplateIteratorModel templateIterator()
            throws TemplateModelException {

        Iterator iterator;

        try {
            iterator = (Iterator)Wrapper.invokeMethod( method, object, paramValues );
        } catch( IllegalAccessException e ) {
            throw new TemplateModelException( "Method call " + method.getName() +
                    "() could not be accessed", e );
        } catch( InvocationTargetException e ) {
            throw new TemplateModelException( "Method call " + method.getName() +
                    "() threw an exception: " + e.getTargetException().getMessage(),
                    (Exception)e.getTargetException() );
        }
        if( iterator instanceof ListIterator ) {
            return new ListIteratorWrapper<ListIterator>( (ListIterator) iterator );
        } else if( iterator != null ) {
            return new IteratorWrapper<Iterator>( iterator );
        }
        return null;
    }

    /**
     * <p>Returns the used iterator to the list model. Implement this method
     * when you want to use an object pool of <code>TemplateIterator</code> objects.
     * Otherwise, leave the implementation of this method blank.</p>
     *
     * <p>Note that if the iterator returned in {@link #templateIterator} is
     * <code>null</code>, this method will not be called for the
     * <code>null</code> iterator.</p>
     *
     * @param iterator the iterator to be returned to the object pool, if any
     */
    public void releaseIterator( TemplateIteratorModel iterator ) {
        // Nothing to see here
    }

}
