/*
 * FreeMarker: a tool that allows Java programs to generate HTML
 * output using templates.
 * Copyright (C) 1998-2005 Benjamin Geer
 * Email: beroul@users.sourceforge.net
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA  02111-1307, USA.
 */
package freemarker.ext.beans2;

import java.util.Comparator;
import java.lang.reflect.Method;

/**
 * Compares two given methods by the following algorithm:
 * <ul>
 * <li>Methods with fewer parameters are sorted before methods with more
 * parameters
 * <li>Methods declared from a subclass are sorted before methods declared
 * in a superclass
 * </ul>
 * This determines which method will be matched for a given method call.
 *
 * @author Nicholas Cull
 * @version $Id: MethodComparator.java 1067 2011-06-06 10:55:28Z run2000 $
 * @since 1.9
 */
class MethodComparator implements Comparator<Method> {
    /**
     * Compares its two arguments for order. Returns a negative integer,
     * zero, or a positive integer as the first argument is less than, equal
     * to, or greater than the second.
     *
     * @param m1 the first method to be compared.
     * @param m2 the second method to be compared.
     * @return a negative integer, zero, or a positive integer as the
     * first argument is less than, equal to, or greater than the second.
     * @throws ClassCastException if the arguments' types prevent them from
     * being compared by this Comparator.
     */
    public int compare( Method m1, Method m2 ) {

        int args1 = m1.getParameterTypes().length;
        int args2 = m2.getParameterTypes().length;

        // Put the method with the fewest parameters first
        if( args1 != args2 ) {
            return args1 - args2;
        }

        // Put the subclasses before the superclasses
        Class<?> c1 = m1.getDeclaringClass();
        Class<?> c2 = m2.getDeclaringClass();

        if( c1 == c2 ) {
            return 0;
        }

        if( c1.isAssignableFrom( c2 )) {
            return 1;
        } else {
            return -1;
        }
    }
}
