/*
 * FreeMarker: a tool that allows Java programs to generate HTML
 * output using templates.
 * Copyright (C) 1998-2005 Benjamin Geer
 * Email: beroul@users.sourceforge.net
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA  02111-1307, USA.
 */
package freemarker.ext.beans2;

import java.util.List;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.Enumeration;
import java.lang.reflect.Method;
import java.lang.reflect.InvocationTargetException;

import freemarker.template.TemplateMethodModel2;
import freemarker.template.TemplateModelException;
import freemarker.template.TemplateModel;

/**
 * Wraps a static method identifier. The method identifier is resolved to a Method
 * object when the method is invoked. This allows different Method objects
 * to be invoked depending on the number and type of objects supplied to the
 * call.
 *
 * @author Nicholas Cull
 * @version $Id: StaticMethodWrapper.java 1067 2011-06-06 10:55:28Z run2000 $
 * @since 1.9
 */
public class StaticMethodWrapper implements TemplateMethodModel2 {
    /** Convenience value for methods that take zero arguments */
    protected static final Object[] emptyParamValues = new Object[] { };
    /** Convenience value for methods that need no argument conversions */
    protected static final Converter[] emptyConverters = new Converter[] { };
    /** Convenience value for methods where no arguments were supplied */
    protected static final TemplateModel[] emptyModelArguments = new TemplateModel[] { };

    /** The class on which the method will be called */
    protected Class clazz;
    /** The name of the method, which may be overloaded, to be resolved against the class */
    protected String identifier;

    /** Default constructor where no target class or identifier are supplied. */
    public StaticMethodWrapper() {
    }

    /**
     * Constructor that takes a class on which the method will be called,
     * and an identifier to resolve to a method.
     *
     * @param clazz the class on which the method will be called
     * @param identifier the identifier for the method
     */
    public StaticMethodWrapper( Class clazz, String identifier ) {
        this.clazz = clazz;
        this.identifier = identifier;
    }

    /**
     * Sets the class on which the method will be resolved and called.
     *
     * @param clazz the class on which the method will be called
     */
    public void setClazz( Class clazz ) {
        this.clazz = clazz;
    }

    /**
     * Sets the identifier for the method, possibly overloaded.
     *
     * @param identifier the identifier for the method
     */
    public void setIdentifier( String identifier ) {
        this.identifier = identifier;
    }

    public boolean isEmpty() throws TemplateModelException {
        return clazz == null;
    }

    public TemplateModel exec( List<TemplateModel> arguments ) throws TemplateModelException {
        TemplateModel[] args;

        if( arguments.isEmpty() ) {
            args = emptyModelArguments;
        } else {
            args = new TemplateModel[ arguments.size() ];
            arguments.toArray( args );
        }

        Features features = Wrapper.getClassFeatures( clazz );
        List<Method> methods = features.getMethods( identifier );
        List<Method> candidates = getCandidates( methods, args.length );

        if( candidates.isEmpty() ) {
            throw new TemplateModelException( "Cannot find a candidate method model, wrong number of arguments" );
        }

        // Figure out which method we're trying to call here, unwrap
        // the argument list as appropriate, and call the target method
        for( Method method : candidates ) {
            Class params[] = method.getParameterTypes();

            // Match up argument types
            Converter[] converters = getConversion( params, args );
            if( converters != null ) {
                // Perform the argument conversions and call the method
                Object[] paramValues = convertParameters( converters, args );
                Class returnType = method.getReturnType();

                if( Iterator.class.isAssignableFrom( returnType )) {
                    return new IteratorMethodWrapper( method, paramValues );
                } else if( Enumeration.class.isAssignableFrom( returnType )) {
                    return new EnumerationMethodWrapper( method, paramValues );
                }
                return invokeMethod( method, paramValues );
            }
        }

        throw new TemplateModelException( "Could not find method call " + identifier +
                "() with correct argument types" );
    }

    /**
     * Determine a preliminary list of methods which may be called based solely
     * on the number of arguments expected versus the number of arguments
     * supplied to the method model.
     *
     * @param methods a list of Method objects with the method identifier
     * equals to the identifier for this object
     * @param argSize the number of arguments supplied to the method model
     * @return a new list of methods where the number of arguments expected
     * is equal to the number of arguments supplied
     */
    private static List<Method> getCandidates( List<Method> methods, int argSize ) {
        List<Method> candidates = new ArrayList<Method>( methods.size() );

        for( Method method : methods ) {
            int length = method.getParameterTypes().length;

            if( length == argSize ) {
                candidates.add( method );
            } else if( length > argSize ) {
                break;
            }
        }
        return candidates;
    }

    /**
     * Determine the best conversion from the template model argumentss
     * supplied by the method model invocation to the argument classes expected
     * by a given method.
     *
     * @param params parameter types expected by a given method
     * @param arguments arguments supplied to this method model
     * @return an array of Converters from template model types to the target
     * method class types, or <code>null</code> if one or more parameters
     * could not be converted
     */
    private static Converter[] getConversion( Class params[], TemplateModel[] arguments ) {

        if( params.length == 0 ) {
            return emptyConverters;
        }

        Converter[] converters = new Converter[ params.length ];
        for( int i = 0; i < params.length; i++) {
            converters[ i ] = Unwrapper.getConverter( arguments[ i ], params[ i ]);
            if( converters[ i ] == null ) {
                return null;
            }
        }
        return converters;
    }

    /**
     * Convert the supplied template model parameters using the given converters.
     * The size of both arrays should be the same.
     *
     * @param converters the converters for converting the supplied arguments
     * @param arguments the template model arguments to be converted
     * @return an array of converted template models
     * @throws TemplateModelException there was a problem converting a
     * template model
     */
    private static Object[] convertParameters( Converter[] converters,
            TemplateModel[] arguments ) throws TemplateModelException {

        if( converters.length == 0 ) {
            return emptyParamValues;
        }

        Object[] paramValues = new Object[ converters.length ];
        for( int i = 0; i < converters.length; i++ ) {
            paramValues[ i ] = converters[ i ].convert( arguments[ i ]);
        }
        return paramValues;
    }

    /**
     * Invoke the given static method with the converted parameters.
     *
     * @param method the method to be invoked
     * @param paramValues the parameters to be supplied to the method
     * @return the wrapped return value of the method call
     * @throws TemplateModelException there was a problem invoking the method,
     * or the method threw an exception
     */
    private static TemplateModel invokeMethod( Method method,
            Object[] paramValues ) throws TemplateModelException {

        try {
            Object result = Wrapper.invokeMethod( method, null, paramValues );
            return Wrapper.wrap( result );
        } catch( IllegalAccessException e ) {
            throw new TemplateModelException( "Method call " + method.getName() +
                    "() could not be accessed", e );
        } catch( InvocationTargetException e ) {
            throw new TemplateModelException( "Method call " + method.getName() +
                    "() threw an exception", (Exception)e.getTargetException() );
        }
    }
}
