/*
 * FreeMarker: a tool that allows Java programs to generate HTML
 * output using templates.
 * Copyright (C) 1998-2005 Benjamin Geer
 * Email: beroul@users.sourceforge.net
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA  02111-1307, USA.
 */
package freemarker.ext.beans2;

import java.util.*;
import java.beans.IntrospectionException;
import java.lang.reflect.Method;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Modifier;

import freemarker.template.TemplateModel;
import freemarker.template.FastScalar;
import freemarker.template.FastBoolean;

/**
 * Wraps a given object, class or enum in a FM-Classic template model. An appropriate
 * model will be determined and returned based on introspection of the supplied
 * object or class.
 *
 * @author Nicholas Cull
 * @version $Id: Wrapper.java 1067 2011-06-06 10:55:28Z run2000 $
 * @since 1.9
 */
public class Wrapper {

    /** Maps Class objects to Features objects */
    private static final Map<Class,Features> classFeatures = new Hashtable<Class,Features>();
    private static final Map<Class,Features> objectFeatures = new Hashtable<Class,Features>();

    /** Maps a class object to a prototype of its Wrapper */
    private static final Map<Class,ObjectWrapper> wrappers = new Hashtable<Class,ObjectWrapper>();

    /* Prototypes for various wrapper types */
    private static final ObjectWrapper arrayWrapper = new ArrayWrapper();
    private static final ObjectWrapper beanWrapper = new BeanWrapper();
    private static final ObjectWrapper collectionWrapper = new CollectionWrapper();
    private static final ObjectWrapper enumWrapper = new EnumWrapper();
    private static final ObjectWrapper enumerationWrapper = new EnumerationWrapper();
    private static final ObjectWrapper iterableWrapper = new IterableWrapper();
    private static final ObjectWrapper iteratorWrapper = new IteratorWrapper();
    private static final ObjectWrapper listIteratorWrapper = new ListIteratorWrapper();
    private static final ObjectWrapper listWrapper = new ListWrapper();
    private static final ObjectWrapper mapWrapper = new MapWrapper();
    private static final ObjectWrapper numberWrapper = new NumberWrapper();
    private static final ObjectWrapper propertiesWrapper = new PropertiesWrapper();
    private static final ObjectWrapper resourceBundleWrapper = new ResourceBundleWrapper();

    /**
     * Private constructor indicating this is a static class.
     */
    private Wrapper() {
    }

    /**
     * Wrap the given object in a FM-Classic template model using reflection.
     * Methods, fields, and properties are available to the template engine
     * using standard template model implementations.
     *
     * @param object the object to be wrapped in a template model
     * @return a FM-Classic template model providing an interface to the
     * underlying object
     */
    public static TemplateModel wrap( Object object )
            throws ClassCastException, NullPointerException {

        if( object == null ) {
            return null;
        }
        if( object instanceof String ) {
            return new FastScalar( (String) object );
        }
        if( object instanceof Boolean ) {
            return FastBoolean.getInstance( ((Boolean)object).booleanValue() );
        }

        ObjectWrapper wrapper = getObjectWrapper( object );
        wrapper.setObject( object );
        return (TemplateModel)wrapper;
    }

    /**
     * Wraps the given class or enum in a FM-Classic template model using reflection.
     * Static methods and fields are available to the template engine using
     * standard template model implementations.
     *
     * @param clazz the class to be wrapped in a template model
     * @return a FM-Classic template model providing an interface to the
     * underlying class
     */
    public static <T> TemplateModel wrap( Class<T> clazz ) {

        if( clazz.isEnum() ) {
            return new StaticEnumWrapper( clazz );
        }

        return new StaticClassWrapper( clazz );
    }

    /**
     * Get object features for the given object. Features are cached by the
     * wrapper so that subsequent lookups for objects of the same class are fast.
     * This method is used internally by the various {@link ObjectWrapper}
     * implementations.
     *
     * @param object the object for which features should be determined
     * @return a Features object containing all relevant introspection information
     * @throws IntrospectionException there was a problem determining object
     * features
     */
    static <T> Features<T> getObjectFeatures( Object object ) throws IntrospectionException {
        Class aClass = object.getClass();
        Features<T> features = objectFeatures.get( aClass );

        if( features == null ) {
            features = new ObjectFeatures<T>( aClass );
            objectFeatures.put( aClass, features );
        }
        return features;
    }

    /**
     * Get class features for the given class. Features are cached by the wrapper
     * so that subsequent lookups for this class are fast.
     *
     * @param clazz the class for which features should be determined
     * @return a Features object containing all relevant introspection information
     */
    static <T> Features<T> getClassFeatures( Class<T> clazz ) {
        Features<T> features = classFeatures.get( clazz );

        if( features == null ) {
            features = new ClassFeatures<T>( clazz );
            classFeatures.put( clazz, features );
        }
        return features;
    }

    /**
     * Invokes a given method on a particular object with the given parameters.
     * If the method call fails, try various superinterfaces or superclasses
     * to try and resolve a public method. Examples of when this code may be
     * required include the <code>java.util.Map.Entry</code> implementation
     * classes.
     *
     * @param method the method to be invoked
     * @param target the target object for the object, may be <code>null</code>
     * @param params the parameters to be passed to the method
     * @return the <code>Object</code> result of the method call
     * @throws IllegalAccessException there was a problem invoking the method
     * on the target class
     * @throws InvocationTargetException the target method threw an exception
     */
    static Object invokeMethod( Method method, Object target, Object[] params )
            throws IllegalAccessException, InvocationTargetException {

        Class clazz = method.getDeclaringClass();
        if( Modifier.isPublic( clazz.getModifiers() )) {
            try {
                return method.invoke( target, params );
            } catch( IllegalAccessException e ) {
                // Ignore and continue
            }
        }

        String name = method.getName();
        Class[] paramTypes = method.getParameterTypes();

        // Check interfaces for accessible method
        Class[] interfaces = clazz.getInterfaces();
        for( int i = 0; i < interfaces.length; i++ ) {
            if( Modifier.isPublic( interfaces[ i ].getModifiers() )) {
                try {
                    Method candidate = interfaces[ i ].getMethod( name, paramTypes );
                    return candidate.invoke( target, params );
                } catch( IllegalAccessException e ) {
                    // Ignore and continue
                } catch( NoSuchMethodException e ) {
                    // Do nothing and continue
                }
            }
        }

        // Last resort: check superclasses for accessible method
        // Note: this was particularly slow when profiling under Java 1.4.2
        Class superclass = clazz.getSuperclass();
        while( superclass != null ) {
            if( Modifier.isPublic( superclass.getModifiers() )) {
                try {
                    Method candidate = superclass.getMethod( name, paramTypes );
                    return candidate.invoke( target, params );
                } catch( IllegalAccessException e ) {
                    // Ignore and continue
                } catch( NoSuchMethodException e ) {
                    // Do nothing and continue
                }
            }
            superclass = superclass.getSuperclass();
        }

        throw new IllegalAccessException( "Cannot invoke method " + name +
                ": no accessible methods on " + target.getClass().getName() );
    }

    /**
     * Returns an appropriate object wrapper for the given object. Wrapper
     * information is cached so that future calls to wrap the same class can
     * be immediately wrapped. Wrappers are created using a prototype pattern,
     * so that prototype objects can be stored in a hash table.
     *
     * @param object the object to be wrapped
     * @return an ObjectWrapper corresponding to the given object
     */
    private static ObjectWrapper getObjectWrapper( Object object ) {
        Class clazz = object.getClass();
        ObjectWrapper prototype = wrappers.get( clazz );

        // The results of these tests should be cached
        if( prototype == null ) {
            if( object instanceof Number ) {
                prototype = numberWrapper;
            } else if( object instanceof Iterable ) {
                if( object instanceof Collection ) {
                    if( object instanceof List ) {
                        prototype = listWrapper;
                    } else {
                        prototype = collectionWrapper;
                    }
                } else {
                    prototype = iterableWrapper;
                }
            } else if( object instanceof Map ) {
                if( object instanceof Properties ) {
                    prototype = propertiesWrapper;
                } else {
                    prototype = mapWrapper;
                }
            } else if( object instanceof Iterator ) {
                if( object instanceof ListIterator ) {
                    prototype = listIteratorWrapper;
                } else {
                    prototype = iteratorWrapper;
                }
            } else if( object instanceof Enumeration ) {
                prototype = enumerationWrapper;
            } else if( clazz.isEnum() ) {
                prototype = enumWrapper;
            } else if( clazz.isArray() ) {
                prototype = arrayWrapper;
            } else if( object instanceof ResourceBundle ) {
                prototype = resourceBundleWrapper;
            } else {
                prototype = beanWrapper;
            }
            wrappers.put( clazz, prototype );
        }
        return (ObjectWrapper) prototype.clone();
    }
}
