/*
 * FreeMarker: a tool that allows Java programs to generate HTML
 * output using templates.
 * Copyright (C) 1998-2005 Benjamin Geer
 * Email: beroul@users.sourceforge.net
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA  02111-1307, USA.
 */
package freemarker.ext.format;

import java.util.List;
import java.util.Map;
import java.util.WeakHashMap;
import java.util.Collections;
import java.io.ObjectStreamException;
import java.io.Serializable;
import java.lang.ref.WeakReference;

import freemarker.template.TemplateMethodModel2;
import freemarker.template.TemplateModelException;
import freemarker.template.TemplateModel;
import freemarker.template.FastScalar;
import freemarker.template.expression.ExpressionUtils;
import freemarker.ext.beans2.Unwrapper;

/**
 * Formats a message using a given substitution pattern. The arguments can be
 * wrapped object or <code>TemplateModel</code> values.
 * <p>
 * Formats are specified using the same pattern as the
 * <code>java.text.MessageFormat</code> class. As a performance optimization,
 * some caching of pre-parsed format templates occurs.
 * <p>
 * <b>Usage:</b><br>
 * From java:
 * <pre>
 * TemplateModelRoot root = new SimpleHash();
 *
 * root.put( "formatMessage", freemarker.ext.format.MessageFormat.getInstance() );
 *
 * ...
 * </pre>
 *
 * <p>From your FM-Classic template:</p>
 * <pre>
 * The following creates a directory path:
 * &lt;assign name = "FM-Classic"&gt;
 * &lt;assign description = "template engine"&gt;
 * &lt;assign output = formatMessage( "{0} is a good {1}. {0} {2}!", name, description, "rocks" )&gt;
 * ${output}
 *
 * ...
 * </pre>
 *
 * @author Nicholas Cull
 * @version $Id: MessageFormat.java 1067 2011-06-06 10:55:28Z run2000 $
 * @since 1.9
 */
public class MessageFormat implements TemplateMethodModel2, Serializable {

    private static final MessageFormat instance = new MessageFormat();

    /** Cache for pre-compiled date format objects */
    private static final Map<String, WeakReference<java.text.MessageFormat>> formatCache =
            Collections.synchronizedMap(new WeakHashMap<String, WeakReference<java.text.MessageFormat>>());

    /** Class UUID for serialization */
    private static final long serialVersionUID = -1517804805129161097L;

    /** Private constructor, indicating that this is a singleton. */
    private MessageFormat() { }

    /**
     * Return a singleton instance of this class.
     *
     * @return a new DateFormat object
     */
    public static MessageFormat getInstance() {
        return instance;
    }

    /**
     * Is the object empty?
     *
     * @return <code>true</code> if this object is empty, otherwise <code>false</code>
     */
    public boolean isEmpty() throws TemplateModelException {
        return false;
    }

    /**
     * Executes a method call. The first value is used as a pattern to a
     * MessageFormat object for formatting the remaining arguments. The
     * remaining arguments are unwrapped using defaultUnwrap before being
     * supplied as arguments to the MessageFormat object.
     *
     * @param arguments a <code>List</code> of <code>TemplateModel</code> objects
     *                  containing the values of the arguments passed to the method.
     * @return the <code>TemplateModel</code> produced by the method, or <code>null</code>.
     */
    public TemplateModel exec( List<TemplateModel> arguments ) throws TemplateModelException {
        int size = arguments.size();

        if( size < 2 ) {
            throw new TemplateModelException( "MessageFormat method call must contain at least two arguments" );
        }

        TemplateModel keyModel = arguments.get( 0 );
        String formatTemplate = ExpressionUtils.getAsStringOrEmpty( keyModel );

        Object[] params = new Object[ size - 1 ];

        for( int i = 1; i < size; i++ ) {
            params[ i - 1 ] = Unwrapper.defaultUnwrap( arguments.get( i ));
        }

        java.text.MessageFormat format = getMessageFormat( formatTemplate );
        return new FastScalar( format.format( params ));
    }

    /**
     * Handle caching of format objects. Always return a clone of the cached
     * format object, to avoid multi-threading issues.
     *
     * @param formatTemplate the format template for the value to be formatted
     * @return a MessageFormat corresponding to the given format template
     */
    private static java.text.MessageFormat getMessageFormat( String formatTemplate ) {
        WeakReference<java.text.MessageFormat> ref = formatCache.get( formatTemplate );
        java.text.MessageFormat format =  null;

        if( ref != null ) {
            format = ref.get();
        }

        if( format == null ) {
            format = new java.text.MessageFormat( formatTemplate );
            ref = new WeakReference<java.text.MessageFormat>( format );
            formatCache.put( new String( formatTemplate ), ref );
        }

        return (java.text.MessageFormat)format.clone();
    }

    /**
     * For serialization purposes, always resolve a de-serialized object
     * back to the singleton instance. This ensures that any strict equality
     * tests remain valid after serialization.
     */
    private Object readResolve() throws ObjectStreamException {
        return instance;
    }
}
