/*
 * ====================================================================
 *
 * Copyright (c) 2000 Attila Szegedi.  All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution, if
 *    any, must include the following acknowlegement:
 *       "This product includes the Expose library
 *        (http://www.szegedi.org/expose)."
 *    Alternately, this acknowlegement may appear in the software itself,
 *    if and wherever such third-party acknowlegements normally appear.
 *
 * 4. The names "Expose" and "Attila Szegedi" must not be used to endorse
 *    or promote products derived from this software without prior written
 *    permission. For written permission, please contact expose@szegedi.org.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR ITS CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
 * OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT
 * OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
 * BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
 * OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN
 * IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * ====================================================================
 *
 */

package freemarker.ext.jdom;

import org.jaxen.jdom.JDOMXPath;
import org.jaxen.JaxenException;
import freemarker.template.*;
import freemarker.ext.misc.Methods;
import freemarker.ext.misc.Transforms;
import freemarker.ext.misc.WriterExceptionListener;

import java.io.Writer;
import java.io.IOException;
import java.io.File;
import java.io.OutputStreamWriter;
import java.io.Serializable;
import java.util.*;
import org.jdom.*;
import org.jdom.output.*;

/**
 * <p>Provides a template for wrapping JDOM objects. It is capable of storing not only
 * a single JDOM node, but a list of JDOM nodes at once (hence the name).
 * Each node is an instance of any of the core JDOM node classes (except namespaces,
 * which are not supported at the moment), or String for representing text.</p>
 *
 * <p>See individual method documentation for exact details on how the class works. In
 * short:</p>
 * <ul>
 * <li>{@link #getAsString()} will render all contained nodes as XML fragment,</code>
 * <li>{@link #exec(List)} provides full XPath functionality implemented on top of
 * the <a href="http://jaxen.codehaus.org/" target="_top">Jaxen</a> library,</li>
 * <li>{@link #get(String)} provides node traversal, copying and filtering - somewhat
 * less expressive than XPath, however it does not require the external library and
 * it evaluates somewhat faster.</li>
 * <li>{@link freemarker.template.TemplateListModel2} methods simply iterate over
 * the contained node list.</li>
 * </ul>
 *
 * <p>Finally, <code>TemplateObjectModel</code> is implemented so that operations
 * on the node list can be performed directly.</p>
 *
 * @author Attila Szegedi, szegedia@users.sourceforge.net
 * @version $Id: NodeListModel.java 1159 2013-04-20 05:53:37Z run2000 $
 */
public class NodeListModel implements TemplateHashModel, TemplateMethodModel,
    TemplateIndexedModel, TemplateListModel2, TemplateScalarModel,
    TemplateObjectModel, Serializable
{
    private static final AttributeXMLOutputter OUTPUT = new AttributeXMLOutputter();
    // A convenience singleton for representing a node set without nodes.
    private static final NodeListModel EMPTY = new NodeListModel(null, false);

    // Cache of already parsed XPath expressions
    private static final Map<String, JDOMXPath> XPATH_CACHE =
            new WeakHashMap<String, JDOMXPath>();

    /**
     * A singleton empty scalar. It is an instance of <code>EmptyScalar</code>.
     */
    public static final TemplateScalarModel EMPTY_SCALAR = new EmptyScalar();

    private static final NamedNodeOperator NAMED_CHILDREN_OP = new NamedChildrenOp();
    private static final NamedNodeOperator NAMED_ATTRIBUTE_OP = new NamedAttributeOp();
    private static final NodeOperator ALL_ATTRIBUTES_OP = new AllAttributesOp();
    private static final NodeOperator ALL_CHILDREN_OP = new AllChildrenOp();
    private static final Map OPERATIONS = createOperations();
    private static final Map SPECIAL_OPERATIONS = createSpecialOperations();
    private static final int SPECIAL_OPERATION_COPY = 0;
    private static final int SPECIAL_OPERATION_UNIQUE = 1;
    private static final int SPECIAL_OPERATION_FILTER_NAME = 2;
    private static final int SPECIAL_OPERATION_FILTER_TYPE = 3;
    private static final int SPECIAL_OPERATION_QUERY_TYPE = 4;
    private static final int SPECIAL_OPERATION_REGISTER_NAMESPACE = 5;

    // The contained nodes
    private List<?> nodes;
    // The namespace(s) for these nodes
    private Map<String,org.jdom.Namespace> namespaces;

    /** Serialization UUID for this class. */
    private static final long serialVersionUID = -4082615505723062770L;

    /**
     * Creates a node set template that holds a single {@link Document} node.
     */
    public NodeListModel(Document document)
    {
        nodes = document == null ? Collections.emptyList() : Collections.singletonList(document);
        namespaces = Collections.<String, org.jdom.Namespace>emptyMap();
    }

    /**
     * Creates a node set template that holds a single {@link Element} node.
     */
    public NodeListModel(Element element)
    {
        nodes = element == null ? Collections.emptyList(): Collections.singletonList(element);
        namespaces = Collections.<String, org.jdom.Namespace>emptyMap();
    }

    private NodeListModel(Object object)
    {
        this(object, Collections.<String, Namespace>emptyMap());
    }

    NodeListModel(Object object, Map<String, org.jdom.Namespace> namespaces)
    {
        nodes = object == null ? Collections.EMPTY_LIST : Collections.singletonList(object);
        this.namespaces = (namespaces.size() == 0) ?
                Collections.<String, org.jdom.Namespace>emptyMap() :
                (Map<String, org.jdom.Namespace>)((HashMap<String, org.jdom.Namespace>)namespaces).clone();
    }


    /**
     * Creates a node set template that holds a list of nodes.
     * @param nodes the list of nodes this template should hold. The created template
     * will copy the passed nodes list, so changes to the passed list will not affect
     * the model.
     */
    public NodeListModel(List<?> nodes)
    {
        this(nodes, true);
        namespaces = Collections.<String, org.jdom.Namespace>emptyMap();
    }

    /**
     * Creates a node set template that holds a list of nodes.
     * @param nodes the list of nodes this template should hold.
     * @param copy if true, the created template will copy the passed nodes list,
     * so changes to the passed list will not affect the model. If false, the model
     * will reference the passed list and will sense changes in it, altough no
     * operations on the list will be synchronized.
     */
    public NodeListModel(List<?> nodes, boolean copy)
    {
        this.nodes = copy && nodes != null ? new ArrayList<Object>(nodes) :
                (nodes == null ? Collections.emptyList() : nodes);
        this.namespaces = Collections.<String, org.jdom.Namespace>emptyMap();
    }

    private NodeListModel(List<?> nodes, Map<String, Namespace> namespaces)
    {
        this.nodes = nodes == null ? Collections.emptyList() : nodes;
        this.namespaces = (namespaces.size() == 0) ?
                Collections.<String, org.jdom.Namespace>emptyMap() :
                (Map<String, org.jdom.Namespace>)((HashMap<String, org.jdom.Namespace>)namespaces).clone();
    }

    /**
     * Factory method for creating a NodeListModel that best suits the passed list.
     * @return the singleton empty model for null or empty input list, a model backed
     * by a singleton collection for a single-element input list, or a generic model
     * backed by the passed list, or a copy of it for multiple-element input list.
     */
    public static NodeListModel createNodeListModel(List<?> list, boolean copy)
    {
        if(list == null || list.isEmpty()) return EMPTY;
        if(list.size() == 1) return new NodeListModel(list.get(0));
        return new NodeListModel(list, copy);
    }

    private static NodeListModel createNodeListModel(List<?> list, Map<String, Namespace> namespaces)
    {
        if(list == null || list.isEmpty())
        {
            if(namespaces == Collections.EMPTY_MAP)
            {
                return EMPTY;
            }
            else
            {
                return new NodeListModel(Collections.emptyList(), namespaces);
            }
        }
        if(list.size() == 1) return new NodeListModel(list.get(0), namespaces);
        return new NodeListModel(list, namespaces);
    }

    /**
     * Returns true if this model contains no nodes.
     */
    public boolean isEmpty()
    {
        return nodes.isEmpty();
    }

    /**
     * This method returns the string resulting from concatenation
     * of string representations of its nodes. Each node is rendered using its XML
     * serialization format, while text (String) is rendered as itself. This greatly
     * simplifies creating XML-transformation templates, as to output a node contained
     * in variable x as XML fragment, you simply write ${x} in the template.
     */
    public String getAsString()
    throws
        TemplateModelException
    {
        if(isEmpty())
            return "";

        java.io.StringWriter sw = new java.io.StringWriter(nodes.size() * 128);
        try
        {
            for(Object node : nodes)
            {
                if(node instanceof Element)
                    OUTPUT.output((Element)node, sw);
                else if(node instanceof Attribute)
                    OUTPUT.output((Attribute)node, sw);
                else if(node instanceof String)
                    sw.write( OUTPUT.escapeElementEntities(node.toString() ));
                else if(node instanceof Number)
                    sw.write( OUTPUT.escapeElementEntities(node.toString() ));
                else if(node instanceof Boolean)
                    sw.write( OUTPUT.escapeElementEntities(node.toString() ));
                else if(node instanceof CDATA)
                    OUTPUT.output((CDATA)node, sw);
                else if(node instanceof Text)
                    OUTPUT.output((Text)node, sw);
                else if(node instanceof Document)
                    OUTPUT.output((Document)node, sw);
                else if(node instanceof ProcessingInstruction)
                    OUTPUT.output((ProcessingInstruction)node, sw);
                else if(node instanceof Comment)
                    OUTPUT.output((Comment)node, sw);
                else if(node instanceof DocType)
                    OUTPUT.output((DocType)node, sw);
                else if(node instanceof EntityRef)
                    OUTPUT.output((EntityRef)node, sw);
                else
                    throw new TemplateModelException(node.getClass().getName() + " is not a core JDOM class");
            }
        }
        catch(IOException e)
        {
            throw new TemplateModelException(e.getMessage());
        }
        return sw.toString();
    }

    /**
     * Provides node set traversal as well as special functions: filtering by name,
     * filtering by node type, shallow-copying, and duplicate removal.
     * While not as powerful as the full XPath support built into the
     * {@link #exec(List)} method, it does not require the external Jaxen
     * library to be present at run time. Below are listed the recognized keys.
     * In key descriptions, "applicable to this-and-that node type" means that if
     * a key is applied to a node set that contains a node of non-applicable type
     * a TemplateMethodModel will be thrown. However, you can use <code>_ftype</code>
     * key to explicitly filter out undesired node types prior to applying the
     * restricted-applicability key. Also "current nodes" means nodes contained in this
     * set.
     * <ul>
     *    <li><code>*</code> or <code>_children</code>: all direct element children of current nodes (non-recursive). Applicable
     *  to element and document nodes.</li>
     *    <li><code>@*</code> or <code>_attributes</code>: all attributes of current nodes. Applicable to elements only.</li>
     *    <li><code>_content</code> the complete content of current nodes (non-recursive).
     *  Applicable to elements and documents.</li>
     *    <li><code>_text</code>: the text of current nodes, one string per node (non-recursive).
     *  Applicable to elements, attributes, comments, processing instructions (returns its data)
     *  and CDATA sections.</li>
     *    <li><code>_name</code>: the names of current nodes, one string per node (non-recursive).
     *  Applicable to elements, attributes, entities, processing instructions (returns its target),
     *  doctypes (returns its public ID)</li>
     *    <li><code>_parent</code>: parent elements of current nodes. Applicable to element, attribute, comment,
     *  entity, processing instruction.</li>
     *    <li><code>_ancestor</code>: all ancestors up to root element (recursive) of current nodes. Applicable
     *  to same node types as <code>_parent</code>.</li>
     *    <li><code>_ancestorOrSelf</code>: all ancestors of current nodes plus current nodes. Applicable
     *  to same node types as <code>_parent</code>.</li>
     *    <li><code>_descendant</code>: all recursive descendant element children of current nodes. Applicable to
     *  document and element nodes.
     *    <li><code>_descendantOrSelf</code>: all recursive descendant element children of current nodes
     *  plus current nodes. Applicable to document and element nodes.
     *    <li><code>_document</code>: all documents the current nodes belong to.
     *  Applicable to all nodes except text.
     *    <li><code>_doctype</code>: doctypes of the current nodes.
     *  Applicable to document nodes only.
     *    <li><code>_fname</code>: is a filter-by-name template method model. When called,
     *  it will yield a node set that contains only those current nodes whose name
     *  matches one of names passed as argument. Attribute names should NOT be prefixed with the
     *  at sign (@). Applicable on all node types, however has no effect on unnamed nodes.</li>
     *    <li><code>_ftype</code>: is a filter-by-type template method model. When called,
     *  it will yield a node set that contains only those current nodes whose type matches one
     *  of types passed as argument. You should pass a single string to this method
     *  containing the characters of all types to keep. Valid characters are:
     *  e (Element), a (Attribute), n (Entity), d (Document), t (DocType),
     *  c (Comment), p (ProcessingInstruction), x (text). If the string anywhere contains
     *  the exclamation mark (!), the filter's effect is negated.</li>
     *    <li><code>_type</code>: Returns a one-character String SimpleScalar containing
     *    the typecode of the first node in the node list. Valid characters are:
     *  e (Element), a (Attribute), n (Entity), d (Document), t (DocType),
     *  c (Comment), p (ProcessingInstruction), x (text). If the type of the node
     *  is unknown, returns '?'. If the node list is empty, returns an empty string scalar.</li>
     *    <li><code>_unique</code>: a copy of the current nodes that keeps only the
     *  first occurrence of every node, eliminating duplicates. Duplicates can
     *  occur in the node set by applying uptree-traversals <code>_parent</code>,
     *  <code>_ancestor</code>, <code>_ancestorOrSelf</code>, and <code>_document</code>.
     *  I.e. <code>foo._children._parent</code> will return a node set that has
     *  duplicates of nodes in foo - each node will have the number of occurrences
     *  equal to the number of its children. In these cases, use
     *  <code>foo._children._parent._unique</code> to eliminate duplicates. Applicable
     *  to all node types.
     *    <li><code>_copy</code>: a copy of the current node set. It is a shallow copy that
     *  shares the underlying node list with this node set, however it has a
     *  separate iterator, so it can be used to guarantee correct behavior of
     *  TemplateListModel methods in multithreaded environment. However, this concern
     *  arises only when using a shared NodeListModel instance, i.e. the one directly
     *  bound into the <code>TemplateModelRoot</code>. All instances returned through {@link #get(String)}
     *  and {@link #exec(List)} during template processing are created new, thus they
     *  are private to the thread that processes the template and need not be copied
     *  before they are used as a TemplateListModel. Applicable to all node types.</li>
     *    <li><code>_registerNamespace(prefix, uri)</code>: register a XML namespace
     *  with the specified prefix and URI for the current node list and all node
     *  lists that are derived from the current node list. After registering,
     *  you can use the <code>nodelist["prefix:localname"]</code> or
     *  <code>nodelist["@prefix:localname"]</code> syntaxes to reach elements and
     *  attributes whose names are namespace scoped. Note that the namespace
     *  prefix need not match the actual prefix used by the XML document itself
     *  since namespaces are compared solely by their URI. You can also register
     *  namespaces from Java code using the
     *  {@link #registerNamespace(String, String)} method.
     * </li>
     *    <li><code>@attributeName</code>: named attributes of current nodes. Applicable to
     *  elements, doctypes and processing instructions. On doctypes it supports
     *  attributes <code>publicId</code>, <code>systemId</code> and <code>elementName</code>. On processing
     *  instructions, it supports attributes <code>target</code> and <code>data</code>, as
     *  well as any other attribute name specified in data as <code>name="value"</code> pair.
     *  The attribute nodes for doctype and processing instruction are synthetic, and
     *  as such have no parent. Note, however that <code>@*</code> does NOT operate on
     *  doctypes or processing instructions.</li>
     *    <li>any other key: element children of current nodes with name matching the key.
     *  This allows for convenience child traversal in <code>book.chapter.title</code> style syntax.
     *  Note that <code>nodeset.childname</code> is technically equivalent to
     *  <code>nodeset._children._fname("childname")</code>, but is both shorter to write
     *  and evaluates faster. Applicable to document and element nodes.</li>
     * </ul>
     * The order of nodes in the resulting set is the order of evaluation of the key
     * on each node in this set from left to right. Evaluation of the key on a single
     * node always yields the results in "natural" order (that of the document preorder
     * traversal), even for uptree traversals. As a consequence, if this node set's nodes
     * are listed in natural order, applying any of the keys will produce a node set that
     * is also naturally ordered. As a special case, all node sets that are directly or
     * indirectly generated from a single Document or Element node through repeated
     * invocations of this method will be naturally ordered.
     * @param key a key that identifies a required set of nodes
     * @return a new NodeListModel that represents the requested set of nodes.
     */
    public TemplateModel get(String key)
    throws
        TemplateModelException
    {
        if(isEmpty())
            return EMPTY;

        if(key == null || key.length() == 0)
            throw new TemplateModelException("Invalid key [" + key + ']');

        NodeOperator op = null;
        NamedNodeOperator nop = null;
        String name = null;

        switch(key.charAt(0))
        {
            case '@':
            {
                if(key.length() != 2 || key.charAt(1) != '*')
                {
                    // Generic attribute key
                    nop = NAMED_ATTRIBUTE_OP;
                    name = key.substring(1);
                }
                else
                    // It is @*
                    op = ALL_ATTRIBUTES_OP;

                break;
            }
            case '*':
            {
                if(key.length() == 1)
                    op = ALL_CHILDREN_OP;
                else
                    // Explicitly disallow any other identifier starting with asterisk
                    throw new TemplateModelException("Invalid key [" + key + ']');

                break;
            }
            case 'x':
            case '_':
            {
                op = (NodeOperator)OPERATIONS.get(key);
                if(op == null)
                {
                    // Some special operation?
                    Integer specop = (Integer)SPECIAL_OPERATIONS.get(key);
                    if(specop != null)
                    {
                        switch(specop.intValue())
                        {
                            case SPECIAL_OPERATION_COPY:
                                return new NodeListModel(nodes, namespaces);
                            case SPECIAL_OPERATION_UNIQUE:
                                return new NodeListModel(removeDuplicates(nodes), namespaces);
                            case SPECIAL_OPERATION_FILTER_NAME:
                                return new NameFilter();
                            case SPECIAL_OPERATION_FILTER_TYPE:
                                return new TypeFilter();
                            case SPECIAL_OPERATION_QUERY_TYPE:
                                return getType();
                            case SPECIAL_OPERATION_REGISTER_NAMESPACE:
                                return new RegisterNamespace();
                        }
                    }
                }
                break;
            }
        }

        if(op == null && nop == null)
        {
            nop = NAMED_CHILDREN_OP;
            name = key;
        }

        List<?> list;
        if(op != null)
            list = evaluateElementOperation(op, nodes);
        else
        {
            String localName = name;
            Namespace namespace = Namespace.NO_NAMESPACE;
            int colon = name.indexOf(':');
            if(colon != -1)
            {
                localName = name.substring(colon + 1);
                String nsPrefix = name.substring(0, colon);
                namespace = namespaces.get(nsPrefix);
                if(namespace == null)
                {
                    if("xml".equals(nsPrefix))
                        namespace = Namespace.XML_NAMESPACE;
                    else
                        throw new TemplateModelException("Unregistered namespace prefix '" + nsPrefix + '\'');
                }
            }

            list = evaluateNamedElementOperation(nop, localName, namespace, nodes);
        }
        return createNodeListModel(list, namespaces);
    }

    private TemplateModel getType()
    {
        if(nodes.size() == 0)
            return new SimpleScalar("");
        Object firstNode = nodes.get(0);
        char code;
        if(firstNode instanceof Element)
            code = 'e';
        else if(firstNode instanceof Text || firstNode instanceof String)
            code = 'x';
        else if(firstNode instanceof Attribute)
            code = 'a';
        else if(firstNode instanceof EntityRef)
            code = 'n';
        else if(firstNode instanceof Document)
            code = 'd';
        else if(firstNode instanceof DocType)
            code = 't';
        else if(firstNode instanceof Comment)
            code = 'c';
        else if(firstNode instanceof ProcessingInstruction)
            code = 'p';
        else
            code = '?';
        return new SimpleScalar(new String(new char[] { code }));
    }

    /**
     * Retrieves the i-th element of the node list.
     *
     * @param i the index of the underlying value we're interested in
     * @return a <code>TemplateModel</code> representing the value for the given index
     * @throws TemplateModelException the value could not be determined, possibly
     * due to an index out-of-bounds, or an otherwise undefined value
     */
    public TemplateModel getAtIndex(long i) throws TemplateModelException {
        try
        {
            return new NodeListModel(nodes.get((int)i), namespaces);
        }
        catch(IndexOutOfBoundsException e)
        {
            throw new TemplateModelException("Index out of bounds: " + e.getMessage(), e);
        }
    }

    /**
     * Retrieves an iterator to iterate over this list.
     *
     * @return an iterator to iterate over the current list.
     * @throws TemplateModelException the next item in the list can't be
     * retrieved, or no next item exists.
     */
    public TemplateIteratorModel templateIterator() throws TemplateModelException {
        return new NodeListIterator( nodes, namespaces );
    }

    /**
     * Returns the used iterator to the list model.
     *
     * @param iterator the iterator to be returned to the object pool, if any
     */
    public void releaseIterator(TemplateIteratorModel iterator) {
        // No implementation, since we don't implement an iterator pool
    }

    /**
     * Applies an XPath expression to the node set and returns the resulting node set.
     * In order for this method to work, your application must have access
     * <a href="http://jaxen.codehaus.org/">Jaxen</a> library classes. The
     * implementation does cache the parsed format of XPath expressions in a weak hash
     * map, keyed by the string representation of the XPath expression. As the string
     * object passed as the argument is usually kept in the parsed FM-Classic template,
     * this ensures that each XPath expression is parsed only once during the lifetime
     * of the FM-Classic template that contains it.
     * @param arguments the list of arguments. Must contain exactly one string that is
     * the XPath expression you wish to apply.
     * @return a NodeListModel representing the nodes that are the result of application
     * of the XPath to the current node set.
     */
    public TemplateModel exec(List<String> arguments)
    throws
        TemplateModelException
    {
        if(arguments == null || arguments.size() != 1)
            throw new TemplateModelException("Exactly one argument required for execute() on NodeTemplate");

        String xpathString = arguments.get(0);
        JDOMXPath xpath;
        try {
            synchronized(XPATH_CACHE)
            {
                xpath = XPATH_CACHE.get(xpathString);
                if(xpath == null)
                {
                    xpath = new JDOMXPath(xpathString);
                    XPATH_CACHE.put(xpathString, xpath);
                }
            }
            Object result = xpath.evaluate(nodes);

            // Deal with the possible return types from Jaxen
            if( result instanceof List ) {
                return createNodeListModel((List)result, false);
            } else if( result instanceof String ) {
                return new FastScalar( (String) result );
            } else if( result instanceof Number ) {
                return new FastNumber( (Number) result );
            } else if( result instanceof Boolean ) {
                return FastBoolean.getInstance( ((Boolean)result).booleanValue() );
            } else {
                return null;
            }
        } catch (JaxenException e) {
            throw new TemplateModelException(
                    "XPath could not be determined: " + e.getMessage(), e );
        }
    }

    /**
     * Registers an XML namespace with this node list. Once registered, you can
     * refer to the registered namespace using its prefix in the
     * {@link #get(String)} method from this node list and all other
     * node lists that are derived from this node list. Use the
     * <code>nodelist["prefix:localname"]</code> or the
     * <code>nodelist["@prefix:localname"]</code> syntax to reach elements and
     *  attributes whose names are namespace scoped. Note that the namespace
     * prefix need not match the actual prefix used by the XML document itself
     * since namespaces are compared solely by their URI. You can also register
     * namespaces during template evaluation using the
     * <code>nodelist._registerNamespace(prefix, uri)</code> syntax in the template.
     * This mechanism is completely independent from the namespace declarations
     * in the XML document itself; its purpose is to give you an easy way
     * to refer to namespace-scoped elements in {@link #get(String)}.
     */
    public void registerNamespace(String prefix, String uri)
    {
        if(namespaces == Collections.EMPTY_MAP)
            namespaces = new HashMap<String, org.jdom.Namespace>();
        namespaces.put(prefix, Namespace.getNamespace(prefix, uri));
    }

    /**
     * Return the underlying object to the reflection mechanism in the
     * <code>freemarker.ext.beans</code> package. Any variables, methods
     * or properties can be called directly via reflection.</p>
     *
     * @return the underlying object for this template model
     * @throws TemplateModelException the object could not be returned
     */
    public Object getAsObject() throws TemplateModelException {
        return nodes;
    }

    private interface NodeOperator
    {
        List<?> operate(Object node)
        throws
            TemplateModelException;
    }

    private interface NamedNodeOperator
    {
        List<?> operate(Object node, String localName, Namespace namespace)
        throws
            TemplateModelException;
    }

    private static final class AllChildrenOp implements NodeOperator
    {
        public List<?> operate(Object node)
        throws
            TemplateModelException
        {
            if(node instanceof Element)
                return ((Element)node).getChildren();
            else if(node instanceof Document)
            {
                Element root = ((Document)node).getRootElement();
                return root == null ? Collections.EMPTY_LIST : Collections.singletonList(root);
            }
            else
                throw new TemplateModelException("_children can not be applied on " + node.getClass());
        }
    }

    private static final class NamedChildrenOp implements NamedNodeOperator
    {
        public List<?> operate(Object node, String localName, Namespace namespace)
        throws
            TemplateModelException
        {
            if(node instanceof Element)
            {
                return ((Element)node).getChildren(localName, namespace);
            }
            else if(node instanceof Document)
            {
                Element root = ((Document)node).getRootElement();
                if(root != null &&
                   root.getName().equals(localName) &&
                   root.getNamespaceURI().equals(namespace.getURI()))
                {
                    return Collections.singletonList(root);
                }
                else
                    return Collections.EMPTY_LIST;
            }
            else
                throw new TemplateModelException("_namedChildren can not be applied on " + node.getClass());
        }
    }

    private static final class AllAttributesOp implements NodeOperator
    {
        public List<?> operate(Object node)
        throws
            TemplateModelException
        {
            if(node instanceof Element)
                return ((Element)node).getAttributes();
            else
                throw new TemplateModelException("_attributes can not be applied on " + node.getClass());
        }
    }

    private static final class NamedAttributeOp implements NamedNodeOperator
    {
        public List<?> operate(Object node, String localName, Namespace namespace)
        throws
            TemplateModelException
        {
            Attribute attr = null;
            if(node instanceof Element)
            {
                Element element = (Element)node;
                attr = element.getAttribute(localName, namespace);
            }
            else if(node instanceof ProcessingInstruction)
            {
                ProcessingInstruction pi = (ProcessingInstruction)node;
                if("target".equals(localName))
                    attr = new Attribute("target", pi.getTarget());
                else if("data".equals(localName))
                    attr = new Attribute("data", pi.getData());
                else
                    attr = new Attribute(localName, pi.getPseudoAttributeValue(localName));
            }
            else if(node instanceof DocType)
            {
                DocType doctype = (DocType)node;
                if("publicId".equals(localName))
                    attr = new Attribute("publicId", doctype.getPublicID());
                else if("systemId".equals(localName))
                    attr = new Attribute("systemId", doctype.getSystemID());
                else if("elementName".equals(localName))
                    attr = new Attribute("elementName", doctype.getElementName());
            }
            else
                throw new TemplateModelException("_attributes can not be applied on " + node.getClass());

            return attr == null ? Collections.EMPTY_LIST : Collections.singletonList(attr);
        }
    }

    private static final class NameOp implements NodeOperator
    {
        public List<?> operate(Object node)
        throws
            TemplateModelException
        {
            if(node instanceof Element)
                return Collections.singletonList(((Element)node).getName());
            else if(node instanceof Attribute)
                return Collections.singletonList(((Attribute)node).getName());
            else if(node instanceof EntityRef)
                return Collections.singletonList(((EntityRef)node).getName());
            else if(node instanceof ProcessingInstruction)
                return Collections.singletonList(((ProcessingInstruction)node).getTarget());
            else if(node instanceof DocType)
                return Collections.singletonList(((DocType)node).getPublicID());
            else
                throw new TemplateModelException("_name can not be applied on " + node.getClass());
        }
    }

    private static Parent getParent(Object node)
    throws
        TemplateModelException
    {
        if(node instanceof Content)
            return ((Content)node).getParent();
        else if(node instanceof Attribute)
            return ((Attribute)node).getParent();
        else
            throw new TemplateModelException("_parent can not be applied on " + node.getClass());
    }

    private static final class ParentOp implements NodeOperator
    {
        public List<?> operate(Object node)
        throws
            TemplateModelException
        {
            Parent parent = getParent(node);
            return parent == null ? Collections.EMPTY_LIST : Collections.singletonList(parent);
        }
    }

    private static final class AncestorOp implements NodeOperator
    {
        public List<?> operate(Object node)
        throws
            TemplateModelException
        {
            Parent parent = getParent(node);
            if(parent == null) return Collections.EMPTY_LIST;
            LinkedList<Object> list = new LinkedList<Object>();
            do
            {
                list.addFirst(parent);
                parent = parent.getParent();
            }
            while(parent != null);
            return list;
        }
    }

    private static final class AncestorOrSelfOp implements NodeOperator
    {
        public List<?> operate(Object node)
        throws
            TemplateModelException
        {
            Parent parent = getParent(node);
            if(parent == null) return Collections.singletonList(node);
            LinkedList<Object> list = new LinkedList<Object>();
            list.addFirst(node);
            do
            {
                list.addFirst(parent);
                parent = parent.getParent();
            }
            while(parent != null);
            return list;
        }
    }

    private static class DescendantOp implements NodeOperator
    {
        public List<?> operate(Object node)
        throws
            TemplateModelException
        {
            List<Element> list = new LinkedList<Element>();
            Element element;

            if(node instanceof Element) {
                element = (Element)node;
            } else if(node instanceof Document) {
                element = ((Document)node).getRootElement();
                list.add(element);
            } else {
                throw new TemplateModelException("_descendant can not be applied on " + node.getClass());
            }

            addChildren(element, list);
            return list;
        }

        protected void addChildren(Element element, List<Element> list)
        {
            List<Element> children = element.getChildren();
            for(Element child : children)
            {
                list.add(child);
                addChildren(child, list);
            }
        }
    }

    private static final class DescendantOrSelfOp extends DescendantOp
    {
        public List<?> operate(Object node)
        throws
            TemplateModelException
        {
            Element element;
            if(node instanceof Element)
                element = (Element)node;
            else if(node instanceof Document)
                element = ((Document)node).getRootElement();
            else
                throw new TemplateModelException("_descendantOrSelf can not be applied on " + node.getClass());

            List<Element> list = new LinkedList<Element>();
            list.add(element);
            addChildren(element, list);
            return list;
        }
    }

    private static final class DocumentOp implements NodeOperator
    {
        public List<?> operate(Object node)
        throws
            TemplateModelException
        {
            Document doc;
            if(node instanceof Parent)
                doc = ((Parent)node).getDocument();
            else if(node instanceof Content)
                doc = ((Content)node).getDocument();
            else if(node instanceof Attribute)
                doc = ((Attribute)node).getDocument();
            else
                throw new TemplateModelException("_document can not be applied on " + node.getClass());

            return doc == null ? Collections.EMPTY_LIST : Collections.singletonList(doc);
        }
    }

    private static final class DocTypeOp implements NodeOperator
    {
        public List<?> operate(Object node)
        throws
            TemplateModelException
        {
            if(node instanceof Document)
            {
                DocType doctype = ((Document)node).getDocType();
                return doctype == null ? Collections.EMPTY_LIST : Collections.singletonList(doctype);
            }
            else
                throw new TemplateModelException("_doctype can not be applied on " + node.getClass());
        }
    }

    private static final class ContentOp implements NodeOperator
    {
        public List<?> operate(Object node)
        throws
            TemplateModelException
        {
            if(node instanceof Parent)
                return ((Parent)node).getContent();

            throw new TemplateModelException("_content can not be applied on " + node.getClass());
        }
    }

    private static final class TextOp implements NodeOperator
    {
        public List<?> operate(Object node)
        throws
            TemplateModelException
        {
            if(node instanceof Element)
                return Collections.singletonList(((Element)node).getTextTrim());
            if(node instanceof Attribute)
                return Collections.singletonList(((Attribute)node).getValue());
            if(node instanceof Text)
                return Collections.singletonList(((Text)node).getText());
            if(node instanceof Comment)
                return Collections.singletonList(((Comment)node).getText());
            if(node instanceof ProcessingInstruction)
                return Collections.singletonList(((ProcessingInstruction)node).getData());
            throw new TemplateModelException("_text can not be applied on " + node.getClass());
        }
    }

    private static List evaluateElementOperation(NodeOperator op, List nodes)
    throws
        TemplateModelException
    {
        int s = nodes.size();
        List<?>[] lists = new List<?>[s];
        int l = 0;
        {
            int i = 0;
            for(Object node : nodes)
            {
                List<?> list = op.operate(node);
                lists[i++] = list;
                l += list.size();
            }
        }
        List<Object> retval = new ArrayList<Object>(l);
        for(int i = 0; i < s; ++i)
            retval.addAll(lists[i]);
        return retval;
    }

    private static List evaluateNamedElementOperation(NamedNodeOperator op, String localName, Namespace namespace, List nodes)
    throws
        TemplateModelException
    {
        int s = nodes.size();
        List<?>[] lists = new List<?>[s];
        int l = 0;
        {
            int i = 0;
            for(Object node : nodes)
            {
                List<?> list = op.operate(node, localName, namespace);
                lists[i++] = list;
                l += list.size();
            }
        }
        List<Object> retval = new ArrayList<Object>(l);
        for(int i = 0; i < s; ++i)
            retval.addAll(lists[i]);
        return retval;
    }

    private static List<Object> removeDuplicates(List<?> list)
    {
        int s = list.size();
        ArrayList<Object> ulist = new ArrayList<Object>(s);
        Set<Object> set = new HashSet<Object>(s * 4 / 3, .75f);

        for(Object o : list) {
            if(set.add(o))
                ulist.add(o);
        }

        ulist.trimToSize();
        return ulist;
    }

    private static Map<String, NodeListModel.NodeOperator> createOperations()
    {
        ContentOp contentOp = new ContentOp();
        TextOp textOp = new TextOp();
        NameOp nameOp = new NameOp();
        ParentOp parentOp = new ParentOp();
        AncestorOp ancestorOp = new AncestorOp();
        AncestorOrSelfOp ancestorOrSelfOp = new AncestorOrSelfOp();
        DescendantOp descendantOp = new DescendantOp();
        DescendantOrSelfOp descendantOrSelfOp = new DescendantOrSelfOp();
        DocumentOp documentOp = new DocumentOp();
        DocTypeOp docTypeOp = new DocTypeOp();

        Map<String, NodeListModel.NodeOperator> map =
                new HashMap<String, NodeListModel.NodeOperator>();

        map.put("_children", ALL_CHILDREN_OP);
        map.put("_attributes", ALL_ATTRIBUTES_OP);
        map.put("_content", contentOp);
        map.put("_text", textOp);
        map.put("_name", nameOp);
        map.put("_parent", parentOp);
        map.put("_ancestor", ancestorOp);
        map.put("_ancestorOrSelf", ancestorOrSelfOp);
        map.put("_descendant", descendantOp);
        map.put("_descendantOrSelf", descendantOrSelfOp);
        map.put("_document", documentOp);
        map.put("_doctype", docTypeOp);

        // These are in for backward compatibility
        map.put("x_children", ALL_CHILDREN_OP);
        map.put("x_attributes", ALL_ATTRIBUTES_OP);
        map.put("x_content", contentOp);
        map.put("x_mixed", contentOp);
        map.put("x_text", textOp);
        map.put("x_name", nameOp);
        map.put("x_parent", parentOp);
        map.put("x_ancestor", ancestorOp);
        map.put("x_ancestorOrSelf", ancestorOrSelfOp);
        map.put("x_descendant", descendantOp);
        map.put("x_descendantOrSelf", descendantOrSelfOp);
        map.put("x_document", documentOp);
        map.put("x_doctype", docTypeOp);

        return map;
    }

    private static Map<String, Integer> createSpecialOperations()
    {
        Map<String, Integer> map = new HashMap<String, Integer>();

        Integer copy = SPECIAL_OPERATION_COPY;
        Integer unique = SPECIAL_OPERATION_UNIQUE;
        Integer fname = SPECIAL_OPERATION_FILTER_NAME;
        Integer ftype = SPECIAL_OPERATION_FILTER_TYPE;
        Integer type = SPECIAL_OPERATION_QUERY_TYPE;
        Integer regns = SPECIAL_OPERATION_REGISTER_NAMESPACE;

        map.put("_copy", copy);
        map.put("_unique", unique);
        map.put("_fname", fname);
        map.put("_ftype", ftype);
        map.put("_type", type);
        map.put("_registerNamespace", regns);

        // These are in for backward compatibility
        map.put("x_copy", copy);
        map.put("x_unique", unique);
        map.put("x_fname", fname);
        map.put("x_ftype", ftype);
        map.put("x_type", type);

        return map;
    }

    private final class RegisterNamespace implements TemplateMethodModel
    {
        public boolean isEmpty()
        {
            return false;
        }

        public TemplateModel exec(List<String> arguments)
        throws
            TemplateModelException
        {
            if(arguments.size() != 2)
                throw new TemplateModelException("_registerNamespace(prefix, uri) requires two arguments");

            registerNamespace(arguments.get(0), arguments.get(1));

            return EMPTY_SCALAR;
        }
    }

    private final class NameFilter implements TemplateMethodModel
    {
        public boolean isEmpty()
        {
            return false;
        }

        public TemplateModel exec(List<String> arguments)
        {
            Set<String> names = new HashSet<String>(arguments);
            List<Object> list = new LinkedList<Object>(nodes);
            Iterator it = list.iterator();
            while(it.hasNext())
            {
                Object node = it.next();
                String name = null;
                if(node instanceof Element)
                    name = ((Element)node).getName();
                else if(node instanceof Attribute)
                    name = ((Attribute)node).getName();
                else if(node instanceof ProcessingInstruction)
                    name = ((ProcessingInstruction)node).getTarget();
                else if(node instanceof EntityRef)
                    name = ((EntityRef)node).getName();
                else if(node instanceof DocType)
                    name = ((DocType)node).getPublicID();

                if(name == null || !names.contains(name))
                    it.remove();
            }
            return createNodeListModel(list, true);
        }
    }

    private final class TypeFilter implements TemplateMethodModel
    {
        public boolean isEmpty()
        {
            return false;
        }

        public TemplateModel exec(List<String> arguments)
        throws
            TemplateModelException
        {
            if(arguments == null || arguments.size() == 0)
                throw new TemplateModelException("_type expects exactly one argument");
            String arg = arguments.get(0);
            boolean invert = arg.indexOf('!') != -1;
            // NOTE: true in each of these variables means 'remove', not 'keep'
            // This is so we don't invert their values in the loop. So,
            // a is true <--> (a is not present in the string) xor invert.
            boolean a = invert != (arg.indexOf('a') == -1);
            boolean c = invert != (arg.indexOf('c') == -1);
            boolean d = invert != (arg.indexOf('d') == -1);
            boolean e = invert != (arg.indexOf('e') == -1);
            boolean n = invert != (arg.indexOf('n') == -1);
            boolean p = invert != (arg.indexOf('p') == -1);
            boolean x = invert != (arg.indexOf('x') == -1);

            List<Object> list = new LinkedList<Object>(nodes);
            Iterator it = list.iterator();
            while(it.hasNext())
            {
                Object node = it.next();
                if((node instanceof Element && e)
                || (node instanceof Attribute && a)
                || (node instanceof String && x)
                || (node instanceof Text && x)
                || (node instanceof ProcessingInstruction && p)
                || (node instanceof Comment && c)
                || (node instanceof EntityRef && n)
                || (node instanceof Document && d)
                || (node instanceof DocType && d))
                    it.remove();
            }
            return createNodeListModel(list, true);
        }
    }

    /**
     * <p>Loads a template from a file passed as the first argument, loads an XML
     * document from either the second argument or standard input, passes it
     * to the template and writes the result of template processing to
     * standard output.</p>
     *
     * <p>Special variables are used within template processing to provide access
     * to the XML document and related functions. These are:</p>
     * <ul>
     * <li><code>document</code> -- the root of the XML document</li>
     * <li><code>methods</code> -- method models provided by the
     * {@link freemarker.ext.misc.Methods} model</li>
     * <li><code>transforms</code> -- transform models provided by the
     * {@link freemarker.ext.misc.Transforms} model</li>
     * </ul>
     *
     * @param args command-line arguments. The first argument should be the
     * template to be processed, the second argument should be the XML document
     */
    public static void main(String[] args) throws Exception {
        org.jdom.input.SAXBuilder builder = new org.jdom.input.SAXBuilder();
        Document document;

        if( args.length > 1 ) {
            document = builder.build(new File(args[1]));
        } else {
            document = builder.build(System.in);
        }

        Writer errWriter = new OutputStreamWriter( System.err );
        WriterExceptionListener listener = new WriterExceptionListener( errWriter );
        TemplateRuntimeHandler handler = new SimpleEventAdapter( listener );

        TemplateWriteableHashModel model = new FastHash();
        model.put("document", new NodeListModel(document));
        model.put("methods", Methods.getInstance());
        model.put("transforms", Transforms.getInstance());
        Template template = TemplateFactory.INSTANCE.createFromFile(args[0]);
        java.io.Writer writer = new java.io.OutputStreamWriter(System.out);
        TemplateProcessorParameters p = TemplateProcessorParameters
                .newInstance(writer)
                .withModel(model)
                .withEventHandler(handler);
        template.process(p);
        errWriter.flush();
        writer.flush();
        writer.close();
    }

    private static final class AttributeXMLOutputter extends XMLOutputter
    {
        public void output(Attribute attribute, Writer out)
        throws
            IOException
        {
      out.write(" ");
      out.write(attribute.getQualifiedName());
      out.write("=");

      out.write("\"");
      out.write(escapeAttributeEntities(attribute.getValue()));
      out.write("\"");
        }
    }

    /**
     * This inner class is public so that reflected method invocations have access
     * (kind of implementation detail).
     * It is a singleton, whose single instance is {@link NodeListModel#EMPTY_SCALAR}.
     */
    private static final class EmptyScalar implements TemplateScalarModel,
            Serializable
    {
        /** Serialization UUID for this class. */
        private static final long serialVersionUID = -4867105975619160749L;

        /**
         * Returns a zero-length string.
         */
        public String getAsString() { return ""; }
        /**
         * Returns true.
         */
        public boolean isEmpty() { return true; }
    }
}
