/*
 * FreeMarker: a tool that allows Java programs to generate HTML
 * output using templates.
 * Copyright (C) 1998-2004 Benjamin Geer
 * Email: beroul@users.sourceforge.net
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA  02111-1307, USA.
 */
package freemarker.ext.misc;

import java.util.List;
import java.util.Collection;
import java.util.Map;
import java.io.Serializable;
import java.io.ObjectStreamException;

import freemarker.template.*;

/**
 * <p>Determines the size of the underlying <code>Collection</code> or
 * <code>Map</code> object given a <code>TemplateObjectModel</code>. Does this
 * by calling the <code>getAsObject</code> method. If the object is of the
 * expected type, the size is returned as a <code>FastNumber</code>. Otherwise,
 * a <code>TemplateModelException</code> is thrown.</p>
 *
 * <p>Usage:<br />
 * From java:</p>
 * <pre>
 * TemplateModelRoot root = new SimpleHash();
 *
 * root.put( "size", freemarker.ext.misc.CollectionSize.getInstance() );
 *
 * ...
 * </pre>
 *
 * <p>From your FM-Classic template:</p>
 * <pre>
 * &lt;assign list1 = [ "test1", "test2", "test3" ]&gt;
 * &lt;assign listSize = size( list1 )&gt;
 * &lt;p&gt;The size of list1 is: ${listSize}&lt;/p&gt;
 *
 * ...
 * </pre>
 *
 * <p><b>Note:</b><br />
 * <code>CollectionSize</code> is a singleton instance. Use the
 * {@link #getInstance()} method to retrieve instances of this model.</p>
 *
 * @author Nicholas Cull
 * @version $Id: CollectionSize.java 1101 2013-04-01 04:17:32Z run2000 $
 * @since 1.8
 */
public final class CollectionSize implements TemplateMethodModel2,
        Serializable {

    private static final CollectionSize instance = new CollectionSize();

    /** Serialization UUID for this class. */
    private static final long serialVersionUID = -9103419191882498128L;

    /** Private constructor, to indicate this is a singleton class. */
    private CollectionSize() {
    }

    /**
     * Retrieve a singleton instance of this class. Since there is no
     * state information held between calls, there is only a need for
     * one instance.
     *
     * @return an instance of this <code>CollectionSize</code> class
     */
    public static CollectionSize getInstance() {
        return instance;
    }

    /**
     * Is the object empty?
     *
     * @return <code>false</code>, since the method is never empty
     */
    public boolean isEmpty() throws TemplateModelException {
        return false;
    }

    /**
     * <p>Retrieves the size of the given Collection or Map. Performs an
     * unwrapping by calling <code>getAsObject</code>, then calls the
     * <code>size</code> method of the underlying object.</p>
     *
     * @param arguments a <code>List</code> containing a
     * <code>TemplateObjectMode</code> as its single item
     * @return the size of the Collection or Map as a <code>TemplateNumberModel</code>
     * @throws TemplateModelException the size could not be determined
     */
    public TemplateModel exec( List<TemplateModel> arguments ) throws TemplateModelException {
        if( arguments.size() != 1 ) {
            throw new TemplateModelException( "CollectionSize method must have exactly one argument" );
        }

        if(!( arguments.get(0) instanceof TemplateObjectModel )) {
            throw new TemplateModelException( "Argument must implement the TemplateObjectModel interface" );
        }
        TemplateObjectModel model = (TemplateObjectModel)arguments.get(0);
        Object object = model.getAsObject();

        if( object instanceof Collection ) {
            Collection collection = (Collection)object;
            return new FastNumber( collection.size() );
        } else if( object instanceof Map ) {
            Map map = (Map)object;
            return new FastNumber( map.size() );
        } else {
            throw new TemplateModelException( "Object must be a Collection or Map" );
        }
    }

    /**
     * For serialization purposes, always resolve a de-serialized object
     * back to the singleton instance. This ensures that any strict equality
     * tests remain valid after serialization.
     */
    private Object readResolve() throws ObjectStreamException {
        return instance;
    }
}
