/*
 * FreeMarker: a tool that allows Java programs to generate HTML
 * output using templates.
 * Copyright (C) 1998-2013 Benjamin Geer
 * Email: beroul@users.sourceforge.net
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA  02111-1307, USA.
 */

package freemarker.ext.misc;

import freemarker.template.TemplateModelException;
import freemarker.template.TemplateTransformModel;

import java.io.*;

/**
 * <p>A transform model that strips trailing whitespace from every line,
 * and reduces consecutive blank lines to a single blank line.
 * This transformation is useful as a global filter.</p>
 *
 * <p>Note: this is not a content-aware filter. Any content inside a
 * &lt;pre&gt; ... &lt;/pre&gt; block will also have its whitespace transformed.</p>
 *
 * <p>Usage:<br />
 * From java:</p>
 * <pre>
 * TemplateModelRoot root = new SimpleHash();
 *
 * root.put( "compressLines", freemarker.ext.misc.CompressLines.getInstance() );
 *
 * ...
 * </pre>
 *
 * <p>From your FM-Classic template:</p>
 * <pre>
 * The following is compressed:
 * &lt;transform compressLines&gt;
 *   &lt;p&gt;This page consists of multiple lines.&lt;/p&gt;
 *   &lt;p&gt;For every line on the page, all trailing whitespace is removed.&lt;/p&gt;
 *
 *
 *   &lt;p&gt;Any blank lines are reduced to a single blank line.&lt;/p&gt;
 * &lt;/transform&gt;
 *
 * ...
 * </pre>
 *
 * <p><b>Note:</b><br />
 * <code>CompressLines</code> is a singleton instance. Use the
 * {@link #getInstance()} method to retrieve instances of this model.</p>
 *
 * @see LegacyCompress
 * @see CompressWhitespace
 * @version $Id: CompressLines.java 1105 2013-04-09 10:52:59Z run2000 $
 */
public final class CompressLines implements TemplateTransformModel,
        Serializable {

    private static final CompressLines instance = new CompressLines();

    /** Serialization UUID for this class. */
    private static final long serialVersionUID = 1925033139308755755L;

    private CompressLines() {
    }

    /**
     * Retrieve a singleton instance of this class. Since there is no
     * state information held between calls, there is only a need for
     * one instance.
     *
     * @return an instance of this <code>CompressLines</code> class
     */
    public static CompressLines getInstance() {
        return instance;
    }

    /**
     * Is the object empty?
     *
     * @return <code>false</code>, to indicate non-empty transformation
     */
    public boolean isEmpty() throws TemplateModelException {
        return false;
    }

    /**
     * Compresses lines within the marked portion of a FM-Classic template.
     *
     * @param source the input to be transformed
     * @param output the destination of the transformation
     */
    public void transform(Reader source, PrintWriter output) throws IOException, TemplateModelException {
        BufferedReader br = new BufferedReader(source);
        boolean blankPending = false;

        String line = br.readLine();
        while(line != null) {
            int len = line.length();
            while(len > 0) {
                char c = line.charAt(len - 1);
                if(Character.isWhitespace(c)) {
                    len--;
                } else {
                    break;
                }
            }
            if(len == 0) {
                blankPending = true;
            } else {
                if(blankPending) {
                    output.println();
                }
                if(len < line.length()) {
                    output.println(line.substring(0, len));
                } else {
                    output.println(line);
                }
                blankPending = false;
            }
            line = br.readLine();
        }
    }

    /**
     * For serialization purposes, always resolve a de-serialized object
     * back to the singleton instance. This ensures that any strict equality
     * tests remain valid after serialization.
     */
    private Object readResolve() throws ObjectStreamException {
        return instance;
    }
}
