/*
 * FreeMarker: a tool that allows Java programs to generate HTML
 * output using templates.
 * Copyright (C) 1998-2004 Benjamin Geer
 * Email: beroul@users.sourceforge.net
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA  02111-1307, USA.
 */

package freemarker.ext.misc;

import freemarker.template.*;
import java.io.*;

/**
 * <p>A transform model that compresses each occurrence of consecutive
 * whitespace down to a single space character. Leading and trailing whitespace
 * is also removed.</p>
 *
 * <p>Usage:<br />
 * From java:</p>
 * <pre>
 * TemplateModelRoot root = new SimpleHash();
 *
 * root.put( "compressWhitespace", freemarker.ext.misc.CompressWhitespace.getInstance() );
 *
 * ...
 * </pre>
 *
 * <p>From your FM-Classic template:</p>
 * <pre>
 * The following is compressed:
 * &lt;transform compressWhitespace&gt;
 *   &lt;p&gt;This paragraph has all whitespace reduced to a single
 *   space character.&lt;/p&gt;
 * &lt;/transform&gt;
 *
 * ...
 * </pre>
 *
 * <p><b>Note:</b><br />
 * <code>CompressWhitespace</code> is a singleton instance. Use the
 * {@link #getInstance()} method to retrieve instances of this model.</p>
 *
 * @version $Id: CompressWhitespace.java 1105 2013-04-09 10:52:59Z run2000 $
 * @see LegacyCompress
 * @see CompressLines
 */
public final class CompressWhitespace implements TemplateTransformModel2,
        Serializable {

    private static final int READER_BUFFER_SIZE = 4096;
    private static final CompressWhitespace instance = new CompressWhitespace();

    /** Serialization UUID for this class. */
    private static final long serialVersionUID = 5379995644307790605L;

    /** Private constructor, to indicate this is a singleton class. */
    private CompressWhitespace () {
    }

    /**
     * Retrieve a singleton instance of this class. Since there is no
     * state information held between calls, there is only a need for
     * one instance.
     *
     * @return an instance of this <code>CompressWhitespace</code> class
     */
    public static CompressWhitespace getInstance() {
        return instance;
    }

    /**
     * Compresses whitespace within the marked portion of a FreeMarker template.
     *
     * @param source the input to be transformed
     * @param output the destination of the transformation
     */
    public void transform(Reader source, Writer output) throws IOException,
            TemplateModelException {

        // Output the string, stripping whitespace characters back to single space
        boolean whitespaceSeen = false;
        boolean atBeginning = true;

        char[]  aBuffer = new char[ READER_BUFFER_SIZE ];
        int buffSize = source.read( aBuffer );
        while(buffSize >= 0) {
            int currOffset = 0;
            int prevOffset = 0;
            while( currOffset < buffSize ) {
                char c = aBuffer[currOffset];

                if( Character.isWhitespace( c )) {
                    if( ! whitespaceSeen ) {
                        output.write( aBuffer, prevOffset, currOffset - prevOffset );
                        whitespaceSeen = true;
                    }
                } else {
                    if( atBeginning ) {
                        atBeginning = false;
                        whitespaceSeen = false;
                        prevOffset = currOffset;
                    } else if( whitespaceSeen ) {
                        output.write(' ');
                        whitespaceSeen = false;
                        prevOffset = currOffset;
                    }
                }
                currOffset++;
            }
            if( ! whitespaceSeen ) {
                output.write( aBuffer, prevOffset, currOffset - prevOffset );
            }
            buffSize = source.read( aBuffer );
        }
    }

    /**
     * Is the object empty?
     *
     * @return <code>false</code>, to indicate this object is not empty
     */
    public boolean isEmpty() throws TemplateModelException {
        return false;
    }

    /**
     * For serialization purposes, always resolve a de-serialized object
     * back to the singleton instance. This ensures that any strict equality
     * tests remain valid after serialization.
     */
    private Object readResolve() throws ObjectStreamException {
        return instance;
    }
}
