/*
 * FreeMarker: a tool that allows Java programs to generate HTML
 * output using templates.
 * Copyright (C) 1998-2004 Benjamin Geer
 * Email: beroul@users.sourceforge.net
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA  02111-1307, USA.
 */

package freemarker.ext.misc;
import freemarker.template.*;

import java.util.List;
import java.io.ObjectInputStream;
import java.io.IOException;
import java.io.Serializable;

/**
 * <p>Simple extension to the {@link freemarker.template.SimpleList} implementation.
 * This class implements a number of convenience facilities in the form of template
 * hash keys. The hash keys are:</p>
 * <ul>
 * <li><code>_size</code> - retrieves the size of the list</li>
 * <li><code>_first</code> - retrieves the first item in the list (always), or
 * <code>null</code> if the list is empty</li>
 * <li><code>_last</code> - retrieves the last item in the list (always), or
 * <code>null</code> if the list is empty</li>
 * </ul>
 *
 * <p>Note: the <code>_isLast</code> key has now been moved to the
 * {@link ExtendedIterator} class.</p>
 *
 * <p>An <code>ExtendedList</code> can also be given a shelf life; it
 * then keeps track of the time elapsed since the last time it copied
 * another {@link freemarker.template.TemplateListModel},
 * {@link freemarker.template.TemplateModel} array, or the last time its
 * shelf life was set.  Its owner can check whether it has become stale,
 * and refresh it if necessary by copying data into it again.</p>
 *
 * <p>When this model is serialized and deserialized, the newly deserialized
 * object has its <code>timeCreated</code> member variable reset to the
 * current time. In other words, the time until expiry will be reset to
 * the current shelf life following deserialization.</p>
 *
 * @version $Id: ExtendedList.java 1101 2013-04-01 04:17:32Z run2000 $
 * @since 1.7
 * @see ExtendedHash
 */
public class ExtendedList extends SimpleList implements TemplateHashModel, Serializable {

    protected static final int MS_PER_MINUTE = 60000;

    /** The creation time to be used for determining staleness. */
    protected transient long timeCreated;

    /**
     * @serial The shelf life in milliseconds of this <code>ExtendedList</code>.
     */
    protected long shelfLife;

    /**
     * @serial Whether this <code>ExtendedList</code> has an indefinite
     * shelf life.
     */
    protected boolean isPermanent = true;

    /** Class UUID for serialization. */
    private static final long serialVersionUID = 1503259903034496279L;

    /** Creates new ExtendedList. */
    public ExtendedList() {
    }

    /**
     * Creates new ExtendedList given a <code>List</code> of values.
     * A defensive copy of the list is made.
     *
     * @param list the list of values to be copied into this
     * <code>ExtendedList</code>
     * @throws NullPointerException the list value is null
     */
    public ExtendedList( List<? extends TemplateModel> list ) {
        super( list );
    }

    /**
     * Creates new ExtendedList populated by the <code>TemplateModel</code> array.
     *
     * @param arr the array to be copied into the underlying <code>List</code>.
     */
    public ExtendedList( TemplateModel[] arr ) {
        super( arr );
    }

    /**
     * Creates new ExtendedList populated by copying the given
     * <code>TemplateListModel</code>.
     *
     * @param listToCopy the list to be copied into this one.
     * @throws TemplateModelException something went wrong while copying the
     * given list into the <code>ExtendedList</code>.
     */
    public ExtendedList( TemplateListModel listToCopy ) throws TemplateModelException {
        super( listToCopy );
    }

    /**
     * Creates new ExtendedList populated by copying the given
     * <code>TemplateListModel2</code>.
     *
     * @param listToCopy the list to be copied into this one.
     * @throws TemplateModelException something went wrong while copying the
     * given list into the <code>ExtendedList</code>.
     */
    public ExtendedList( TemplateListModel2 listToCopy ) throws TemplateModelException {
        super( listToCopy );
    }

     /**
     * Gets one of the special keys that the <code>ExtendedList</code>
     * implementation provides.
     *
     * @param key the name of the special value to be retrieved.
     * @return the <code>TemplateModel</code> referred to by the key,
     * or null if not found.
     */
    public synchronized TemplateModel get( String key ) throws TemplateModelException {
        if( "_size".equals(key)) {
            return new FastNumber( list.size() );
        }
        if( "_first".equals(key)) {
            if( list.size() > 0 ) {
                return list.get( 0 );
            } else {
                return null;
            }
        }
        if( "_last".equals(key)) {
            if( list.size() > 0 ) {
                return list.get( list.size() - 1 );
            } else {
                return null;
            }
        }
        return null;
    }

    /**
     * Delegates to the underlying <code>SimpleList</code> copy method. This
     * variation sets the time that this list was created.
     */
    public synchronized void copy( TemplateListModel listToCopy )
            throws TemplateModelException {
        setTimeCreated();
        super.copy( listToCopy );
    }

    /**
     * Delegates to the underlying <code>SimpleList</code> copy method. This
     * variation sets the time that this list was created.
     *
     * @param listToCopy the list to be copied into this one.
     * @throws TemplateModelException something went wrong while copying the
     * given list into the <code>SimpleList</code>.
     */
    public synchronized void copy(TemplateListModel2 listToCopy)
            throws TemplateModelException {
        setTimeCreated();
        super.copy( listToCopy );
    }

    /**
     * Delegates to the underlying <code>SimpleList</code> copy method. This
     * variation sets the time that this list was created.
     */
    public synchronized void copy( TemplateModel[] arr ) {
        setTimeCreated();
        super.copy( arr );
    }

    /**
     * Sets the amount of time before this <code>ExtendedList</code>
     * will become stale.
     *
     * @param minutes the number of minutes before this
     * <code>ExtendedList</code> will become stale.
     */
    public synchronized void setShelfLife(int minutes) {
        isPermanent = false;
        shelfLife = minutes * MS_PER_MINUTE;
        setTimeCreated();
    }

    /**
     * @return this <code>ExtendedList</code>'s shelf life in minutes.
     */
    public synchronized int getShelfLife() {
        return (int)(shelfLife / MS_PER_MINUTE);
    }

    /**
     * Gives this <code>ExtendedList</code> an indefinite shelf life.
     */
    public synchronized void setPermanent() {
        isPermanent = true;
    }

    /**
     * @return <code>true</code> if this <code>ExtendedList</code> has an
     * indefinite shelf life.
     */
    public synchronized boolean isPermanent() {
        return isPermanent;
    }

    /**
     * @return <code>true</code> if this <code>ExtendedList</code> has become stale.
     */
    public synchronized boolean isStale() {
        return ( ! isPermanent &&
                 (System.currentTimeMillis() - timeCreated >= shelfLife));
    }

    /**
     * Retrieves an iterator to iterate over this list.
     *
     * @return an iterator to iterate over the current list.
     * @throws TemplateModelException the next item in the list can't be
     * retrieved, or no next item exists.
     */
    public TemplateIteratorModel templateIterator() throws TemplateModelException {
        return new ExtendedIterator( list );
    }

    /**
     * Returns the used iterator to the list model.
     *
     * @param iterator the iterator to be returned to the object pool, if any
     */
    public void releaseIterator(TemplateIteratorModel iterator) {
        // Do nothing
    }

    /**
     * Sets the time created to the current time.
     */
    private synchronized void setTimeCreated() {
        timeCreated = System.currentTimeMillis();
    }

    /**
     * Attempt to compare ExtendedList objects for equality.
     *
     * @param o the object to compare against
     * @return <code>true</code> if the objects are equal, otherwise
     * <code>false</code>
     */
    public boolean equals( Object o ) {
        if( this == o ) {
            return true;
        }
        if( !( o instanceof ExtendedList )) {
            return false;
        }
        if( !super.equals( o )) {
            return false;
        }

        final ExtendedList extendedList = ( ExtendedList ) o;

        if( isPermanent != extendedList.isPermanent ) {
            return false;
        }
        if( isPermanent ) {
            return true;
        }
        return( shelfLife == extendedList.shelfLife );
    }

    /**
     * Generate a hash code for this object based on the hash code for the
     * <code>SimpleList</code>.
     *
     * @return a hash code corresponding to the value of this object
     */
    public int hashCode() {
        int result = super.hashCode();
        if( isPermanent ) {
            result = 29 * result + 1;
        } else {
            result = 29 * result + ( int ) ( shelfLife ^ ( shelfLife >>> 32 ) );
        }
        return result;
    }

    /**
     * For serialization purposes, read this object in a specialized manner.
     * This resets the shelf life to the current time.
     *
     * @param stream the input stream to read serialized objects from
     */
    private void readObject(ObjectInputStream stream)
            throws IOException, ClassNotFoundException {
        stream.defaultReadObject();
        timeCreated = System.currentTimeMillis();
    }

}
