/*
 * FreeMarker: a tool that allows Java programs to generate HTML
 * output using templates.
 * Copyright (C) 1998-2004 Benjamin Geer
 * Email: beroul@users.sourceforge.net
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA  02111-1307, USA.
 */
package freemarker.ext.misc;

import java.util.*;
import java.io.Serializable;
import java.io.ObjectStreamException;

import freemarker.template.*;

/**
 * <p>Determines the values of the underlying <code>Map</code> object given a
 * <code>TemplateObjectModel</code>. Does this by calling the
 * <code>getAsObject</code> method. If the object is of the expected type,
 * the values are returned as a <code>FastList</code>. Otherwise, a
 * <code>TemplateModelException</code> is thrown.</p>
 *
 * <p>Usage:<br />
 * From java:</p>
 * <pre>
 * TemplateModelRoot root = new SimpleHash();
 *
 * root.put( "values", freemarker.ext.misc.HashValues.getInstance() );
 *
 * ...
 * </pre>
 *
 * <p>From your FM-Classic template:</p>
 * <pre>
 * &lt;assign hash1 = { "test1", "value1",
 *                      "test2", "value2",
 *                      "test3", "value3" }&gt;
 * &lt;assign valueList = values( hash1 )&gt;
 * &lt;p&gt;The values of hash1 are:&lt;/p&gt;
 * &lt;p&gt;&lt;foreach value in valueList&gt;
 * ${value}&lt;br&gt;
 * &lt;/foreach&gt;&lt;/p&gt;
 *
 * ...
 * </pre>
 *
 * <p><b>Note:</b><br />
 * <code>HashValues</code> is a singleton instance. Use the
 * {@link #getInstance()} method to retrieve instances of this model.</p>
 *
 * @author Nicholas Cull
 * @version $Id: HashValues.java 1101 2013-04-01 04:17:32Z run2000 $
 * @since 1.8
 */
public final class HashValues implements TemplateMethodModel2, Serializable {

    private static final HashValues instance = new HashValues();

    /** Class UUID for serialization. */
    private static final long serialVersionUID = -2622939369036499463L;

    /** Private constructor, to indicate this is a singleton class. */
    private HashValues() {
    }

    /**
     * Retrieve a singleton instance of this class. Since there is no
     * state information held between calls, there is only a need for
     * one instance.
     *
     * @return an instance of this <code>HashValues</code> class
     */
    public static HashValues getInstance() {
        return instance;
    }

    /**
     * Is the object empty?
     *
     * @return <code>false</code>, since the method is never empty
     */
    public boolean isEmpty() throws TemplateModelException {
        return false;
    }

    /**
     * Executes a method call. Arguments are passed as a <code>List</code> of
     * {@link TemplateModel} objects.
     *
     * @param arguments a <code>List</code> of <code>TemplateModel</code> objects
     * containing the values of the arguments passed to the method.
     * @return the <code>TemplateModel</code> produced by the method, or <code>null</code>.
     */
    public TemplateModel exec( List<TemplateModel> arguments ) throws TemplateModelException {
        if( arguments.size() != 1 ) {
            throw new TemplateModelException( "HashValues method must have exactly one argument" );
        }

        if(!( arguments.get(0) instanceof TemplateObjectModel )) {
            throw new TemplateModelException( "Argument must implement the TemplateObjectModel interface" );
        }
        TemplateObjectModel model = (TemplateObjectModel)arguments.get(0);
        Object object = model.getAsObject();

        if( !( object instanceof Map )) {
            throw new TemplateModelException( "Could not determine values of this object" );
        }

        Collection<TemplateModel> values = ((Map<?,TemplateModel>)object).values();
        List<TemplateModel> list = new ArrayList<TemplateModel>(values);
        return new FastList<TemplateModel>( list );
    }

    /**
     * For serialization purposes, always resolve a de-serialized object
     * back to the singleton instance. This ensures that any strict equality
     * tests remain valid after serialization.
     */
    private Object readResolve() throws ObjectStreamException {
        return instance;
    }
}
