/*
 * FreeMarker: a tool that allows Java programs to generate HTML
 * output using templates.
 * Copyright (C) 1998-2004 Benjamin Geer
 * Email: beroul@users.sourceforge.net
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA  02111-1307, USA.
 */

package freemarker.ext.misc;

import java.util.logging.Logger;
import java.util.logging.Level;
import freemarker.template.*;

/**
 * <p>Adapts FM-Classic's exception events to the <code>java.util.logging</code>
 * logging methods. The Java logging package is provided with Java 1.4 or later.</p>
 *
 * <p><b>Usage example:</b></p>
 * <pre>
 * // Configure the logger
 * Logger global = Logger.getLogger("");
 * Formatter simpleFormatter = new SimpleFormatter();
 * ConsoleHandler consoleHandler = new ConsoleHandler();
 * consoleHandler.setFormatter(simpleFormatter);
 * global.addHandler(consoleHandler);
 *
 * // Set up the template
 * Template template = new Template(filename);
 *
 * // Set up the model
 * TemplateModelRoot modelRoot = (...)
 *
 * // Set up the listener
 * SimpleEventAdapter adapter = new SimpleEventAdapter( new JavaLoggingExceptionListener() );
 *
 * // Call the template
 * template.process( modelRoot, writer, adapter );
 *
 * // Flush the handler
 * consoleHandler.flush();
 * </pre>
 *
 * @author  Nicholas Cull
 * @version $Id: JavaLoggingExceptionListener.java 1152 2013-04-15 10:34:50Z run2000 $
 * @since 1.8.5
 */
public final class JavaLoggingExceptionListener implements TemplateExceptionListener {

    /** Creates new JavaLoggingExceptionListener */
    public JavaLoggingExceptionListener() {
    }

    /**
     * Passes exceptions raised by FM-Classic onto the Java logging API.
     * The log name supplied to the logger match the class and method name
     * that caused the event to be fired.
     */
    public void exceptionThrown(TemplateExceptionEvent e) {
        String source = e.getSourceName();
        Logger logger = Logger.getLogger( source );
        StringBuilder message = new StringBuilder();
        Exception ex = e.getException();

        message.append( ex.getMessage() );
        if( ex instanceof TemplateException ) {
            Exception cause = ((TemplateException)ex).getCauseException();
            if( cause != null ) {
                message.append( ": " );
                message.append( cause.getMessage() );
            }
        }
        Level priority = null;

        switch( e.getSeverity() ) {
            case ERROR:
                priority = Level.SEVERE;
                break;

            case WARNING:
                priority = Level.WARNING;
                break;

            case DEPRECATION:
                priority = Level.INFO;
                break;
        }

        int index = source.lastIndexOf('.');
        String className, methodName;

        // Figure out the class and method names for the original exception
        if( index > 0 ) {
            className = source.substring( 0, index );
            methodName = source.substring( index + 1 );
        } else {
            className = source;
            methodName = "";
        }

        logger.logp( priority, className, methodName, message.toString() );
    }
}
