/*
 * FreeMarker: a tool that allows Java programs to generate HTML
 * output using templates.
 * Copyright (C) 1998-2004 Benjamin Geer
 * Email: beroul@users.sourceforge.net
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA  02111-1307, USA.
 */
package freemarker.ext.misc;

import freemarker.template.*;
import freemarker.template.expression.ExpressionUtils;

import java.io.Serializable;
import java.io.ObjectStreamException;
import java.util.List;
import java.util.ArrayList;

/**
 * Behaves the same way as MapList, but results are returned "live", that is,
 * evaluated on the fly. This is useful in situations where the underlying
 * list model may be unbounded, and a loop over such a list is terminated
 * by a break or return statement.
 * <p>
 * The method model supplied to the LiveMapList call <i>should</i>
 * be idempotent, otherwise unexpected results may appear when re-evaluating
 * list elements. For MapList, this is not such a problem, as long as the side
 * effects are predictable over the contents of the underlying list.
 *
 * @author Nicholas Cull
 * @version $Id: LiveMapList.java 1090 2013-03-29 11:04:07Z run2000 $
 * @since 1.9
 */
public final class LiveMapList implements TemplateMethodModel2, Serializable {
    private static final LiveMapList instance = new LiveMapList();

    private static final long serialVersionUID = 7347013161239690231L;

    /**
     * Private constructor, meaning this shouldn't be instantiated directly.
     * Use the {@link #getInstance} method to return an instance of this
     * class.
     */
    private LiveMapList() {
    }

    /**
     * Retrieve a singleton instance of this class. Since there is no
     * state information held between calls, there is only a need for
     * one instance.
     *
     * @return an instance of this <code>MapList</code> class
     */
    public static LiveMapList getInstance() {
        return instance;
    }

    /**
     * Is the object empty?
     *
     * @return <code>true</code> if this object is empty, otherwise <code>false</code>
     */
    public boolean isEmpty() throws TemplateModelException {
        return false;
    }

    /**
     * Executes a method call. Arguments are passed as a <code>List</code> of
     * {@link freemarker.template.TemplateModel} objects.
     *
     * @param arguments a <code>List</code> of <code>TemplateModel</code> objects
     *                  containing the values of the arguments passed to the method.
     * @return the <code>TemplateModel</code> produced by the method, or <code>null</code>.
     */
    public TemplateModel exec( List<TemplateModel> arguments ) throws TemplateModelException {
        if( arguments.size() != 2 ) {
            throw new TemplateModelException( "Map takes two arguments" );
        }

        TemplateModel list = (TemplateModel) arguments.get( 0 );
        TemplateModel method = (TemplateModel) arguments.get( 1 );

        if( list instanceof TemplateListModel ) {
            if(( method instanceof TemplateMethodModel ) || ( method instanceof TemplateMethodModel2 )) {
                return new LiveMapList1( (TemplateListModel) list, method );
            } else {
                throw new TemplateModelException( "Second argument to map must be a method" );
            }
        } else if( list instanceof TemplateListModel2 ) {
            if(( method instanceof TemplateMethodModel ) || ( method instanceof TemplateMethodModel2 )) {
                return new LiveMapList2( (TemplateListModel2)list, method );
            } else {
                throw new TemplateModelException( "Second argument to map must be a method" );
            }
        } else {
            throw new TemplateModelException( "First argument to map must be a list" );
        }
    }

    /**
     * For serialization purposes, always resolve a de-serialized object
     * back to the singleton instance. This ensures that any strict equality
     * tests remain valid after serialization.
     */
    private Object readResolve() throws ObjectStreamException {
        return instance;
    }

    private static class LiveMapList1 implements TemplateListModel {
        private final TemplateListModel listModel;
        private final TemplateModel mapMethod;

        public LiveMapList1( TemplateListModel listModel, TemplateModel mapMethod ) {
            this.listModel = listModel;
            this.mapMethod = mapMethod;
        }

        /**
         * Is the object empty?
         *
         * @return <code>true</code> if this object is empty, otherwise <code>false</code>
         */
        public boolean isEmpty() throws TemplateModelException {
            return listModel == null || listModel.isEmpty();
        }

        /**
         * Is the cursor at the beginning of the list?
         *
         * @return <code>true</code> if the cursor is at the beginning of the list,
         *         otherwise <code>false</code>.
         */
        public boolean isRewound() throws TemplateModelException {
            return listModel.isRewound();
        }

        /**
         * Resets the cursor to the beginning of the list.
         */
        public void rewind() throws TemplateModelException {
            listModel.rewind();
        }

        /**
         * Is there a next item in the list?
         *
         * @return <code>true</code> if there is a next element, otherwise
         *         <code>false</code>.
         */
        public boolean hasNext() throws TemplateModelException {
            return listModel.hasNext();
        }

        /**
         * Retrieves the next item in the list.
         *
         * @return the next element in the list.
         * @throws freemarker.template.TemplateModelException
         *          the next item in the list can't be
         *          retrieved, or no next item exists.
         */
        public TemplateModel next() throws TemplateModelException {
            TemplateModel nextModel = listModel.next();

            if( mapMethod instanceof TemplateMethodModel ) {
                ArrayList<String> args = new ArrayList<String>(1);
                args.add( ExpressionUtils.getAsStringOrEmpty( nextModel ));
                return ((TemplateMethodModel)mapMethod).exec( args );
            } else {
                ArrayList<TemplateModel> args = new ArrayList<TemplateModel>(1);
                args.add( nextModel );
                return ((TemplateMethodModel2)mapMethod).exec( args );
            }
        }

        /**
         * Retrieves the specified item from the list.
         *
         * @param index the index of the item to be retrieved.
         * @return the specified index in the list.
         * @throws freemarker.template.TemplateModelException
         *          the specified item in the list can't be
         *          retrieved, or the index is out of bounds.
         * @since 1.5.3
         */
        public TemplateModel get( int index ) throws TemplateModelException {
            TemplateModel indexedModel = listModel.get( index );

            if( mapMethod instanceof TemplateMethodModel ) {
                ArrayList<String> args = new ArrayList<String>(1);
                args.add( ExpressionUtils.getAsStringOrEmpty( indexedModel ));
                return ((TemplateMethodModel)mapMethod).exec( args );
            } else {
                ArrayList<TemplateModel> args = new ArrayList<TemplateModel>(1);
                args.add( indexedModel );
                return ((TemplateMethodModel2)mapMethod).exec( args );
            }
        }
    }

    private static class LiveMapList2 implements TemplateListModel2 {
        private final TemplateListModel2 listModel;
        private final TemplateModel mapModel;

        public LiveMapList2( TemplateListModel2 listModel, TemplateModel mapModel ) {
            this.listModel = listModel;
            this.mapModel = mapModel;
        }

        /**
         * Is the object empty?
         *
         * @return <code>true</code> if this object is empty, otherwise <code>false</code>
         */
        public boolean isEmpty() throws TemplateModelException {
            return listModel == null || listModel.isEmpty();
        }

        /**
         * Retrieves an iterator to iterate over this list. Can choose to return
         * <code>null</code> if there are no elements in the list.
         *
         * @return an iterator to iterate over the current list.
         * @throws freemarker.template.TemplateModelException
         *          the next item in the list can't be
         *          retrieved, or no next item exists.
         */
        public TemplateIteratorModel templateIterator() throws TemplateModelException {
            if( mapModel instanceof TemplateMethodModel ) {
                return new LiveMapIterator1( listModel.templateIterator(), (TemplateMethodModel) mapModel );
            } else {
                return new LiveMapIterator2( listModel.templateIterator(), (TemplateMethodModel2) mapModel );
            }
        }

        /**
         * <p>Returns the used iterator to the list model. Implement this method
         * when you want to use an object pool of <code>TemplateIterator</code> objects.
         * Otherwise, leave the implementation of this method blank.</p>
         *
         * <p>Note that if the iterator returned in {@link #templateIterator} is
         * <code>null</code>, this method will not be called for the
         * <code>null</code> iterator.</p>
         *
         * @param iterator the iterator to be returned to the object pool, if any
         */
        public void releaseIterator( TemplateIteratorModel iterator ) {
            listModel.releaseIterator( ((LiveMapIterator)iterator).getWrapperIterator() );
        }
    }

    interface LiveMapIterator extends TemplateIteratorModel {
        TemplateIteratorModel getWrapperIterator();
    }

    private static class LiveMapIterator1 implements LiveMapIterator {
        private final TemplateIteratorModel iterator;
        private final TemplateMethodModel mapMethod;

        public LiveMapIterator1( TemplateIteratorModel iterator, TemplateMethodModel mapMethod ) {
            this.iterator = iterator;
            this.mapMethod = mapMethod;
        }

        /**
         * Is the object empty?
         *
         * @return <code>true</code> if this object is empty, otherwise <code>false</code>
         */
        public boolean isEmpty() throws TemplateModelException {
            return iterator == null || iterator.isEmpty();
        }

        /**
         * Do we have another item in the list?
         *
         * @return <code>true</code> if there are more items to be iterated over,
         *         otherwise <code>false</code>
         * @throws freemarker.template.TemplateModelException
         *          there was a problem determining the
         *          next item in the list
         */
        public boolean hasNext() throws TemplateModelException {
            return iterator.hasNext();
        }

        /**
         * Retrieve the next item in the list. The item will be a <code>TemplateModel</code>
         * containing the underlying value.
         *
         * @return the next item in the list
         * @throws freemarker.template.TemplateModelException
         *          the next item couldn't be retrieved,
         *          or we're at the end of the list
         */
        public TemplateModel next() throws TemplateModelException {
            TemplateModel result = iterator.next();
            ArrayList<String> args = new ArrayList<String>(1);
            args.add( ExpressionUtils.getAsStringOrEmpty( result ));
            return mapMethod.exec( args );
        }

        public TemplateIteratorModel getWrapperIterator() {
            return iterator;
        }
    }

    private static class LiveMapIterator2 implements LiveMapIterator {
        private final TemplateIteratorModel iterator;
        private final TemplateMethodModel2 mapMethod;

        public LiveMapIterator2( TemplateIteratorModel iterator, TemplateMethodModel2 mapMethod ) {
            this.iterator = iterator;
            this.mapMethod = mapMethod;
        }

        /**
         * Is the object empty?
         *
         * @return <code>true</code> if this object is empty, otherwise <code>false</code>
         */
        public boolean isEmpty() throws TemplateModelException {
            return iterator == null || iterator.isEmpty();
        }

        /**
         * Do we have another item in the list?
         *
         * @return <code>true</code> if there are more items to be iterated over,
         *         otherwise <code>false</code>
         * @throws freemarker.template.TemplateModelException
         *          there was a problem determining the
         *          next item in the list
         */
        public boolean hasNext() throws TemplateModelException {
            return iterator.hasNext();
        }

        /**
         * Retrieve the next item in the list. The item will be a <code>TemplateModel</code>
         * containing the underlying value.
         *
         * @return the next item in the list
         * @throws freemarker.template.TemplateModelException
         *          the next item couldn't be retrieved,
         *          or we're at the end of the list
         */
        public TemplateModel next() throws TemplateModelException {
            ArrayList<TemplateModel> args = new ArrayList<TemplateModel>(1);
            args.add( iterator.next() );
            return mapMethod.exec( args );
        }

        public TemplateIteratorModel getWrapperIterator() {
            return iterator;
        }
    }
}
