/*
 * FreeMarker: a tool that allows Java programs to generate HTML
 * output using templates.
 * Copyright (C) 1998-2004 Benjamin Geer
 * Email: beroul@users.sourceforge.net
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA  02111-1307, USA.
 */
package freemarker.ext.misc;

import freemarker.template.*;
import freemarker.template.expression.ExpressionUtils;

import java.util.List;
import java.util.ArrayList;
import java.io.Serializable;
import java.io.ObjectStreamException;

/**
 * @author Nicholas Cull
 * @version $Id: MapList.java 1090 2013-03-29 11:04:07Z run2000 $
 * @since 1.9
 */
public final class MapList implements TemplateMethodModel2, Serializable {
    private static final MapList instance = new MapList();

    private static final long serialVersionUID = 7347013161239690231L;

    /**
     * Private constructor, meaning this shouldn't be instantiated directly.
     * Use the {@link #getInstance} method to return an instance of this
     * class.
     */
    private MapList() {
    }

    /**
     * Retrieve a singleton instance of this class. Since there is no
     * state information held between calls, there is only a need for
     * one instance.
     *
     * @return an instance of this <code>MapList</code> class
     */
    public static MapList getInstance() {
        return instance;
    }

    /**
     * Is the object empty?
     *
     * @return <code>true</code> if this object is empty, otherwise <code>false</code>
     */
    public boolean isEmpty() throws TemplateModelException {
        return false;
    }

    /**
     * Executes a method call. Arguments are passed as a <code>List</code> of
     * {@link freemarker.template.TemplateModel} objects.
     *
     * @param arguments a <code>List</code> of <code>TemplateModel</code> objects
     *                  containing the values of the arguments passed to the method.
     * @return the <code>TemplateModel</code> produced by the method, or <code>null</code>.
     */
    public TemplateModel exec( List<TemplateModel> arguments ) throws TemplateModelException {
        if( arguments.size() != 2 ) {
            throw new TemplateModelException( "Map takes two arguments" );
        }

        TemplateModel list = arguments.get( 0 );
        TemplateModel method = arguments.get( 1 );

        if( list instanceof TemplateListModel ) {
            if( method instanceof TemplateMethodModel ) {
                return generateListMethod( (TemplateListModel)list, (TemplateMethodModel)method );
            } else if( method instanceof TemplateMethodModel2 ) {
                return generateListMethod( (TemplateListModel)list, (TemplateMethodModel2)method );
            } else {
                throw new TemplateModelException( "Second argument to map must be a method" );
            }
        } else if( list instanceof TemplateListModel2 ) {
            if( method instanceof TemplateMethodModel ) {
                return generateListMethod( (TemplateListModel2)list, (TemplateMethodModel)method );
            } else if( method instanceof TemplateMethodModel2 ) {
                return generateListMethod( (TemplateListModel2)list, (TemplateMethodModel2)method );
            } else {
                throw new TemplateModelException( "Second argument to map must be a method" );
            }
        } else {
            throw new TemplateModelException( "First argument to map must be a list" );
        }
    }

    private static TemplateListModel2 generateListMethod( TemplateListModel list, TemplateMethodModel method )
            throws TemplateModelException {
        ArrayList<TemplateModel> resultList = new ArrayList<TemplateModel>();

        if( ! list.isEmpty() ) {
            ArrayList<String> args = new ArrayList<String>( 1 );

            if( ! list.isRewound() ) {
                list.rewind();
            }

            try {
                while( list.hasNext() ) {
                    TemplateModel item = list.next();
                    args.add( ExpressionUtils.getAsStringOrEmpty( item ));
                    resultList.add( method.exec( args ));
                    args.clear();
                }
            } finally {
                list.rewind();
            }
        }

        return new FastList<TemplateModel>( resultList );
    }

    private static TemplateListModel2 generateListMethod( TemplateListModel list, TemplateMethodModel2 method )
            throws TemplateModelException {
        ArrayList<TemplateModel> resultList = new ArrayList<TemplateModel>();

        if( ! list.isEmpty() ) {
            ArrayList<TemplateModel> args = new ArrayList<TemplateModel>( 1 );

            if( ! list.isRewound() ) {
                list.rewind();
            }

            try {
                while( list.hasNext() ) {
                    TemplateModel item = list.next();
                    args.add( item );
                    resultList.add( method.exec( args ));
                    args.clear();
                }
            } finally {
                list.rewind();
            }
        }

        return new FastList<TemplateModel>( resultList );
    }

    private static TemplateListModel2 generateListMethod( TemplateListModel2 list, TemplateMethodModel method )
            throws TemplateModelException {
        ArrayList<TemplateModel> resultList = new ArrayList<TemplateModel>();

        if( ! list.isEmpty() ) {
            ArrayList<String> args = new ArrayList<String>( 1 );
            TemplateIteratorModel iterator = list.templateIterator();

            try {
                while( iterator.hasNext() ) {
                    TemplateModel item = iterator.next();
                    args.add( ExpressionUtils.getAsStringOrEmpty( item ));
                    resultList.add( method.exec( args ));
                    args.clear();
                }
            } finally {
                list.releaseIterator( iterator );
            }
        }

        return new FastList<TemplateModel>( resultList );
    }

    private static TemplateListModel2 generateListMethod( TemplateListModel2 list, TemplateMethodModel2 method )
            throws TemplateModelException {
        ArrayList<TemplateModel> resultList = new ArrayList<TemplateModel>();

        if( ! list.isEmpty() ) {
            ArrayList<TemplateModel> args = new ArrayList<TemplateModel>( 1 );
            TemplateIteratorModel iterator = list.templateIterator();

            try {
                while( iterator.hasNext() ) {
                    TemplateModel item = iterator.next();
                    args.add( item );
                    resultList.add( method.exec( args ));
                    args.clear();
                }
            } finally {
                list.releaseIterator( iterator );
            }
        }

        return new FastList<TemplateModel>( resultList );
    }

    /**
     * For serialization purposes, always resolve a de-serialized object
     * back to the singleton instance. This ensures that any strict equality
     * tests remain valid after serialization.
     */
    private Object readResolve() throws ObjectStreamException {
        return instance;
    }
}
