/*
 * FreeMarker: a tool that allows Java programs to generate HTML
 * output using templates.
 * Copyright (C) 1998-2004 Benjamin Geer
 * Email: beroul@users.sourceforge.net
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA  02111-1307, USA.
 */
package freemarker.ext.misc;

import java.util.List;
import java.util.Arrays;
import java.io.*;

import freemarker.template.TemplateMethodModel;
import freemarker.template.TemplateModelException;
import freemarker.template.TemplateModel;
import freemarker.template.TemplateTransformModel;
import org.apache.oro.text.PatternCache;
import org.apache.oro.text.PatternCacheLRU;
import org.apache.oro.text.perl.Perl5Util;
import org.apache.oro.text.perl.MalformedPerl5PatternException;

/**
 * <p>Perform Perl 5 substitutions using the Jakarta ORO regular expression
 * library. A list of regular expressions is passed in, and executed
 * in order on a line-by-line basis. In this respect, the Perl 5 modifiers
 * "m" and "s" make no difference here. The result of this transformation
 * is always terminated by a newline sequence.</p>
 *
 * <p><b>Usage:</b><br />
 * From java:</p>
 * <pre>
 * TemplateModelRoot root = new SimpleHash();
 *
 * root.put( "substitute", freemarker.ext.misc.Perl5Substitute.getInstance() );
 *
 * ...
 * </pre>
 *
 * <p>From your FM-Classic template:</p>
 * <pre>
 * The following is stripped of HTML tags:
 * &lt;transform substitute( "s/&lt;[^&gt;]*&gt;//g" )&gt;
 *   &lt;p&gt;This paragraph has &lt;em&gt;all&lt;/em&gt; tags removed.&lt;/p&gt;
 * &lt;/transform&gt;
 *
 * ...
 * </pre>
 *
 * <p><b>Note:</b><br />
 * <code>Perl5Substitute</code> is a singleton instance. Use the
 * {@link #getInstance()} method to retrieve instances of this model.</p>
 *
 * @version $Id: Perl5Substitute.java 1152 2013-04-15 10:34:50Z run2000 $
 * @author Nicholas Cull
 * @since 1.8
 */
public final class Perl5Substitute implements TemplateMethodModel,
        Serializable {

    private static final Perl5Substitute instance = new Perl5Substitute();

    /** Class UUID for serialization. */
    private static final long serialVersionUID = -2721279165395212285L;

    /**
     * Private constructor indicating the Perl5Substitute is a singleton.
     * Note however, that the transformation class itself is instantiated
     * for each new call to the {@link #exec} method.
     */
    private Perl5Substitute() {
    }

    /**
     * Return a new instance of this class. No state is held in the method
     * model itself. Rather, the method model creates a new instance
     * of the transformation class itself whenever it is called. The
     * transformation class holds all state information about the regular
     * expressions.
     *
     * @return a new instance of the Perl5Substitute method model
     */
    public static Perl5Substitute getInstance() {
        return instance;
    }

    /**
     * Is the object empty?
     *
     * @return <code>true</code> if this object is empty, otherwise <code>false</code>
     */
    public boolean isEmpty() throws TemplateModelException {
        return false;
    }

    /**
     * Creates a TemplateTransformModel with the specified arguments as
     * regular expressions. Arguments are passed as a <code>List</code> of
     * <code>String</code> objects. Each regular expression is evaluated
     * in order, once per line.
     *
     * @param arguments a <code>List</code> of <code>String</code> objects
     * containing the regular expressions to be applied
     * @return the <code>TemplateTransformModel</code> produced by the method
     */
    public TemplateModel exec( List<String> arguments ) throws TemplateModelException {
        return new OroTransform( arguments );
    }

    /**
     * For serialization purposes, always resolve a de-serialized object
     * back to the singleton instance. This ensures that any strict equality
     * tests remain valid after serialization.
     */
    private Object readResolve() throws ObjectStreamException {
        return instance;
    }

    /**
     * Implementation of the transformation itself. We do some pre-conversions
     * and cache initialization up-front to make sure that runtime performance
     * of the transform itself is as fast as we can make it.
     */
    private static class OroTransform implements TemplateTransformModel,
            Serializable {

        private transient PatternCache patternCache;
        private transient Perl5Util util;
        private final String []substitutions;

        /** Class UUID for serialization. */
        private static final long serialVersionUID = -1935085075483027358L;

        /**
         * Create a new OroTransform model with the given regular expression
         * patterns.
         *
         * @param patterns the Perl regular expression patterns
         */
        public OroTransform( List<String> patterns ) {
            patternCache = new PatternCacheLRU( patterns.size() );
            util = new Perl5Util( patternCache );
            substitutions = new String[ patterns.size() ];
            patterns.toArray( substitutions );
        }

        /**
         * Is the object empty?
         *
         * @return <code>true</code> if this object is empty, otherwise <code>false</code>
         */
        public boolean isEmpty() throws TemplateModelException {
            return false;
        }

        /**
         * Performs a transformation/filter on FM-Classic output.
         *
         * @param source the input to be transformed
         * @param output the destination of the transformation
         */
        public void transform( Reader source, PrintWriter output ) throws IOException,
                TemplateModelException {

            BufferedReader reader = new BufferedReader(source);
            String line;

            try {
                while((line = reader.readLine()) != null) {
                    for(String pattern : substitutions) {
                        line = util.substitute(pattern, line);
                    }
                    output.println(line);
                }
            } catch( MalformedPerl5PatternException e ) {
                throw new TemplateModelException( "Perl 5 regular expression could not be compiled: "
                        + e.getMessage(), e );
            }
        }

        /**
         * Returns a string representation of the object.
         *
         * @return a <code>String</code> representation of the object.
         */
        public String toString() {
            StringBuilder buffer = new StringBuilder();

            buffer.append( "Perl5Substitute transform: " );
            buffer.append( Arrays.asList( substitutions ));
            return buffer.toString();
        }

        /**
         * For serialization, read this object normally, then reconstruct
         * the Perl 5 pattern cache utility class.
         *
         * @param stream the input stream to read serialized objects from
         */
        private void readObject(ObjectInputStream stream) throws IOException,
                ClassNotFoundException {

            stream.defaultReadObject();
            patternCache = new PatternCacheLRU( substitutions.length );
            util = new Perl5Util( patternCache );
        }
    }
}
