/*
 * FreeMarker: a tool that allows Java programs to generate HTML
 * output using templates.
 * Copyright (C) 1998-2005 Benjamin Geer
 * Email: beroul@users.sourceforge.net
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA  02111-1307, USA.
 */
package freemarker.ext.misc;

import java.io.*;
import java.util.List;
import java.util.Arrays;

import freemarker.template.TemplateMethodModel;
import freemarker.template.TemplateModelException;
import freemarker.template.TemplateModel;
import freemarker.template.TemplateTransformModel;
import org.jtr.transliterate.Perl5Parser;
import org.jtr.transliterate.CharacterReplacer;
import org.jtr.transliterate.CharacterParseException;

/**
 * <p>Perform Perl 5 translation using the JTR library. A list of translate
 * expressions is passed in, and executed in order on a line-by-line basis.
 * The result of this transformation is always terminated by a newline
 * sequence.</p>
 *
 * <p><b>Usage:</b><br />
 * From java:</p>
 * <pre>
 * TemplateModelRoot root = new SimpleHash();
 *
 * root.put( "translate", freemarker.ext.misc.Perl5Translate.getInstance() );
 *
 * ...
 * </pre>
 *
 * <p>From your FM-Classic template:</p>
 * <pre>
 * The following removes all duplicate spaces:
 * &lt;transform translate( "tr/ //s" )&gt;
 *    This paragraph has  all  excess  spaces  removed.
 * &lt;/transform&gt;
 *
 * ...
 * </pre>
 *
 * <p><b>Note:</b><br />
 * <code>Perl5Translate</code> is a singleton instance. Use the
 * {@link #getInstance()} method to retrieve instances of this model.</p>
 *
 * @author Nicholas Cull
 * @version $Id: Perl5Translate.java 1153 2013-04-15 10:59:37Z run2000 $
 * @since 1.9
 */
public final class Perl5Translate implements TemplateMethodModel, Serializable {
    private static final Perl5Translate instance = new Perl5Translate();

    /** Class UUID for serialization. */
    private static final long serialVersionUID = 7255864681955590135L;

    /**
     * Private constructor indicating the Perl5Translate is a singleton.
     * Note however, that the transformation class itself is instantiated
     * for each new call to the {@link #exec} method.
     */
    private Perl5Translate() {
    }

    /**
     * Return a new instance of this class. No state is held in the method
     * model itself. Rather, the method model creates a new instance
     * of the transformation class itself whenever it is called. The
     * transformation class holds all state information about the translations.
     *
     * @return a new instance of the Perl5Substitute method model
     */
    public static Perl5Translate getInstance() {
        return instance;
    }

    public boolean isEmpty() throws TemplateModelException {
        return false;
    }

    public TemplateModel exec( List<String> arguments ) throws TemplateModelException {
        try {
            return new TranslateTransform( arguments );
        } catch( CharacterParseException e ) {
            throw new TemplateModelException( "Cannot compile translation: " +
                    e.getMessage(), e );
        }
    }

    /**
     * For serialization purposes, always resolve a de-serialized object
     * back to the singleton instance. This ensures that any strict equality
     * tests remain valid after serialization.
     */
    private Object readResolve() throws ObjectStreamException {
        return instance;
    }

    /**
     * Implementation of the transformation itself. We store the translations
     * as a String to optimize the serialized form, as well as a transient
     * compiled form for speedy processing.
     */
    private static class TranslateTransform implements TemplateTransformModel, Serializable {
        private transient CharacterReplacer[] replacers;
        private final String[] translations;

        /** Class UUID for serialization. */
        private static final long serialVersionUID = -6370566533077695240L;

        public TranslateTransform( List<String> translations ) throws CharacterParseException {
            this.translations = new String[translations.size()];
            translations.toArray( this.translations );
            compileTranslations();
        }

        private void compileTranslations() throws CharacterParseException {
            replacers = new CharacterReplacer[ translations.length ];

            for( int i = 0; i < translations.length; i++ ) {
                replacers[ i ] = Perl5Parser.makeReplacer( translations[ i ]);
            }
        }

        public boolean isEmpty() throws TemplateModelException {
            return (translations == null) || (translations.length == 0);
        }

        public void transform( Reader source, PrintWriter output )
                throws IOException, TemplateModelException {

            BufferedReader reader = new BufferedReader(source);
            String line;

            while((line = reader.readLine()) != null) {
                for(int i = 0; i < translations.length; i++) {
                    line = replacers[ i ].doReplacement(line);
                }
                output.println(line);
            }
        }

        /**
         * Returns a string representation of the object.
         *
         * @return a <code>String</code> representation of the object.
         */
        public String toString() {
            StringBuilder buffer = new StringBuilder();

            buffer.append( "Perl5Translate transform: " );
            buffer.append( Arrays.asList( translations ));
            return buffer.toString();
        }

        /**
         * For serialization, read this object normally, then reconstruct
         * the Perl 5 pattern cache utility class.
         *
         * @param stream the input stream to read serialized objects from
         */
        private void readObject(ObjectInputStream stream) throws IOException,
                ClassNotFoundException {

            stream.defaultReadObject();
            try {
                compileTranslations();
            } catch( CharacterParseException e ) {
                throw new InvalidObjectException(
                        "Cannot compile translations from serialized form: " +
                        e.getMessage() );
            }
        }
    }
}
