/*
 * FreeMarker: a tool that allows Java programs to generate HTML
 * output using templates.
 * Copyright (C) 1998-2004 Benjamin Geer
 * Email: beroul@users.sourceforge.net
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA  02111-1307, USA.
 */
package freemarker.ext.misc;

import freemarker.template.*;
import org.apache.oro.text.regex.MatchResult;

/**
 * A simple model that corresponds with the Jakarta ORO <code>MatchResult</code>
 * interface. This is returned by instances of the {@link Match} class.
 *
 * @author Nicholas Cull
 * @version $Id: PerlMatcher.java 1101 2013-04-01 04:17:32Z run2000 $
 * @since 1.8.2
 */
final class PerlMatcher implements TemplateScalarModel, TemplateIndexedModel,
        TemplateHashModel, TemplateObjectModel {
    private MatchResult result;

    /**
     * Construct a PerlMatcher that takes a <code>MatchResult</code> object
     * to be wrapped.
     *
     * @param result an ORO <code>MatchResult</code> from which we retrieve
     * results
     */
    public PerlMatcher( MatchResult result ) {
        this.result = result;
    }

    /**
     * Is the object empty?
     *
     * @return <code>true</code> if this object is empty, otherwise <code>false</code>
     */
    public boolean isEmpty() {
        return result == null;
    }

    /**
     * Returns the matched value as a <code>String</code>.
     *
     * @return the <code>String</code> value of this scalar.
     */
    public String getAsString() {
        return result.toString();
    }

    /**
     * Get the matched value corresponding to the given index. An index of 0
     * returns the complete match. Other indexes correspond with subgroup
     * matches.
     *
     * @param index the index of the underlying value we're interested in
     * @return a <code>TemplateModel</code> representing the matched value
     * for the given index
     */
    public TemplateModel getAtIndex( long index ) {
        return new FastScalar( result.group( (int)index ));
    }

    /**
     * Gets a {@link TemplateModel} from the hash. The only key of interest
     * here is the "groups" key, which returns the number of groups as a
     * number.
     *
     * @param key the name by which the <code>TemplateModel</code>
     * is identified in the template.
     * @return the <code>TemplateModel</code> referred to by the key,
     * or <code>null</code> if not found.
     */
    public TemplateModel get( String key ) throws TemplateModelException {
        if( key == null ) {
            return null;
        }
        if( "groups".equals(key)) {
            return new FastNumber( result.groups() );
        }
        return null;
    }

    /**
     * Return the underlying object to the reflection mechanism in the
     * <code>freemarker.ext.beans</code> package. Any variables, methods
     * or properties can be called directly via reflection.</p>
     *
     * @return the underlying object for this template model
     * @throws TemplateModelException the object could not be returned
     */
    public Object getAsObject() throws TemplateModelException {
        return result;
    }
}
