/*
 * FreeMarker: a tool that allows Java programs to generate HTML
 * output using templates.
 * Copyright (C) 1998-2005 Benjamin Geer
 * Email: beroul@users.sourceforge.net
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA  02111-1307, USA.
 */
package freemarker.ext.misc;

import java.util.List;
import java.io.*;

import freemarker.template.TemplateMethodModel2;
import freemarker.template.TemplateModelException;
import freemarker.template.TemplateModel;
import freemarker.template.TemplateTransformModel;
import freemarker.template.TemplateTransformModel2;

/**
 * <p>A class for chaining multiple transformations into a single model without
 * having to nest &lt;transform&gt; tags. A slightly more elegant way to
 * handle transformation pipelines.</p>
 *
 * <p>Transformations are called from leftmost argument to rightmost in order.
 * This differs from nesting &lt;transform&gt; tags where the innermost is
 * called first, and the outermost is called last.</p>
 *
 * <p>Usage:<br />
 * From java:</p>
 * <pre>
 * TemplateModelRoot root = new SimpleHash();
 *
 * root.put( "pipeline", freemarker.ext.misc.Pipeline.getInstance() );
 *
 * ...
 * </pre>
 *
 * <p>From your FM-Classic template:</p>
 * <pre>
 * The following is transformed with a pipeline:
 * &lt;transform pipeline( transform1, transform2, transform3, transform4 )&gt;
 *   This output is transformed by all the transformations in the pipeline.
 * &lt;/transform&gt;
 *
 * ...
 * </pre>
 *
 * <p>In the example, the transformations are applied in the following order:</p>
 * <p> Source =&gt; Transform 1 =&gt; Transform 2 =&gt; Transform 3 =&gt;
 * Transform 4 = &gt; Output</p>
 *
 * <p><b>Note:</b><br />
 * <code>Pipeline</code> is a singleton instance. Use the
 * {@link #getInstance()} method to retrieve instances of this model.</p>
 *
 * @author Nicholas Cull
 * @version $Id: Pipeline.java 1088 2013-03-29 05:52:17Z run2000 $
 * @since 1.9
 */
public final class Pipeline implements TemplateMethodModel2, Serializable {

    private static final Pipeline instance = new Pipeline();

    /** Class UUID for serialization. */
    private static final long serialVersionUID = -1492993253658875861L;

    /**
     * Private constructor indicating the Pipeline is a singleton.
     * Note however, that the transformation pipeline class itself is
     * instantiated for each new call to the {@link #exec} method.
     */
    private Pipeline() { }

    /**
     * Return a new instance of this class. No state is held in the method
     * model itself. Rather, the method model creates a new instance
     * of the transformation class itself whenever it is called. The
     * transformation class holds all state information about the pipeline.
     *
     * @return a new instance of the Pipeline method model
     */
    public static Pipeline getInstance() {
        return instance;
    }

    /**
     * Is the current method model empty.
     *
     * @return <code>false</code>, indicating that this method model is not
     * empty
     */
    public boolean isEmpty() {
        return false;
    }

    /**
     * Construct a new transform pipeline with the given transform models.
     * The models will be called in leftmost to rightmost order.
     *
     * @param arguments a List of TemplateTransformModel or
     * TemplateTransformModel2 models to be chained into a single pipeline
     * @return a new TemplateTransformModel2 that represents all the
     * transform models in a single pipeline
     * @throws TemplateModelException
     */
    public TemplateModel exec( List<TemplateModel> arguments ) throws TemplateModelException {
        // Need to determine up-front if all list arguments are
        // TemplateTransformModel or TemplateTransformModel2 implementations
        int size = arguments.size();
        TemplateModel model;

        for( int i = 0; i < size; i++ ) {
            model = (TemplateModel) arguments.get( i );
            if((model != null) && !( model instanceof TemplateTransformModel ) && !( model instanceof TemplateTransformModel2 )) {
                throw new TemplateModelException( "Argument " + i +
                        " is not a transform model, it is a " + model.getClass().getName() );
            }
        }

        return new TransformPipeline( arguments );
    }

    /**
     * For serialization purposes, always resolve a de-serialized object
     * back to the singleton instance. This ensures that any strict equality
     * tests remain valid after serialization.
     */
    private Object readResolve() throws ObjectStreamException {
        return instance;
    }
}
