/*
 * FreeMarker: a tool that allows Java programs to generate HTML
 * output using templates.
 * Copyright (C) 1998-2004 Benjamin Geer
 * Email: beroul@users.sourceforge.net
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA  02111-1307, USA.
 */
package freemarker.ext.misc;

import freemarker.template.*;
import freemarker.template.expression.ExpressionUtils;

import java.util.List;
import java.util.ArrayList;
import java.io.*;

/**
 * @author Nicholas Cull
 * @version $Id: ReduceTransform.java 1090 2013-03-29 11:04:07Z run2000 $
 * @since 1.9
 */
public final class ReduceTransform implements TemplateMethodModel2, Serializable {
    private static final ReduceTransform instance = new ReduceTransform();

    private static final long serialVersionUID = -3274602358874454517L;

    /**
     * Private constructor, meaning this shouldn't be instantiated directly.
     * Use the {@link #getInstance} method to return an instance of this
     * class.
     */
    private ReduceTransform() {
    }

    /**
     * Retrieve a singleton instance of this class. Since there is no
     * state information held between calls, there is only a need for
     * one instance.
     *
     * @return an instance of this <code>ReduceTransform</code> class
     */
    public static ReduceTransform getInstance() {
        return instance;
    }

    /**
     * Is the object empty?
     *
     * @return <code>true</code> if this object is empty, otherwise <code>false</code>
     */
    public boolean isEmpty() throws TemplateModelException {
        return false;
    }

    /**
     * Executes a method call. Arguments are passed as a <code>List</code> of
     * {@link freemarker.template.TemplateModel} objects.
     *
     * @param arguments a <code>List</code> of <code>TemplateModel</code> objects
     *                  containing the values of the arguments passed to the method.
     * @return the <code>TemplateModel</code> produced by the method, or <code>null</code>.
     */
    public TemplateModel exec( List<TemplateModel> arguments ) throws TemplateModelException {
        int argSize = arguments.size();
        if( argSize != 1 ) {
            throw new TemplateModelException( "Reduce takes one argument" );
        }

        TemplateModel model = arguments.get( 0 );

        if( model instanceof TemplateMethodModel ) {
            return new ReduceTransform1( (TemplateMethodModel) model );
        } else if( model instanceof TemplateMethodModel2 ) {
            return new ReduceTransform2( (TemplateMethodModel2) model );
        } else {
            throw new TemplateModelException( "ReduceTransform takes a method model" );
        }
    }

    /**
     * For serialization purposes, always resolve a de-serialized object
     * back to the singleton instance. This ensures that any strict equality
     * tests remain valid after serialization.
     */
    private Object readResolve() throws ObjectStreamException {
        return instance;
    }

    private static class ReduceTransform1 implements TemplateTransformModel, Serializable {
        private final TemplateMethodModel method;

        private static final long serialVersionUID = 3424787698372478652L;

        public ReduceTransform1( TemplateMethodModel method ) {
            this.method = method;
        }

        /**
         * Is the object empty?
         *
         * @return <code>true</code> if this object is empty, otherwise <code>false</code>
         */
        public boolean isEmpty() throws TemplateModelException {
            return method == null;
        }

        /**
         * Performs a transformation/filter on FM-Classic output.
         *
         * @param source the input to be transformed
         * @param output the destination of the transformation
         */
        public void transform( Reader source, PrintWriter output ) throws IOException, TemplateModelException {
            BufferedReader br = new BufferedReader( source );
            ArrayList<String> args = new ArrayList<String>(2);
            String line = br.readLine();
            String result;

            if( line != null ) {
                result = line;
                while((line = br.readLine()) != null) {
                    args.add( result );
                    args.add( line );
                    TemplateModel outModel = method.exec( args );
                    result = ExpressionUtils.getAsStringOrEmpty( outModel );
                    args.clear();
                }
                output.println( result );
            }
        }
    }

    private static class ReduceTransform2 implements TemplateTransformModel, Serializable {
        private final TemplateMethodModel2 method;

        private static final long serialVersionUID = 5524744614372471232L;

        public ReduceTransform2( TemplateMethodModel2 method ) {
            this.method = method;
        }

        /**
         * Is the object empty?
         *
         * @return <code>true</code> if this object is empty, otherwise <code>false</code>
         */
        public boolean isEmpty() throws TemplateModelException {
            return method == null;
        }

        /**
         * Performs a transformation/filter on FM-Classic output.
         *
         * @param source the input to be transformed
         * @param output the destination of the transformation
         */
        public void transform( Reader source, PrintWriter output ) throws IOException, TemplateModelException {
            BufferedReader br = new BufferedReader( source );
            ArrayList<TemplateModel> args = new ArrayList<TemplateModel>(2);
            String line = br.readLine();
            TemplateModel result;

            if( line != null ) {
                result = new FastScalar( line );
                while((line = br.readLine()) != null) {
                    args.add( result );
                    args.add( new FastScalar( line ));
                    result = method.exec( args );
                    args.clear();
                }
                output.println( ExpressionUtils.getAsStringOrEmpty( result ));
            }
        }
    }
}
