/*
 * FreeMarker: a tool that allows Java programs to generate HTML
 * output using templates.
 * Copyright (C) 1998-2004 Benjamin Geer
 * Email: beroul@users.sourceforge.net
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA  02111-1307, USA.
 */
package freemarker.ext.misc;

import java.io.*;

import freemarker.template.TemplateTransformModel2;
import freemarker.template.TemplateModelException;

/**
 * <p>A transform model that removes all blank lines. In addition, trailing
 * whitespace is also deleted. Thus, for a line containing only whitespace,
 * the entire line will be removed. Finally, the last line will not have
 * a linefeed at the end of it</p>
 *
 * <p>Newlines are detected by the <code>BufferedReader.readLine()</code> method,
 * and are rewritten using the platform default (using the
 * <code>System.getProperty()</code> method).</p>
 *
 * <p>Usage:<br />
 * From java:</p>
 * <pre>
 * TemplateModelRoot root = new SimpleHash();
 *
 * root.put( "removeBlankLines", freemarker.ext.misc.RemoveBlankLines.getInstance() );
 *
 * ...
 * </pre>
 *
 * <p>From your FM-Classic template:</p>
 * <pre>
 * The following is has empty lines removed:
 * &lt;transform removeBlankLines&gt;
 *   &lt;ul&gt;
 *   &lt;foreach item in listItems&gt;
 *     &lt;li&gt;${item}&lt;/li&gt;
 *   &lt;foreach&gt;
 *   &lt;/ul&gt;
 * &lt;/transform&gt;
 *
 * ...
 * </pre>
 *
 * <p><b>Note:</b><br />
 * <code>RemoveBlankLines</code> is a singleton instance. Use the
 * {@link #getInstance()} method to retrieve instances of this model.</p>
 *
 * @author Nicholas Cull
 * @version $Id: RemoveBlankLines.java 1101 2013-04-01 04:17:32Z run2000 $
 * @since 1.8.1
 */
public final class RemoveBlankLines implements TemplateTransformModel2, Serializable {
    private static final RemoveBlankLines instance = new RemoveBlankLines();

    /** Class UUID for serialization. */
    private static final long serialVersionUID = 2549618176944246810L;

    /**
     * Private constructor, meaning this shouldn't be instantiated directly.
     * Use the {@link #getInstance} method to return an instance of this
     * class.
     */
    private RemoveBlankLines() {
    }

    /**
     * Retrieve a singleton instance of this class. Since there is no
     * state information held between calls, there is only a need for
     * one instance.
     *
     * @return an instance of this <code>RemoveBlankLines</code> class
     */
    public static RemoveBlankLines getInstance() {
        return instance;
    }

    /**
     * Is the object empty?
     *
     * @return <code>false</code> to indicate that this model is non-empty
     */
    public boolean isEmpty() throws TemplateModelException {
        return false;
    }

    /**
     * Strips all blank lines from the source. In addition, removes any trailing
     * spaces from each line. As a consequence, if any line consists only of
     * whitespace, then all whitespace will be removed, and the line will be
     * stripped.
     *
     * @param source the input to be transformed
     * @param output the destination of the transformation
     */
    public void transform( Reader source, Writer output ) throws IOException,
            TemplateModelException {
        BufferedReader reader = new BufferedReader( source );
        String line = reader.readLine();
        String newline = System.getProperty( "line.separator" );

        while( line != null ) {
            // Strip trailing spaces
            int end = line.length();
            while(( end > 0 ) && Character.isWhitespace( line.charAt( end - 1 ))) {
                end--;
            }
            if( end > 0 ) {
                output.write( line, 0, end );
            }
            line = reader.readLine();
            if(( end > 0 ) && ( line != null )) {
                output.write( newline );
            }
        }
    }

    /**
     * For serialization purposes, always resolve a de-serialized object
     * back to the singleton instance. This ensures that any strict equality
     * tests remain valid after serialization.
     */
    private Object readResolve() throws ObjectStreamException {
        return instance;
    }
}
