/*
 * FreeMarker: a tool that allows Java programs to generate HTML
 * output using templates.
 * Copyright (C) 1998-2004 Benjamin Geer
 * Email: beroul@users.sourceforge.net
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA  02111-1307, USA.
 */
package freemarker.ext.misc;

import java.util.List;
import java.util.ArrayList;
import java.io.Serializable;
import java.io.ObjectInputStream;
import java.io.IOException;

import freemarker.template.*;
import freemarker.template.expression.ExpressionUtils;
import org.apache.oro.text.PatternCacheLRU;
import org.apache.oro.text.perl.Perl5Util;
import org.apache.oro.text.perl.MalformedPerl5PatternException;

/**
 * A method model that implements the Perl 5 split function. This uses the
 * Jakarta ORO library to perform the split.
 *
 * <p>Usage:<br />
 * From java:</p>
 * <pre>
 * TemplateModelRoot root = new SimpleHash();
 *
 * root.put( "split", new freemarker.ext.misc.Split() );
 *
 * ...
 * </pre>
 *
 * <p>From your FM-Classic template:</p>
 * <pre>
 * The following lists each directory in the path:
 * &lt;assign path = "/var/log/apache/"&gt;
 * &lt;assign output = split( "/\\//", path )&gt;
 * &lt;foreach item in output&gt;
 * %{item}
 * &lt;/foreach&gt;
 *
 * ...
 * </pre>
 *
 * @author Nicholas Cull
 * @version $Id: Split.java 1101 2013-04-01 04:17:32Z run2000 $
 * @since 1.8.2
 */
public final class Split implements TemplateMethodModel2, Serializable {

    private transient PatternCacheLRU patternCache;

    /** Class UUID for serialization. */
    private static final long serialVersionUID = -743213181532021558L;

    /**
     * Create a new split method.
     */
    public Split() {
        patternCache = new PatternCacheLRU();
    }

    /**
     * Is the object empty?
     *
     * @return <code>false</code> to indicate the method is not empty
     */
    public boolean isEmpty() {
        return false;
    }

    /**
     * Executes a method call. Arguments are passed as a <code>List</code> of
     * {@link TemplateModel} objects.
     *
     * @param arguments a <code>List</code> of <code>TemplateModel</code> objects
     * containing the values of the arguments passed to the method.
     * @return the <code>TemplateModel</code> produced by the method, or <code>null</code>.
     */
    public TemplateModel exec( List<TemplateModel> arguments ) throws TemplateModelException {
        int size = arguments.size();

        if( size < 1 || size > 3 ) {
            throw new TemplateModelException( "Split needs between one and three arguments" );
        }

        Perl5Util perl5Util = new Perl5Util(patternCache);
        List<String> results = new ArrayList<String>();
        String input, pattern;
        int limit;

        try {
            switch( size ) {
                case 1:
                    input = ExpressionUtils.getAsStringOrEmpty((TemplateModel)arguments.get( 0 ));
                    try {
                        perl5Util.split( results, input );
                    } catch( MalformedPerl5PatternException e ) {
                        throw new TemplateModelException(
                                "Perl 5 regular expression could not be compiled: "
                                + e.getMessage(), e );
                    }
                    break;

                case 2:
                    pattern = ExpressionUtils.getAsStringOrEmpty((TemplateModel)arguments.get( 0 ));
                    input = ExpressionUtils.getAsStringOrEmpty((TemplateModel)arguments.get( 1 ));
                    try {
                        perl5Util.split( results, pattern, input );
                    } catch( MalformedPerl5PatternException e ) {
                        throw new TemplateModelException(
                                "Perl 5 regular expression could not be compiled: "
                                + e.getMessage(), e );
                    }
                    break;

                case 3:
                    pattern = ExpressionUtils.getAsStringOrEmpty((TemplateModel)arguments.get( 0 ));
                    input = ExpressionUtils.getAsStringOrEmpty((TemplateModel)arguments.get( 1 ));
                    limit = (int)ExpressionUtils.getAsNumber((TemplateModel)arguments.get( 2 ));
                    try {
                        perl5Util.split( results, pattern, input, limit );
                    } catch( MalformedPerl5PatternException e ) {
                        throw new TemplateModelException(
                                "Perl 5 regular expression could not be compiled: "
                                + e.getMessage(), e );
                    }
                    break;
            }
        } catch( TemplateException e ) {
            throw new TemplateModelException( e.getMessage(), e );
        } catch( MalformedPerl5PatternException e ) {
            throw new TemplateModelException( "Malformed split expression: " +
                    e.getMessage(), e );
        } catch( NullPointerException e ) {
            // This occurs if getAsString returns null

        }

        // Now that we have the results, return them as an array of TemplateModels
        List<TemplateModel> templateList = new ArrayList<TemplateModel>( results.size() );

        for( String item : results ) {
            templateList.add( new FastScalar( item ));
        }
        return new FastList<TemplateModel>( templateList );
    }

    /**
     * For serialization, read this object normally, then reconstruct
     * the Perl 5 pattern cache utility class.
     *
     * @param stream the input stream to read serialized objects from
     */
    private void readObject(ObjectInputStream stream) throws IOException,
            ClassNotFoundException {
        stream.defaultReadObject();
        patternCache = new PatternCacheLRU();
    }
}
