/*
 * FreeMarker: a tool that allows Java programs to generate HTML
 * output using templates.
 * Copyright (C) 1998-2013 Benjamin Geer
 * Email: beroul@users.sourceforge.net
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA  02111-1307, USA.
 */

package freemarker.ext.misc;

import freemarker.template.TemplateModel;
import freemarker.template.TemplateModelException;
import freemarker.template.TemplateTransformModel;
import freemarker.template.TemplateTransformModel2;

import java.io.*;
import java.util.Arrays;
import java.util.List;

/**
 * A pipeline of transform models that is itself a transform model.
 * Transformations occur in leftmost to rightmost order. Null or
 * empty models are considered equivalent to the null transformation,
 * where data is simply copied from input to output without alteration.
 * Similarly, a pipeline consisting of no transformations is also handled
 * as a null transformation.
 *
 * @author Nicholas Cull
 * @version $Id$
 * @since 1.9
 */
public final class TransformPipeline implements TemplateTransformModel2, Serializable {
    /**
     * @serial the list of transform models to be pipelined. In practise,
     * serialization depends on all the contained models also being
     * serializable.
     */
    private final TemplateModel[] transforms;

    /** Class UUID for serialization. */
    private static final long serialVersionUID = -3179369201409975041L;

    /**
     * Create a new transform pipeline with the given transform models.
     *
     * @param transforms an ordered list of TemplateTransformModel and
     * TemplateTransformModel2 objects to be pipelined
     */
    public TransformPipeline(List<TemplateModel> transforms) {
        this.transforms = new TemplateModel[ transforms.size() ];
        transforms.toArray( this.transforms );
    }

    /**
     * Is there anything in this pipeline.
     *
     * @return <code>true</code> if the pipeline is empty, otherwise
     * <code>false</code>
     * @throws freemarker.template.TemplateModelException
     */
    public boolean isEmpty() throws TemplateModelException {
        return ( transforms == null ) || ( transforms.length == 0 );
    }

    /**
     * Perform the requested transformations on the given data source in
     * the given order.
     *
     * @param source the input data to be transformed
     * @param target the data after being fed through all transformations
     * @throws java.io.IOException
     * @throws freemarker.template.TemplateModelException
     */
    public void transform( Reader source, Writer target )
            throws IOException, TemplateModelException {

        // This shouldn't be necessary, but deal with empty transformations
        // just in case.
        if(( transforms == null ) || ( transforms.length == 0 )) {
            pipelineStage( source, target, null);
            return;
        }

        // Two special cases to be considered:
        // * the beginning, where the input is the source reader
        // * the end, where the target is the output writer
        String intermediate = "";

        for( int i = 0; i < transforms.length; i++ ) {
            Reader input;

            if( i == 0 ) {
                input = source;
            } else {
                input = new StringReader( intermediate );
            }
            if( i + 1 == transforms.length ) {
                pipelineStage( input, target, transforms[ i ]);
            } else {
                StringWriter output = new StringWriter();
                pipelineStage( input, output, transforms[ i ]);
                intermediate = output.toString();
            }
        }
    }

    /**
     * Perform a single pipeline stage from the given source to the given
     * target using the supplied transformation. If we're given a
     * <code>null</code> or empty transform model, perform a null
     * transformation on the source. Otherwise, determine the type of model,
     * and call with the appropriate arguments.
     *
     * @param source the source data to be transformed
     * @param target the target where transformed data will be written
     * @param transform the transformation to be applied, possibly
     * <code>null</code> or empty
     * @throws java.io.IOException an input/output error occured during the
     * transformation
     * @throws freemarker.template.TemplateModelException the transformation threw an exception
     */
    private void pipelineStage( Reader source, Writer target,
            TemplateModel transform ) throws IOException, TemplateModelException {

        // Deal with null transformations
        if(( transform == null ) || ( transform.isEmpty() )) {
            int bufferLen = 1024;
            char[] buffer = new char[ bufferLen ];
            int size = 0;
            size = source.read( buffer );
            while( size > 0 ) {
                target.write( buffer, 0, size );
            }
        }

        if( transform instanceof TemplateTransformModel) {
            TemplateTransformModel pipeline = (TemplateTransformModel) transform;
            pipeline.transform( source, new PrintWriter( target ));
        } else if( transform instanceof TemplateTransformModel2 ) {
            TemplateTransformModel2 pipeline = (TemplateTransformModel2) transform;
            pipeline.transform( source, target );
        }
    }

    /**
     * Returns a string representation of the object.
     *
     * @return a <code>String</code> representation of the object.
     */
    public String toString() {
        StringBuilder buffer = new StringBuilder();

        buffer.append( "Pipelined transforms: " );
        buffer.append( Arrays.asList(transforms));
        return buffer.toString();
    }
}
