/*
 * FreeMarker: a tool that allows Java programs to generate HTML
 * output using templates.
 * Copyright (C) 1998-2005 Benjamin Geer
 * Email: beroul@users.sourceforge.net
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA  02111-1307, USA.
 */
package freemarker.ext.misc;

import java.util.List;
import java.io.Serializable;
import java.io.ObjectInputStream;
import java.io.IOException;

import freemarker.template.*;
import freemarker.ext.util.CacheMapLRU;

import org.jtr.transliterate.Perl5Parser;
import org.jtr.transliterate.CharacterReplacer;
import org.jtr.transliterate.CharacterParseException;

/**
 * <p>A method model that implements the Perl 5 translate function. This uses
 * the JTR library to perform the translation.</p>
 *
 * <p><b>Usage:</b><br />
 * From java:</p>
 * <pre>
 * TemplateModelRoot root = new SimpleHash();
 *
 * root.put( "translate", new freemarker.ext.misc.Translate() );
 *
 * ...
 * </pre>
 *
 * <p>From your FM-Classic template:</p>
 * <pre>
 * The following is stripped of all duplicate spaces:
 * &lt;assign text = "   This paragraph has  all  excess  spaces  removed."&gt;
 * &lt;assign output = translate( "tr/ //s", text )&gt;
 *
 * ${output}
 *
 * The result is ${output.result}, the number of matches was ${output.matches}.
 * ...
 * </pre>
 *
 * @author Nicholas Cull
 * @version $Id: Translate.java 1084 2013-03-27 10:12:06Z run2000 $
 * @since 1.9
 */
public final class Translate implements TemplateMethodModel, Serializable {
    private transient CacheMapLRU<String, CharacterReplacer> translateCache;

    /** Class UUID for serialization. */
    private static final long serialVersionUID = 1929541287462943061L;

    public Translate() {
        translateCache = new CacheMapLRU<String, CharacterReplacer>();
    }

    public boolean isEmpty() throws TemplateModelException {
        return false;
    }

    public TemplateModel exec( List<String> arguments ) throws TemplateModelException {

        if( arguments.size() != 2 ) {
            throw new TemplateModelException( "Translate needs two arguments" );
        }

        CharacterReplacer replacer;
        String pattern = arguments.get( 0 );
        synchronized( this ) {
            replacer = translateCache.get( pattern );
            if( replacer == null ) {
                try {
                    replacer = Perl5Parser.makeReplacer( pattern );
                    translateCache.put( pattern, replacer );
                } catch( CharacterParseException e ) {
                    throw new TemplateModelException( "Could not parse translation: " + e.getMessage(), e );
                }
            }
        }

        String result = replacer.doReplacement( arguments.get( 1 ));
        int matches = replacer.getMatches();
        return new TranslateResult( result, matches );
    }

    /**
     * For serialization, read this object normally, then reconstruct
     * the translate cache.
     *
     * @param stream the input stream to read serialized objects from
     */
    private void readObject( ObjectInputStream stream) throws IOException,
            ClassNotFoundException {
        stream.defaultReadObject();
        translateCache = new CacheMapLRU<String, CharacterReplacer>();
    }

    /**
     * Contains the result of the translation. The result can be returned
     * directly as a scalar model, or accessed using the "result" key. The
     * number of matches can be accessed using the "matches" key.
     */
    private static class TranslateResult implements TemplateScalarModel, TemplateHashModel {
        private String result;
        private int matches;

        public TranslateResult( String result, int matches ) {
            this.result = result;
            this.matches = matches;
        }

        public boolean isEmpty() throws TemplateModelException {
            return result == null;
        }

        public String getAsString() throws TemplateModelException {
            return result;
        }

        public TemplateModel get( String key ) throws TemplateModelException {
            if( "result".equals( key )) {
                return new FastScalar( result );
            } else if( "matches".equals( key )) {
                return new FastNumber( matches );
            }
            return null;
        }
    }
}
