/*
 * FreeMarker: a tool that allows Java programs to generate HTML
 * output using templates.
 * Copyright (C) 1998-2004 Benjamin Geer
 * Email: beroul@users.sourceforge.net
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA  02111-1307, USA.
 */

package freemarker.ext.misc;

import java.io.Writer;
import java.io.IOException;

import freemarker.template.*;

/**
 * <p>Adapts FM-Classic's exception events to an IO Writer class. This would
 * typically be either a File writer, or adapted to another output such
 * as <code>System.err</code>. By default the writer is flushed after each
 * message that is written.</p>
 *
 * <p><b>Usage example:</b></p>
 * <pre>
 * // Create a writer for error messages
 * Writer errorWriter = new OutputStreamWriter( System.err );
 *
 * // Set up the template
 * Template template = new Template(filename);
 *
 * // Set up the model
 * TemplateModelRoot modelRoot = (...)
 *
 * // Set up the listener
 * SimpleEventAdapter adapter = new SimpleEventAdapter( new WriterExceptionListener( errorWriter ));
 *
 * // Call the template
 * template.process( modelRoot, writer, adapter );
 * </pre>
 *
 * @author Nicholas Cull
 * @since 1.8.2
 * @version $Id: WriterExceptionListener.java 1101 2013-04-01 04:17:32Z run2000 $
 */
public final class WriterExceptionListener implements TemplateExceptionListener {
    /** The writer to send exception messages */
    private Writer writer;
    /** Whether to autoflush after every log event, defaults to true. */
    private boolean autoFlush;

    /**
     * Constructor that takes a Writer as the target for error messages.
     * Autoflush is set to <code>true</code> by default.
     *
     * @param writer the Writer to which error messages will be sent
     */
    public WriterExceptionListener(Writer writer) {
        this( writer, true );
    }

    /**
     * Constructor that takes a Writer as the target for error messages, and
     * a flag to indicate whether the <code>flush()</code> method should be
     * called after every event.
     *
     * @param writer the Writer to which error messages will be sent
     * @param autoFlush should the Writer be flushed after every message
     */
    public WriterExceptionListener( Writer writer, boolean autoFlush ) {
        this.writer = writer;
        this.autoFlush = autoFlush;
    }

    /**
     * This method is called whenever a {@link TemplateExceptionEvent}
     * is generated by a FM-Classic template. Implement this method to
     * decide how to respond to such events.
     */
    public void exceptionThrown( TemplateExceptionEvent e )
            throws TemplateException {
        String newline = System.getProperty( "line.separator" );

        try {
            switch( e.getSeverity() ) {
                case ERROR:
                    writer.write( "ERROR " );
                    break;

                case WARNING:
                    writer.write( "WARNING ");
                    break;

                case DEPRECATION:
                    writer.write( "DEPRECATION ");
                    break;
            }

            Exception ex = e.getException();

            writer.write( e.getSourceName() );
            writer.write( " -  " );
            writer.write( ex.getMessage() );

            if( ex instanceof TemplateException ) {
                Exception cause = ((TemplateException)ex).getCauseException();
                if( cause != null ) {
                    writer.write( ": " );
                    writer.write( cause.getMessage() );
                }
            }

            writer.write( newline );
            if( autoFlush ) {
                writer.flush();
            }
        } catch( IOException e1 ) {
            // do nothing
        }
    }
}
