/*
 * FreeMarker: a tool that allows Java programs to generate HTML
 * output using templates.
 * Copyright (C) 1998-2005 Benjamin Geer
 * Email: beroul@users.sourceforge.net
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA  02111-1307, USA.
 */
package freemarker.ext.util;

/**
 * A simple cache of fixed size that implements a least recently used algorithm.
 * The access is unsyncrhonized. The default capacity of the cache is of ten
 * items. Attempts to store more than ten unique keys will result in the least
 * recently used item being evicted, and the new item inserted in its place.
 * Keys are assumed to always be non-<code>null</code> values.
 *
 * @author Nicholas Cull
 * @version $Id: CacheMapLRU.java 1090 2013-03-29 11:04:07Z run2000 $
 * @since 1.9
 */
public class CacheMapLRU<K,V> {
    private MapEntry[] entries;
    private int size;
    public static final int DEFAULT_CAPACITY = 10;

    public CacheMapLRU() {
        entries = new MapEntry[ DEFAULT_CAPACITY ];
        size = 0;
    }

    public CacheMapLRU( int capacity ) {
        if( capacity <= 0 ) {
            throw new IllegalArgumentException( "Capacity should be a positive value" );
        }

        entries = new MapEntry[ capacity ];
        size = 0;
    }

    public V get( K key ) {
        for( int i = 0; i < size; i++ ) {
            if( key.equals( entries[ i ].key )) {
                entries[ i ].timestamp = System.currentTimeMillis();
                return (V) entries[ i ].value;
            }
        }
        return null;
    }

    public void put( K key, V value ) {
        if( size < entries.length ) {
            entries[ size++ ] = new MapEntry( key, value );
        } else {
            // Implement LRU algorithm: look for the least recently updated
            // entry, and replace
            int index = 0;
            long timestamp = entries[ index ].timestamp;
            for( int i = 1; i < entries.length; i++ ) {
                if( entries[ index ].timestamp < timestamp ) {
                    index = i;
                }
            }
            entries[ index ] = new MapEntry( key, value );
        }
    }

    public void clear() {
        entries = new MapEntry[ entries.length ];
        size = 0;
    }

    private static final class MapEntry {
        private long timestamp;
        private Object key;
        private Object value;

        public MapEntry(Object key, Object value) {
            this.timestamp = System.currentTimeMillis();
            this.key = key;
            this.value = value;
        }

        /**
         * Returns a string representation of the object.
         *
         * @return a string representation of the object
         */
        public String toString() {
            return "MapEntry { key = '" + key + "', value = " + value +
                    ", timestamp = " + new java.util.Date( timestamp ).toString() +
                    " }";
        }
    }
}
