/*
 * FreeMarker: a tool that allows Java programs to generate HTML
 * output using templates.
 * Copyright (C) 1998-2005 Benjamin Geer
 * Email: beroul@users.sourceforge.net
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA  02111-1307, USA.
 */

package freemarker.template;

import java.io.*;
import freemarker.template.cache.Cache;
import freemarker.template.cache.Cacheable;

/**
 * <p>A base class from which {@link Template} implementations are subclassed.
 * This class implements all the interfaces required, and provides some of the
 * basic machinery required to compile and cache a <code>Template</code>.</p>
 *
 * @see Template
 * @see UnparsedTemplate
 * @version $Id: AbstractTemplate.java 1162 2013-04-20 12:04:32Z run2000 $
 */

public abstract class AbstractTemplate implements TemplateProcessor, Cacheable,
        Compileable, Cloneable {

    /** The <code>Cache</code> to which this template belongs (if any). */
    protected Cache cache;

    // For deserialization purposes.
    protected AbstractTemplate() {
    }

    /**
     * Constructs a template by compiling it from a <code>Reader</code>.
     *
     * @param reader a <code>Reader</code> from which the
     * template can be read.
     */
    public AbstractTemplate(Reader reader) throws IOException,
            IllegalArgumentException {
        compile( reader );
    }

    /**
     * Constructs a template by compiling it from a <code>Reader</code>.
     *
     * @param reader a <code>Reader</code> from which the
     * template can be read.
     */
    public AbstractTemplate(Reader reader, Cache cache) throws IOException,
            IllegalArgumentException {
        compile( reader );
        this.cache = cache;
    }

    /**
     * Compiles the template from a <code>Reader</code>. If the template
     * has already been compiled, this method does nothing.
     *
     * @param reader an <code>Reader</code> from which the
     * template can be read.
     */
    protected abstract void compile(Reader reader) throws IOException,
            IllegalArgumentException;

    /**
     * Retrieve the {@link freemarker.template.cache.Cache} that this object
     * is stored in.
     *
     * @return the <code>Cache</code> that this template belongs to.
     */
    public Cache getCache() {
        return this.cache;
    }

    /**
     * <p>Clones the current template.</p>
     *
     * <p>Cloning is used in {@link freemarker.template.cache.Cache}s,
     * whenever we need to create a new template: rather than simply creating
     * a new Template, we ask a {@link freemarker.template.cache.TemplateRegistry}
     * to create one for us. <code>TemplateRegistry</code> uses the clone
     * function to take an existing template, copy it, and return the copy
     * to the cache, where it is then populated.</p>
     *
     * @return a copy of the current template.
     */
    public Object clone() {
        try {
            return super.clone();
        } catch( CloneNotSupportedException e ) {
            throw new RuntimeException( e.getMessage() );
        }
    }

    /**
     * Processes the template, using data from the template model, writing
     * any events to the <code>TemplateEventAdapter</code>, and outputs
     * the resulting text to a <code>Writer</code>.
     *
     * @param p template processing parameters
     */
    public abstract ExitStatus process(TemplateProcessorParameters p) throws IOException;

    /**
     * Processes the template, using data from a template model, and outputs
     * the resulting text to a <code>Writer</code>.
     *
     * @param modelRoot the root node of the data model.  If <code>null</code>,
     * an empty data model is used.
     * @param out a <code>Writer</code> to output the text to.
     */
    public abstract void process(TemplateWriteableHashModel modelRoot, Writer out)
            throws IOException;

    /**
     * Processes the template, using an empty data model, and outputs
     * the resulting text to a <code>Writer</code>.
     *
     * @param out a <code>Writer</code> to output the text to.
     */
    public abstract void process(Writer out) throws IOException;

    /**
     * Takes the given Reader, reads it until the end of the stream, and
     * accumulates the contents in a String. The string is returned when
     * the stream is exhausted.
     *
     * @param reader the reader to be turned into a String
     * @return a String representation of the given stream
     * @throws IOException something went wrong while processing the stream
     */
    protected static String getTemplateText( Reader reader )
            throws IOException {

        if( reader == null ) {
            throw new IllegalArgumentException( "Reader is null" );
        }

        StringBuilder textBuf = new StringBuilder();
        BufferedReader br = new BufferedReader( reader );
        char cbuf[] = new char[ 1024 ];
        int nSize;

        try {
            nSize = br.read( cbuf );
            while( nSize > 0 ) {
                textBuf.append( cbuf, 0, nSize );
                nSize = br.read( cbuf );
            }
        } finally {
            try {
                br.close();
            } catch (IOException e) { }
        }

        return textBuf.toString();
    }
}
