/*
 * FreeMarker: a tool that allows Java programs to generate HTML
 * output using templates.
 * Copyright (C) 1998-2004 Benjamin Geer
 * Email: beroul@users.sourceforge.net
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA  02111-1307, USA.
 */

package freemarker.template;

import java.io.*;
import freemarker.template.cache.Cache;
import freemarker.template.cache.Cacheable;

/**
 * <p>An application or servlet can instantiate a <code>BinaryData</code>
 * to retrieve a binary file.
 *
 * <p>You can pass the filename of the binary file to a <code>BinaryDataFactory</code>
 * object, in which case it is read in immediately.
 *
 * <p>To retrieve the binary data, call the {@link #process} method.
 *
 * <p>To facilitate multithreading, <code>BinaryData</code> objects are
 * immutable; if you need to reload a binary file, you must make a new
 * <code>BinaryData</code> object.  In most cases, it will be sufficient
 * to let a {@link freemarker.template.cache.Cache} do this for you.
 *
 * @see FileTemplateCache
 * @see BinaryDataFactory
 * @version $Id: BinaryData.java 1166 2013-04-21 08:21:15Z run2000 $
 */
public class BinaryData implements Cacheable, Cloneable, Compileable,
        Serializable {

    /**
     * The binary data held by this object.
     *
     * @serial The binary data stored as a byte array
     */
    protected byte[] dataArray;
    /** The cache to which this binary data object belongs (if any). */
    protected transient Cache cache;

    /** Serialization UUID for this class. */
    private static final long serialVersionUID = -17848906542414739L;

    /**
     * Clones an existing <code>BinaryData</code> instance.
     *
     * @param stream an <code>InputStream</code> from which the
     * template can be read.
     */
    public BinaryData(InputStream stream) throws IOException {
        compile(stream);
    }

    /**
     * Clones an existing <code>BinaryData</code> instance.
     *
     * @param stream an <code>InputStream</code> from which the
     * template can be read.
     */
    public BinaryData(InputStream stream, Cache cache) throws IOException {
        compile(stream);
        this.cache = cache;
    }

    /**
     * Compiles the template from an <code>InputStream</code>. If the template
     * has already been compiled, this method does nothing.
     *
     * @param stream an <code>InputStream</code> from which the
     *               template can be read.
     */
    public void compile(InputStream stream)
            throws IOException, IllegalArgumentException {

        ByteArrayOutputStream arrayStream = new ByteArrayOutputStream(8192);
        byte[] buffer = new byte[4096];
        int bytes_read;

        if( stream == null ) {
            throw new IllegalArgumentException( "Cannot compile binary data from supplied InputStream" );
        }

        while ((bytes_read = stream.read(buffer)) >= 0) {
            arrayStream.write( buffer, 0, bytes_read );
        }
        arrayStream.close();
        stream.close();

        dataArray = arrayStream.toByteArray();
    }

    /**
     * Processes the binary data file, and output
     * the resulting binary data to an <tt>OutputStream</tt>.
     *
     * @param out an <tt>OutputStream</tt> to output the binary data to.
     */
    public void process(OutputStream out) {
        try {
            if (dataArray != null) {
                out.write( dataArray );
            }
        }
        catch (IOException e) { }
    }

    /**
     * Returns the length of the binary data, in bytes. This is convenient
     * for Servlets that want to return this information to the caller.
     *
     * @return the content length of the binary data, in bytes
     */
    public int getContentLength() {
        return dataArray.length;
    }

    /**
     * Retrieve the {@link freemarker.template.cache.Cache} that this object is stored in.
     *
     * @return the <code>Cache</code> that this template belongs to.
     */
    public Cache getCache() {
        return this.cache;
    }

    /**
     * Clones the current <code>BinaryData</code> object.
     *
     * @return a cloned instance of the current <code>BinaryData</code>
     * object
     */
    public Object clone() {
        try {
            return super.clone();
        } catch( CloneNotSupportedException e ) {
            throw new AssertionError( e.getMessage() );
        }
    }

    /**
     * Returns a string representation of the object.
     *
     * @return  a string representation of the object.
     */
    public String toString() {
        StringBuilder buffer = new StringBuilder();
        buffer.append( "BinaryData, " );
        if( dataArray == null ) {
            buffer.append( "no" );
        } else {
            buffer.append( dataArray.length );
        }
        buffer.append( " bytes." );
        return buffer.toString();
    }
}
