/*
 * FreeMarker: a tool that allows Java programs to generate HTML
 * output using templates.
 * Copyright (C) 1998-2011 Benjamin Geer
 * Email: beroul@users.sourceforge.net
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA  02111-1307, USA.
 */

package freemarker.template;

import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStream;

/**
 * A factory class that can create <tt>BinaryData</tt> objects for various
 * inputs.
 *
 * @author Nicholas Cull
 * @version $Id: BinaryDataFactory.java 1112 2013-04-12 02:59:25Z run2000 $
 * @since 1.9
 */
public final class BinaryDataFactory implements CompileableFactory<BinaryData> {
    /** The singleton instance of this BinaryDataFactory. */
    public static final BinaryDataFactory INSTANCE = new BinaryDataFactory();

    private BinaryDataFactory() {
    }

    /**
     * Return a name for the compilable object created by this factory.
     *
     * @return "binary", to indicate this factory creates a binary compilable
     * object
     */
    public String getName() {
        return "binary";
    }

    /**
     * Create an instance of a <tt>BinaryData</tt> object from the given
     * InputStream.
     *
     * @param params
     * @return a newly created <tt>BinaryData</tt> object from the given
     * input stream
     * @throws IOException there was a problem reading from the input stream
     */
    public BinaryData create(CompileableFactoryParameters params) throws IOException {
        return new BinaryData(params.getStream(), params.getCache());
    }

    /**
     * Create an instance of a <tt>BinaryData</tt> object from the given
     * InputStream.
     *
     * @param stream the input stream from which the <tt>BinaryData</tt> object
     * will be created
     * @return a newly created <tt>BinaryData</tt> object from the given
     * input stream
     * @throws IOException there was a problem reading from the input stream
     */
    public BinaryData createFromStream(InputStream stream) throws IOException {
        return new BinaryData(stream);
    }

    /**
     * Create an instance of a <tt>BinaryData</tt> object from the given
     * InputStream.
     *
     * @param stream the input stream from which the <tt>BinaryData</tt> object
     * will be created
     * @param encoding ignored
     * @return a newly created <tt>BinaryData</tt> object from the given
     * input stream
     * @throws IOException there was a problem reading from the input stream
     */
    public BinaryData createFromStream(InputStream stream, String encoding) throws IOException {
        return new BinaryData(stream);
    }

    /**
     * Create an instance of a <tt>BinaryData</tt> object from the given
     * file.
     *
     * @param file the file from which the <tt>BinaryData</tt> object will be
     * created
     * @return a newly created <tt>BinaryData</tt> object from the given
     * input stream
     * @throws IOException there was a problem reading from the input stream
     */
    public BinaryData createFromFile(File file) throws IOException {
        return new BinaryData(new FileInputStream(file));
    }

    /**
     * Create an instance of a <tt>BinaryData</tt> object from the given
     * file path.
     *
     * @param filePath the file path from which the <tt>BinaryData</tt> object
     * will be created
     * @return a newly created <tt>BinaryData</tt> object from the given
     * input stream
     * @throws IOException there was a problem reading from the input stream
     */
    public BinaryData createFromFile(String filePath) throws IOException {
        return new BinaryData(new FileInputStream(filePath));
    }
}
