/*
 * FreeMarker: a tool that allows Java programs to generate HTML
 * output using templates.
 * Copyright (C) 1998-2013 Benjamin Geer
 * Email: beroul@users.sourceforge.net
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA  02111-1307, USA.
 */

package freemarker.template;

import freemarker.template.cache.Cache;

import java.io.InputStream;

/**
 * Parameter class for a Compileable factory. This is a convenience class
 * for passing parameters to a compileable factory to create new instances
 * of compileable objects. Any additional parameters can be added to this parameter
 * class without having to alter the method signature of the processor
 * itself.
 * <p>
 * This is an immutable class. Calls to any of the <code>withXXX()</code>
 * methods return a new object.
 * </p>
 *
 * @author Nicholas Cull
 * @version $Id: CompileableFactoryParameters.java 1112 2013-04-12 02:59:25Z run2000 $
 * @since 1.9
 */
public final class CompileableFactoryParameters {
    private final InputStream stream;
    private final String encoding;
    private final Cache cache;

    /**
     * Create a new CompileableFactory parameter object with all the fields set.
     *
     * @param stream the input stream from which the compilable object will
     * be created
     * @param encoding the text encoding, if appropriate
     * @param cache the containing cache of the item to be loaded
     */
    private CompileableFactoryParameters(InputStream stream, String encoding, Cache cache) {
        if(stream == null) {
            throw new NullPointerException("Input stream cannot be null");
        }
        this.stream = stream;
        this.encoding = encoding;
        this.cache = cache;
    }

    /**
     * Create a new instance of this compileable factory parameter class
     * with the given input stream.
     *
     * @param stream the input stream to be passed in to the compileable factory
     * @return a new CompileableFactoryParameters object with the given
     * input stream
     * @throws NullPointerException the input stream is <code>null</code>
     */
    public static CompileableFactoryParameters newInstance(InputStream stream) {
        return new CompileableFactoryParameters(stream, null, null);
    }

    /**
     * Create a new instance of this compileable factory parameter class
     * with the given input stream and cache to be associated with.
     *
     * @param stream the input stream to be passed in to the compileable factory
     * @param cache the cache the resulting compiled object will be associated
     * with
     * @return a new CompileableFactoryParameters object with the given
     * input stream
     * @throws NullPointerException the writer is <code>null</code>
     */
    public static CompileableFactoryParameters newInstance(InputStream stream, Cache cache) {
        return new CompileableFactoryParameters(stream, null, cache);
    }

    /**
     * Return a new CompileableFactoryParameters object with the given character
     * encoding in place of the existing character encoding. This ensures that
     * every CompileableFactoryParameters object is immutable.
     *
     * @param encoding the character encoding for the compileable factory
     * to use in place of the existing character encoding
     * @return a new CompileableFactoryParameters object with the new
     * character encoding
     */
    public CompileableFactoryParameters withEncoding(String encoding) {
        return new CompileableFactoryParameters(this.stream, encoding, this.cache);
    }

    /**
     * Return a new CompileableFactoryParameters object with the given cache
     * in place of the existing cache. This ensures that every
     * CompileableFactoryParameters object is immutable.
     *
     * @param cache the cache for the compileable factory to use in place of
     * the existing cache
     * @return a new CompileableFactoryParameters object with the new cache
     */
    public CompileableFactoryParameters withCache(Cache cache) {
        return new CompileableFactoryParameters(this.stream, this.encoding, cache);
    }

    /**
     * Return a new CompileableFactoryParameters object with the given input
     * stream in place of the existing input stream. This ensures that every
     * CompileableFactoryParameters object is immutable.
     *
     * @param stream the input stream for the compileable factory to use
     * in place of the existing input stream
     * @return a new CompileableFactoryParameters object with the new
     * input stream
     * @throws NullPointerException the input stream is <code>null</code>
     */
    public CompileableFactoryParameters withInputStream(InputStream stream) {
        return new CompileableFactoryParameters(stream, this.encoding, this.cache);
    }

    /**
     * Get the input stream for this parameter object.
     *
     * @return the input stream
     */
    public InputStream getStream() {
        return stream;
    }

    /**
     * Get the character encoding for this parameter object.
     *
     * @return the character encoding
     */
    public String getEncoding() {
        return encoding;
    }

    /**
     * Get the cache object for this parameter object.
     *
     * @return the cache object
     */
    public Cache getCache() {
        return cache;
    }
}
