/*
 * FreeMarker: a tool that allows Java programs to generate HTML
 * output using templates.
 * Copyright (C) 1998-2005 Benjamin Geer
 * Email: beroul@users.sourceforge.net
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA  02111-1307, USA.
 */

package freemarker.template;

import java.io.*;
import java.util.Collections;
import java.util.HashMap;
import java.util.Map;
import java.util.Set;

import freemarker.template.cache.Cache;
import freemarker.template.compiler.LinkedListTemplateBuilder;
import freemarker.template.compiler.StandardTemplateParser;
import freemarker.template.compiler.ParseException;
import freemarker.template.instruction.TemplateFunctionModel;

/**
 * <p>An application or servlet can instantiate a <code>Template</code>
 * to compile and process an HTML template.</p>
 *
 * <p>You can pass the filename of the template to a <code>TemplateFactory</code>
 * object, in which case it is compiled immediately.  Once compiled,
 * the resulting <code>Template</code> is stored in an an efficient
 * data structure for later use.</p>
 *
 * <p>To process the template and produce HTML, call the
 * {@link #process(TemplateWriteableHashModel, Writer)} method, which takes
 * a tree of {@link TemplateModel} objects as its data model.  The root
 * node of the tree must be a {@link TemplateWriteableHashModel}.</p>
 *
 * <p>Any error messages from exceptions thrown by the data model, or
 * generated by the <code>Template</code> during compilation or
 * processing, will be included as HTML comments in the output.</p>
 *
 * <p>To facilitate multithreading, <code>Template</code> objects are
 * immutable; if you need to recompile a template, you must make a new
 * <code>Template</code> object. In most cases, it will be sufficient
 * to let a {@link TemplateCache} do this for you.</p>
 *
 * @see TemplateCache
 * @see TemplateFactory
 * @version $Id: Template.java 1166 2013-04-21 08:21:15Z run2000 $
 */
public class Template extends AbstractTemplate
        implements FunctionTemplateProcessor, Serializable {

    /** The root node of the compiled template. */
    protected TemplateProcessor compiledTemplate;
    /** A mapping of all function models in this compiled template. */
    protected Map<String, TemplateFunctionModel> functions;
    /** A cached copy of any parser exception thrown during compilation. */
    protected transient ParseException buildError;

    /** Serialization UUID for this class. */
    private static final long serialVersionUID = 2146772088091073346L;

    /**
     * Serialized form is a TemplateProcessor holding the parsed template
     * tree, followed by an array of FunctionInstructions. The array contains
     * functions to be evaluated at runtime if any functions are called.
     * This is done primarily for type correctness, and avoids serializing a
     * Map object.
     *
     * @serialField compiledTemplate TemplateProcessor object holding the
     * entire compiled template parse tree
     * @serialField functionNames String[] an array of function names
     * @serialField functions TemplateFunctionModel[] an array of function
     * models associated with the names
     */
    private static final ObjectStreamField[] serialPersistentFields
        = { new ObjectStreamField("compiledTemplate", TemplateProcessor.class),
            new ObjectStreamField("functionNames", String[].class),
            new ObjectStreamField("functions", TemplateFunctionModel[].class) };

    /**
     * Constructs a template by compiling it from a Reader.  Calls
     * <code>compile()</code>.
     *
     * @param reader the source of the template file to be compiled.
     */
    public Template(Reader reader) throws IOException {
        super( reader );
    }

    /**
     * Constructs a template by compiling it from a Reader.  Calls
     * <code>compile()</code>.
     *
     * @param reader the source of the template file to be compiled.
     */
    public Template(Reader reader, Cache cache) throws IOException {
        super( reader, cache );
    }

    /**
     * Compiles the template from a <code>Reader</code>.
     *
     * @param reader a <code>Reader</code> from which the
     *               template can be read.
     */
    protected void compile(Reader reader) throws IOException, IllegalArgumentException {

        String text = getTemplateText( reader );
        TemplateProcessor newTemplate = compileText( text );
        compiledTemplate = newTemplate;
    }

    /**
     * Compiles the template text using the standard parser and builder
     * classes. To provide different compilation behavior, subclasses
     * need only override this method.
     *
     * @param text the text to compile.
     * @return a <code>TemplateProcessor</code> representing the compiled template.
     */
    protected TemplateProcessor compileText(String text) throws IOException {
        StandardTemplateParser parser =
            new StandardTemplateParser(this, text);

        LinkedListTemplateBuilder builder =
            new LinkedListTemplateBuilder(this, parser);

        try {
            return builder.build();
        } catch( ParseException e ) {
            buildError = e;
            return null;
        }
    }

    /**
     * Adds a function to the template. Called by the
     * <code>TemplateBuilder</code> at compile-time.
     *
     * @param name the name of the function to be stored
     * @param function the function to be stored by the template
     * @throws IllegalStateException the method has been called after the
     * template has been compiled
     */
    public void addFunction(String name, TemplateFunctionModel function)
            throws IllegalStateException {

        if( compiledTemplate != null ) {
            throw new IllegalStateException( "Template has already been compiled" );
        }
        if (functions == null) {
            functions = new HashMap<String, TemplateFunctionModel>();
        }
        functions.put(name, function);
    }

    /**
     * Retrieves a function from the template.  Called by <code>CallInstruction</code>s
     * and <code>IncludeInstruction</code>s at run-time.
     *
     * @param name the name of the function to be retrieved
     */
    public TemplateFunctionModel getFunction(String name) {
        if(functions == null) {
            return null;
        }
        return functions.get(name);
    }

    /**
     * Retrieve a <code>Set</code> of function names for this template.
     *
     * @return a <code>Set</code> of function names (<code>String</code> objects)
     * that have been defined for this template.
     */
    public Set<String> getFunctionNames() {
        if(functions == null) {
            return Collections.emptySet();
        }
        return Collections.unmodifiableSet(functions.keySet());
    }

    /**
     * Processes the contents of this {@link TemplateProcessor} and
     * outputs the resulting text to a <code>PrintWriter</code>.
     *
     * @param modelRoot the root node of the data model.
     * @param out a <code>PrintWriter</code> to send the output to.
     * @param eventHandler a <code>TemplateEventAdapter</code> for handling any
     * events that occur during processing.
     */
    public void process(TemplateWriteableHashModel modelRoot, PrintWriter out,
            TemplateRuntimeHandler eventHandler) {
        try {
            if(modelRoot == null) {
                modelRoot = new FastHash();
            }
            TemplateProcessorParameters p = TemplateProcessorParameters
                    .newInstance(out)
                    .withModel(modelRoot)
                    .withEventHandler(eventHandler);
            process(p);
        } catch( IOException e ) {
            // Do nothing, since PrintWriter cannot throw IOException
        }
    }

    /**
     * Processes the template, using data from a template model, and outputs
     * the resulting text to a <code>PrintWriter</code>.
     *
     * @param modelRoot the root node of the data model. If <code>null</code>,
     * an empty data model is used.
     * @param out a <code>PrintWriter</code> to send the output to.
     */
    public void process(TemplateWriteableHashModel modelRoot, PrintWriter out) {

        try {
            if (modelRoot == null) {
                modelRoot = new FastHash();
            }
            TemplateProcessorParameters p = TemplateProcessorParameters
                    .newInstance(out)
                    .withModel(modelRoot);
            process(p);
        } catch( IOException e ) {
            // Do nothing, since PrintWriter cannot throw IOException
        }
    }

    /**
     * Processes the template, using an empty data model, and outputs
     * the resulting text to a <code>PrintWriter</code>.
     *
     * @param out a <code>PrintWriter</code> to send the output to.
     */
    public void process(PrintWriter out) {
        try {
            TemplateProcessorParameters p = TemplateProcessorParameters.newInstance(out);
            process(p);
        } catch( IOException e ) {
            // Do nothing, since PrintWriter cannot throw IOException
        }
    }

    /**
     * Processes the contents of this {@link TemplateProcessor} and
     * outputs the resulting text to a <code>Writer</code>.
     *
     * @param modelRoot the root node of the data model.
     * @param out a <code>Writer</code> to send the output to.
     * @param eventHandler a <code>TemplateEventAdapter</code> for handling any
     * events that occur during processing.
     * @return a zero or positive value if template processing was successful,
     * otherwise a negative number to indicate that no template has been
     * compiled
     * @throws IOException an IO error occurred with the <code>Writer</code>
     * during processing
     */
    public ExitStatus process(TemplateWriteableHashModel modelRoot, Writer out,
            TemplateRuntimeHandler eventHandler) throws IOException {
            if (modelRoot == null) {
                modelRoot = new FastHash();
            }
            TemplateProcessorParameters p = TemplateProcessorParameters.newInstance(out)
                    .withModel(modelRoot)
                    .withEventHandler(eventHandler);
            return process(p);
    }

    /**
     * Processes the contents of this {@link TemplateProcessor} and
     * outputs the resulting text to a <code>Writer</code>.
     *
     * @param p template processing parameters
     * @return a zero or positive value if template processing was successful,
     * otherwise a negative number to indicate that no template has been
     * compiled
     * @throws IOException an IO error occurred with the <code>Writer</code>
     * during processing
     */
    public ExitStatus process(TemplateProcessorParameters p) throws IOException {
        if (compiledTemplate != null) {
            if((functions != null) && (!functions.isEmpty())) {
                try {
                    copyFunctions(this, p.getModelRoot());
                } catch( TemplateModelException e ) {
                    p.getEventHandler().fireExceptionThrown(this, new TemplateException(
                            "Couldn't copy functions into the model", e),
                            p.getWriter(),
                            "freemarker.template.Template.process",
                            TemplateRuntimeHandler.Severity.ERROR);
                }
            }
            return compiledTemplate.process(p);
        } else if( buildError != null ) {
            p.getEventHandler().fireExceptionThrown(this, buildError, p.getWriter(),
                    "freemarker.template.Template.process",
                    TemplateRuntimeHandler.Severity.ERROR);
        }
        return TemplateProcessor.ExitStatus.UNCOMPILED_TEMPLATE;
    }

    /**
     * Processes the template, using data from a template model, and outputs
     * the resulting text to a <code>Writer</code>.
     *
     * @param modelRoot the root node of the data model.  If <code>null</code>,
     * an empty data model is used.
     * @param out a <code>Writer</code> to output the text to.
     */
    public void process(TemplateWriteableHashModel modelRoot, Writer out)
            throws IOException {
        if (modelRoot == null) {
            modelRoot = new FastHash();
        }
        TemplateProcessorParameters p = TemplateProcessorParameters
                .newInstance(out)
                .withModel(modelRoot);
        process(p);
    }

    /**
     * Processes the template, using an empty data model, and outputs
     * the resulting text to a <code>Writer</code>.
     *
     * @param out a <code>Writer</code> to output the text to.
     */
    public void process(Writer out) throws IOException {
        TemplateProcessorParameters p = TemplateProcessorParameters.newInstance(out);
        process(p);
    }

    /**
     * <p>Clones the current template. A shallow clone is performed, meaning that
     * changes to the underlying structure of one will affect the structure
     * of the other. As Templates are immutable anyway, this shouldn't be
     * an issue.</p>
     *
     * @return a shallow clone of the current template.
     */
    public Object clone() {
        Template newTemplate = (Template)super.clone();

        if(functions != null) {
            newTemplate.functions = (Map<String, TemplateFunctionModel>)((HashMap<String, TemplateFunctionModel>)functions).clone();
        }
        return newTemplate;
    }

    /**
     * Copies functions from a template into a data model. This lets functions
     * from parent and child templates be available in both contexts.
     *
     * @param template the template from which function instructions will
     * be copied
     * @param modelRoot the template model where functions will be copied to
     * @throws TemplateModelException the model could not accept the function
     */
    protected static void copyFunctions(FunctionTemplateProcessor template,
            TemplateWriteableHashModel modelRoot) throws TemplateModelException {
        Set<String> functionNames = template.getFunctionNames();

        for (String functionName : functionNames) {
            TemplateFunctionModel function = template.getFunction(functionName);
            modelRoot.put(functionName, function);
        }
    }

    /**
     * For serialization, write this object a TemplateProcessor containing
     * the compiled template, and an arrays. The array contains the
     * FunctionInstructions that may be called from the template.
     *
     * @param stream the output stream to write this object to
     */
    private void writeObject(ObjectOutputStream stream) throws IOException {
        ObjectOutputStream.PutField fields = stream.putFields();
        int size = (functions == null) ? 0 : functions.size();
        TemplateFunctionModel[] functionArray = new TemplateFunctionModel[size];
        String[] nameArray = new String[size];

        if( functions != null ) {
            int i = 0;

            for (Map.Entry<String, TemplateFunctionModel> entry : functions.entrySet()) {
                nameArray[i] = entry.getKey();
                functionArray[i] = entry.getValue();
                i++;
            }
        }

        fields.put("compiledTemplate", compiledTemplate);
        fields.put("functionNames", nameArray);
        fields.put("functions", functionArray);
        stream.writeFields();
    }

    /**
     * For serialization, read this object a TemplateProcessor containing
     * the compiled template, and an array of FunctionInstructions. These
     * may be called from the template at runtime if any functions were defined
     * in the template source. Check whether the function list is null.
     *
     * @param stream the input stream to read serialized objects from
     */
    private void readObject(ObjectInputStream stream) throws IOException,
            ClassNotFoundException {

        ObjectInputStream.GetField fields = stream.readFields();

        compiledTemplate = (TemplateProcessor)fields.get("compiledTemplate", null);
        String[] functionNames = (String[])fields.get("functionNames", null);
        TemplateFunctionModel[] functionValues = (TemplateFunctionModel[])fields.get("functions", null);

        if(functionValues == null) {
            throw new InvalidObjectException( "Cannot create a Template with a null function list" );
        }
        functions = new HashMap<String, TemplateFunctionModel>();

        int size = functionValues.length;
        for(int i=0; i<size; i++) {
            String name = functionNames[i];
            TemplateFunctionModel functionValue = functionValues[i];
            functions.put( name, functionValue );
        }
    }

    /**
     * Returns a string representation of the object. The value returned
     * represents the parse tree of the template.
     *
     * @return a string representation of the object.
     */
    public String toString() {
        StringBuilder buffer = new StringBuilder();

        buffer.append( "FreeMarker template, " );
        buffer.append( compiledTemplate );
        return buffer.toString();
    }
}
