/*
 * FreeMarker: a tool that allows Java programs to generate HTML
 * output using templates.
 * Copyright (C) 1998-2011 Benjamin Geer
 * Email: beroul@users.sourceforge.net
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA  02111-1307, USA.
 */

package freemarker.template;

import java.io.*;

/**
 * A factory class that can create <tt>Template</tt> objects for various
 * inputs.
 *
 * @author Nicholas Cull
 * @version $Id $
 * @since 1.9
 */
public final class TemplateFactory implements CompileableFactory<Template> {
    /** The singleton instance of this TemplateFactory. */
    public static final TemplateFactory INSTANCE = new TemplateFactory();

    private TemplateFactory() {
    }

    /**
     * Return a name for the compilable object created by this factory.
     *
     * @return "template", to indicate this factory creates a template
     * compilable object
     */
    public String getName() {
        return "template";
    }

    /**
     * Create an instance of a <tt>Template</tt> object from the given
     * InputStream, using the named character encoding.
     *
     * @param params
     * @return a newly created <tt>Template</tt> object from the given
     * input stream
     * @throws IOException there was a problem reading from the input stream
     */
    public Template create(CompileableFactoryParameters params) throws IOException {
        if(params.getEncoding() == null) {
            return new Template(new InputStreamReader(params.getStream()),
                    params.getCache());
        } else {
            return new Template(
                    new InputStreamReader(params.getStream(),
                            params.getEncoding()),
                    params.getCache());
        }
    }

    /**
     * Create an instance of a <tt>Template</tt> object from the given
     * InputStream, using the platform default character encoding.
     *
     * @param stream the input stream from which the <tt>Template</tt> object
     * will be created
     * @return a newly created <tt>Template</tt> object from the given
     * input stream
     * @throws IOException there was a problem reading from the input stream
     */
    public Template createFromStream(InputStream stream) throws IOException {
        return new Template(new InputStreamReader(stream));
    }

    /**
     * Create an instance of a <tt>Template</tt> object from the given
     * InputStream, using the named character encoding.
     *
     * @param stream the input stream from which the <tt>Template</tt> object
     * will be created
     * @param encoding the character encoding of the input stream
     * @return a newly created <tt>Template</tt> object from the given
     * input stream
     * @throws IOException there was a problem reading from the input stream
     */
    public Template createFromStream(InputStream stream, String encoding) throws IOException {
        return new Template(new InputStreamReader(stream, encoding));
    }

    /**
     * Create an instance of a <tt>Template</tt> object from the given
     * file, using the platform default character encoding.
     *
     * @param file the file from which the <tt>Template</tt> object will be
     * created
     * @return a newly created <tt>Template</tt> object from the given
     * file
     * @throws IOException there was a problem reading from the file
     */
    public Template createFromFile(File file) throws IOException {
        return new Template(new InputStreamReader(new FileInputStream(file)));
    }

    /**
     * Create an instance of a <tt>Template</tt> object from the given
     * file, using the named character encoding.
     *
     * @param file the file from which the <tt>Template</tt> object will be
     * created
     * @param encoding the character encoding of the file
     * @return a newly created <tt>Template</tt> object from the given
     * file
     * @throws IOException there was a problem reading from the file
     */
    public Template createFromFile(File file, String encoding) throws IOException {
        return new Template(new InputStreamReader(new FileInputStream(file), encoding));
    }

    /**
     * Create an instance of a <tt>Template</tt> object from the given
     * file path, using the platform default character encoding.
     *
     * @param filePath the file from which the <tt>Template</tt> object will be
     * created
     * @return a newly created <tt>Template</tt> object from the given
     * file path
     * @throws IOException there was a problem reading from the file path
     */
    public Template createFromFile(String filePath) throws IOException {
        return new Template(new InputStreamReader(new FileInputStream(filePath)));
    }

    /**
     * Create an instance of a <tt>Template</tt> object from the given
     * file, using the named character encoding.
     *
     * @param filePath the file from which the <tt>Template</tt> object will be
     * created
     * @param encoding the character encoding of the file
     * @return a newly created <tt>Template</tt> object from the given
     * file
     * @throws IOException there was a problem reading from the file
     */
    public Template createFromFile(String filePath, String encoding) throws IOException {
        return new Template(new InputStreamReader(new FileInputStream(filePath), encoding));
    }

    /**
     * Create an instance of a <tt>Template</tt> object from the given Reader.
     *
     * @param reader the reader from which the <tt>Template</tt> object
     * will be created
     * @return a newly created <tt>Template</tt> object from the given
     * reader
     * @throws IOException there was a problem reading from the reader
     */
    public Template createFromReader(Reader reader) throws IOException {
        return new Template(reader);
    }

    /**
     * Create an instance of a <tt>Template</tt> object from the given String.
     *
     * @param str the string from which the <tt>Template</tt> object
     * will be created
     * @return a newly created <tt>Template</tt> object from the given
     * string
     * @throws IOException there was a problem reading from the string
     */
    public Template createFromString(String str) throws IOException {
        return new Template(new StringReader(str));
    }
}
