/*
 * FreeMarker: a tool that allows Java programs to generate HTML
 * output using templates.
 * Copyright (C) 1998-2013 Benjamin Geer
 * Email: beroul@users.sourceforge.net
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA  02111-1307, USA.
 */

package freemarker.template;

import java.io.OutputStreamWriter;
import java.io.Writer;

/**
 * Parameter class for a Template processor. This is a convenience class
 * for passing parameters to a template processor to apply the template
 * to a model. Any additional parameters can be added to this parameter
 * class without having to alter the method signature of the processor
 * itself.
 * <p>
 * This is an immutable class. Calls to any of the <code>withXXX()</code>
 * methods return a new object.
 * </p>
 * <p>
 *     Typical usage:
 * </p>
 * <pre>
 * Reader reader = ...;
 * Template template = new Template(reader);
 * TemplateProcessorParameters tpp = TemplateProcessorParameters
 *     .newInstance(writer)
 *     .withEventHandler(TemplateEventAdapter.DefaultEventAdapter)
 *     .withEscape(HtmlEscape.getInstance());
 * template.process(tpp);
 * </pre>
 *
 * @author Nicholas Cull
 * @version $Id: TemplateProcessorParameters.java 1112 2013-04-12 02:59:25Z run2000 $
 * @since 1.9
 */
public final class TemplateProcessorParameters {
    private final TemplateWriteableHashModel modelRoot;
    private final TemplateRuntimeHandler eventHandler;
    private final TemplateModel escape;
    private final Writer out;

    /**
     * Create a new TemplateProcessorParameters object with all the
     * fields assigned.
     *
     * @param modelRoot the root node of the data model.
     * @param out a <code>Writer</code> to send the output to.
     * @param eventHandler a <code>TemplateEventAdapter</code> for handling any
     * events that occur during processing.
     * @param escape the TemplateModel to use for escaping variables written
     * to the writer, may be <code>null</code>.
     */
    private TemplateProcessorParameters(TemplateWriteableHashModel modelRoot, Writer out,
                TemplateRuntimeHandler eventHandler, TemplateModel escape) {
        if(out == null) {
            throw new NullPointerException("Writer cannot be null");
        }
        if(modelRoot == null) {
            throw new NullPointerException("Model root cannot be null");
        }
        if(eventHandler == null) {
            throw new NullPointerException("Event handler cannot be null");
        }
        if(escape != null) {
            if(!(escape instanceof TemplateTransformModel) && !(escape instanceof TemplateTransformModel2)) {
                throw new IllegalArgumentException("Escape must be a TemplateTransformModel or a TemplateTransformModel2");
            }
        }
        this.modelRoot = modelRoot;
        this.out = out;
        this.eventHandler = eventHandler;
        this.escape = escape;
    }

    /**
     * Create a new instance of this template processor parameter class.
     * The instance contains an empty model root, the standard output
     * stream, and the default event adapter.
     *
     * @return a new template processor parameter object
     */
    public static TemplateProcessorParameters newInstance() {
        return new TemplateProcessorParameters(
                new FastHash(),
                new OutputStreamWriter(System.out),
                TemplateEventAdapter.DefaultEventAdapter,
                null
        );
    }

    /**
     * Create a new instance of this template processor parameter class
     * with the given writer. The instance contains an empty model root,
     * the default event adapter, and no template escaper.
     *
     * @param writer the writer to be passed in to the template processor
     * @return a new TemplateProcessorParameters object with the given
     * writer
     * @throws NullPointerException the writer is <code>null</code>
     */
    public static TemplateProcessorParameters newInstance(Writer writer) {
        return new TemplateProcessorParameters(
                new FastHash(),
                writer,
                TemplateEventAdapter.DefaultEventAdapter,
                null
        );
    }

    /**
     * Return a new TemplateProcessorParameters object with the given writer
     * in place of the existing writer. This ensures that every
     * TemplateProcessorParameters object is immutable.
     *
     * @param writer the writer to use in place of the existing writer
     * @return a new TemplateProcessorParameters object with the new writer
     * @throws NullPointerException the writer is <code>null</code>
     */
    public TemplateProcessorParameters withWriter(Writer writer) {
        return new TemplateProcessorParameters(this.modelRoot, writer, this.eventHandler, this.escape);
    }

    /**
     * Return a new TemplateProcessorParameters object with the given template
     * model in place of the existing template model. This ensures that every
     * TemplateProcessorParameters object is immutable.
     *
     * @param modelRoot the template model to use in place of the existing
     * template model
     * @return a new TemplateProcessorParameters object with the new template
     * model
     * @throws NullPointerException the model root is <code>null</code>
     */
    public TemplateProcessorParameters withModel(TemplateWriteableHashModel modelRoot) {
        return new TemplateProcessorParameters(modelRoot, this.out, this.eventHandler, this.escape);
    }

    /**
     * Return a new TemplateProcessorParameters object with the given event
     * handler in place of the existing event handler. This ensures that every
     * TemplateProcessorParameters object is immutable.
     *
     * @param eventHandler the event handler to use in place of the existing
     * event handler
     * @return a new TemplateProcessorParameters object with the new event
     * handler
     * @throws NullPointerException the event handler is <code>null</code>
     */
    public TemplateProcessorParameters withEventHandler(TemplateRuntimeHandler eventHandler) {
        return new TemplateProcessorParameters(this.modelRoot, this.out, eventHandler, this.escape);
    }

    /**
     * Return a new TemplateProcessorParameters object with the given escape
     * TemplateModel in place of the existing escape. This ensures that every
     * TemplateProcessorParameters object is immutable.
     *
     * @param escape the escape TemplateModel to use in place of the existing
     * escape
     * @return a new TemplateProcessorParameters object with the new escape
     * TemplateModel
     */
    public TemplateProcessorParameters withEscape(TemplateModel escape) {
        return new TemplateProcessorParameters(this.modelRoot, this.out, this.eventHandler, escape);
    }

    /**
     * Get the Template model root for this parameter object.
     *
     * @return the template model root
     */
    public TemplateWriteableHashModel getModelRoot() {
        return modelRoot;
    }

    /**
     * Get the writer for this parameter object.
     *
     * @return the writer
     */
    public Writer getWriter() {
        return out;
    }

    /**
     * Get the Template runtime handler for this parameter object.
     *
     * @return the runtime handler
     */
    public TemplateRuntimeHandler getEventHandler() {
        return eventHandler;
    }

    /**
     * Get the Template escape model for this parameter object.
     *
     * @return the template escape model, possibly <code>null</code>
     */
    public TemplateModel getEscape() {
        return escape;
    }
}
