/*
 * FreeMarker: a tool that allows Java programs to generate HTML
 * output using templates.
 * Copyright (C) 1998-2005 Benjamin Geer
 * Email: beroul@users.sourceforge.net
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA  02111-1307, USA.
 */

package freemarker.template;

import freemarker.template.cache.Cache;

import java.io.*;
import java.util.Arrays;

/**
 * <p>An <code>UnparsedTemplate</code> consists only of text. No logic
 * beyond simple outputting is performed.
 *
 * <p>You can pass the filename of the template to a <code>UnparsedTemplateFactory</code>
 * object, in which case it is read immediately.  Once read, the unparsed template
 * is stored in an an character array for later use.
 *
 * <p>To process the unparsed template, call the {@link TemplateProcessor#process} method,
 * which takes an optional tree of {@link TemplateModel} objects as its
 * data model.  The root node of the tree must be a {@link TemplateWriteableHashModel}.
 *
 * <p>Any error messages will be included as HTML comments in the output.
 *
 * <p>To facilitate multithreading, <code>UnparsedTemplate</code> objects are
 * immutable; if you need to recompile a template, you must make a new
 * <code>UnparsedTemplate</code> object.  In most cases, it will be sufficient
 * to let a {@link TemplateCache} do this for you.
 *
 * @see TemplateCache
 * @see UnparsedTemplateFactory
 * @version $Id: UnparsedTemplate.java 1166 2013-04-21 08:21:15Z run2000 $
 */
public class UnparsedTemplate extends AbstractTemplate implements Serializable {
    /**
     * The unparsed template text.
     *
     * @serial The unparsed template text
     */
    protected char[] templateText;
    private transient int hashCode;

    /** Serialization UUID for this class. */
    private static final long serialVersionUID = 6100824836126362261L;

    /**
     * Serialized unparsed template as a String object. This means that for the
     * 1.2 series of JVMs, the serialized unparsed template cannot be larger
     * than 64kB.
     *
     * @serialField textValue String the <code>String</code> value of this
     * template
     */
    private static final ObjectStreamField[] serialPersistentFields
        = { new ObjectStreamField("textValue", String.class) };

    /**
     * Constructs an unparsed template by compiling it from a Reader. Calls
     * <code>compile()</code>.
     *
     * @param reader the source of the template file to be compiled.
     */
    public UnparsedTemplate(Reader reader) throws IOException {
        super( reader );
    }

    /**
     * Constructs an unparsed template by compiling it from a Reader. Calls
     * <code>compile()</code>.
     *
     * @param reader the source of the template file to be compiled.
     */
    public UnparsedTemplate(Reader reader, Cache cache) throws IOException {
        super( reader, cache );
    }

    /**
     * Compiles the template from a <code>Reader</code>.
     *
     * @param reader an <code>Reader</code> from which the
     *               template can be read.
     */
    protected void compile(Reader reader) throws IOException,
            IllegalArgumentException {

        String text = getTemplateText( reader );
        templateText = text.toCharArray();
        hashCode = 0;
    }

    /**
     * Processes the contents of this <code>UnparsedTemplate</code> and
     * outputs the resulting text to a <code>Writer</code>.
     *
     * @param modelRoot the root node of the data model.
     * @param out a <code>Writer</code> to send the output to.
     * @param eventHandler a <code>TemplateEventAdapter</code> for handling any
     * events that occur during processing.
     * @throws IOException an IO error occurred with the <code>Writer</code>
     * during processing
     */
    public ExitStatus process(TemplateWriteableHashModel modelRoot, Writer out,
            TemplateRuntimeHandler eventHandler) throws IOException {
        if (templateText != null) {
            out.write( templateText );
        }
        return TemplateProcessor.ExitStatus.OK;
    }

    /**
     * Processes the contents of this <code>UnparsedTemplate</code> and
     * outputs the resulting text to a <code>Writer</code>.
     *
     * @param p template processing parameters
     * @throws IOException an IO error occurred with the <code>Writer</code>
     * during processing
     */
    public ExitStatus process(TemplateProcessorParameters p) throws IOException {
        if (templateText != null) {
            p.getWriter().write(templateText);
        }
        return TemplateProcessor.ExitStatus.OK;
    }

    /**
     * Processes the template, using data from a template model, and outputs
     * the resulting text to a <code>Writer</code>.
     *
     * @param modelRoot the root node of the data model.
     * @param out a <code>Writer</code> to output the text to.
     */
    public void process(TemplateWriteableHashModel modelRoot, Writer out)
            throws IOException {
        if (templateText != null) {
            out.write( templateText );
        }
    }

    /**
     * Processes the template, using an empty data model, and outputs
     * the resulting text to a <code>Writer</code>.
     *
     * @param out a <code>Writer</code> to output the text to.
     */
    public void process(Writer out) throws IOException {
        if (templateText != null) {
            out.write( templateText );
        }
    }

    /**
     * Return the String value of this object
     *
     * @return the template text, if any, of this object
     */
    public String toString() {
        if( templateText == null ) {
            return "null";
        }
        return new String( templateText );
    }

    /**
     * Tests this object for equality with the given object.
     *
     * @param o the object to be compared against
     * @return <code>true</code> if the two objects are equal, otherwise
     * <code>false</code>
     */
    public boolean equals( Object o ) {
        if( this == o ) {
            return true;
        }
        if( !( o instanceof UnparsedTemplate )) {
            return false;
        }

        final UnparsedTemplate unparsedTemplate = ( UnparsedTemplate ) o;
        return Arrays.equals( templateText, unparsedTemplate.templateText );
    }

    /**
     * Retrieve the hash code for this object
     *
     * @return a hash value corresponding to the value of this object
     */
    public int hashCode() {
        if( hashCode == 0 ) {
            if( templateText == null ) {
                hashCode = 3;
            }
            hashCode = new String( templateText ).hashCode() + 3;
        }
        return hashCode;
    }

    /**
     * For serialization, write this object as a String.
     *
     * @param stream the output stream to write this object to
     */
    private void writeObject(ObjectOutputStream stream) throws IOException {
        ObjectOutputStream.PutField fields = stream.putFields();

        if( templateText == null ) {
            fields.put("textValue", null);
        } else {
            fields.put("textValue", new String( templateText ));
        }
        stream.writeFields();
    }

    /**
     * For serialization, read this object as a String.
     *
     * @param stream the input stream to read serialized objects from
     */
    private void readObject(ObjectInputStream stream)
            throws IOException, ClassNotFoundException {

        ObjectInputStream.GetField fields = stream.readFields();
        String value = (String)fields.get("textValue", null);

        // Recreate the original value
        if( value != null ) {
            templateText = value.toCharArray();
        }
    }
}
