/*
 * FreeMarker: a tool that allows Java programs to generate HTML
 * output using templates.
 * Copyright (C) 1998-2011 Benjamin Geer
 * Email: beroul@users.sourceforge.net
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA  02111-1307, USA.
 */

package freemarker.template.cache;

import java.io.File;

/**
 * Parameter class for creating new <tt>FileRetriever</tt> objects.
 * This is a convenience class for passing parameters to a file retriever
 * to configure the retriever. Any additional parameters can be added
 * to this parameter class without having to alter the method signature of
 * the retriever itself.
 * <p>
 * This is an immutable class. Calls to any of the <code>withXXX()</code>
 * methods return a new object.
 * </p>
 *
 * @author Nicholas Cull
 * @version $Id: FileRetrieverParameters.java 1101 2013-04-01 04:17:32Z run2000 $
 * @since 1.9
 */
public final class FileRetrieverParameters implements CacheRetrieverParameters {
    /** The root directory where the retriever will get files. */
    private final File directoryRoot;
    /** The filename suffix required for a file to be retrieved. */
    private final String filenameSuffix;
    /** The text encoding of the template files. */
    private final String encoding;
    /** The template registry to use to instantiate objects. */
    private final TemplateRegistry registry;

    /**
     * Create a new FileRetrieverParameters object with all the fields
     * assigned.
     *
     * @param directoryRoot the directory root containing the templates
     * @param filenameSuffix the common suffix of all templates, if any
     * @param encoding the character encoding of the templates
     * @param registry the registry for compiling the templates
     */
    private FileRetrieverParameters(File directoryRoot, String filenameSuffix, String encoding, TemplateRegistry registry) {
        this.directoryRoot = directoryRoot;
        this.filenameSuffix = filenameSuffix;
        this.encoding = encoding;
        this.registry = registry;
    }

    /**
     * Create a new instances of the FileRetrieverParameters class. This includes
     * a default TemplateRegistry.
     *
     * @return a new FileRetrieverParameters object with a new template registry
     */
    public static FileRetrieverParameters newInstance() {
        return new FileRetrieverParameters(null, null, null, new TemplateRegistry());
    }

    /**
     * Create a new instances of the FileRetrieverParameters class, with
     * the given directory root, and a default TemplateRegistry.
     *
     * @param directoryRoot the root directory containing files to be retrieved
     * @return a new FileRetrieverParameters object with a new template registry
     */
    public static FileRetrieverParameters newInstance(File directoryRoot) {
        return new FileRetrieverParameters(directoryRoot, null, null, new TemplateRegistry());
    }

    /**
     * Create a new instances of the FileRetrieverParameters class, with
     * the given directory root, and a default TemplateRegistry.
     *
     * @param directoryRoot the root directory containing files to be retrieved
     * @return a new FileRetrieverParameters object with a new template registry
     */
    public static FileRetrieverParameters newInstance(String directoryRoot) {
        return new FileRetrieverParameters(new File(directoryRoot), null, null, new TemplateRegistry());
    }

    /**
     * Return a new FileRetrieverParameters object with the given directory root
     * in place of the existing directory root. This ensures that every
     * FileRetrieverParameters object is immutable.
     *
     * @param directoryRoot the directory root to use in place of the existing
     * directory root
     * @return a new FileRetrieverParameters object with the new directory root
     */
    public FileRetrieverParameters withDirectoryRoot(File directoryRoot) {
        return new FileRetrieverParameters(directoryRoot, filenameSuffix, encoding, registry);
    }

    /**
     * Return a new FileRetrieverParameters object with the given directory root
     * in place of the existing directory root. This ensures that every
     * FileRetrieverParameters object is immutable.
     *
     * @param directoryRoot the directory root to use in place of the existing
     * directory root
     * @return a new FileRetrieverParameters object with the new directory root
     */
    public FileRetrieverParameters withDirectoryRoot(String directoryRoot) {
        return new FileRetrieverParameters(new File(directoryRoot), filenameSuffix, encoding, registry);
    }

    /**
     * Return a new FileRetrieverParameters object with the given filename suffix
     * in place of the existing filename suffix. This ensures that every
     * FileRetrieverParameters object is immutable.
     *
     * @param filenameSuffix the filename suffix to use in place of the existing
     * filename suffix
     * @return a new FileRetrieverParameters object with the new filename suffix
     */
    public FileRetrieverParameters withFilenameSuffix(String filenameSuffix) {
        return new FileRetrieverParameters(directoryRoot, filenameSuffix, encoding, registry);
    }

    /**
     * Return a new FileRetrieverParameters object with the given character
     * encoding in place of the existing character encoding. This ensures that
     * every FileRetrieverParameters object is immutable.
     *
     * @param encoding the character encoding to use in place of the existing
     * character encoding
     * @return a new FileRetrieverParameters object with the new character
     * encoding
     */
    public FileRetrieverParameters withEncoding(String encoding) {
        return new FileRetrieverParameters(directoryRoot, filenameSuffix, encoding, registry);
    }

    /**
     * Return a new FileRetrieverParameters object with the given template
     * registry in place of the existing template registry. This ensures that
     * every FileRetrieverParameters object is immutable.
     *
     * @param registry the template registry to use in place of the existing
     * template registry
     * @return a new FileRetrieverParameters object with the new template
     * registry
     */
    public FileRetrieverParameters withRegistry(TemplateRegistry registry) {
        return new FileRetrieverParameters(directoryRoot, filenameSuffix, encoding, registry);
    }

    /**
     * Get the directory root for this parameter object.
     *
     * @return the directory root
     */
    public File getDirectoryRoot() {
        return directoryRoot;
    }

    /**
     * Get the filename suffix for this parameter object.
     *
     * @return the filename suffix
     */
    public String getFilenameSuffix() {
        return filenameSuffix;
    }

    /**
     * Get the character encoding for this parameter object.
     *
     * @return the character encoding
     */
    public String getEncoding() {
        return encoding;
    }

    /**
     * Get the template registry for this parameter object.
     *
     * @return the template registry
     */
    public TemplateRegistry getRegistry() {
        return registry;
    }
}
