/*
 * FreeMarker: a tool that allows Java programs to generate HTML
 * output using templates.
 * Copyright (C) 1998-2004 Benjamin Geer
 * Email: beroul@users.sourceforge.net
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA  02111-1307, USA.
 */

package freemarker.template.cache;

import java.util.HashMap;
import java.util.Iterator;
import java.util.Set;
import java.util.Collections;
import freemarker.template.TemplateException;

/**
 * Implements a load-on-demand caching strategy with periodic refreshes.
 * The main difference between this and the plain load-on-demand strategy
 * is that we only update the templates when the timer fires, rather than
 * every time the cache item is accessed.
 *
 * @author  Nicholas Cull
 * @version $Id: LoDWithRefreshCachingStrategy.java 1153 2013-04-15 10:59:37Z run2000 $
 * @since 1.8
 */
public final class LoDWithRefreshCachingStrategy extends BaseCachingStrategy {
    private HashMap<String, CacheElement> cache = new HashMap<String, CacheElement>();
    private UpdateTimer timer;
    private long delay = 5000; // five seconds
    private int maximumAge = 0; // Default to no expiry
    private String defaultTemplate;

    /** Creates new LoadOnDemandCachingStrategy. */
    public LoDWithRefreshCachingStrategy() {
    }

    /**
     * Creates new LoDWithRefreshCachingStrategy with the given parameters.
     *
     * @param params the parameters with which to initialize this strategy
     */
    public LoDWithRefreshCachingStrategy(StrategyParameters params) {
        eventHandler = params.getEventHandler();
        defaultTemplate = params.getDefaultTemplateType();
        retriever = params.getRetriever();
        delay = params.getDelay() * 1000;
        maximumAge = params.getMaximumAge();
    }

    /**
     * Returns the interval between two cache updates. This is meaningful
     * only if the cache policy is a load-on-demand or preload type.
     *
     * @return the number of seconds between cache updates
     */
    public synchronized long getDelay() {
        return delay / 1000;
    }

    /**
     * Retrieves the maximum age a cache item can be before it is evicted
     * from the cache. The age is determined as the number of cache updates
     * since the item was last accessed. This is meaningful
     * only if the cache policy is a load-on-demand type.
     *
     * @return the maximum age before an item is evicted from the cache
     */
    public synchronized int getMaximumAge() {
        return maximumAge;
    }

    /**
     * Retrieves the default template type to be created when retrieving
     * items from the cache.
     *
     * @return the type of template cached by default
     */
    public String getDefaultTemplateType() {
        return defaultTemplate;
    }

    /**
     * Retrieves an item from the cache, according to the loading policy
     * implemented.
     *
     * @param name the name of the item to retrieve
     * @return the corresponding <code>Cacheable</code> object, or <code>null</code>
     * if not found or an error has occurred
     */
    public Cacheable getItem(String name) {
        return getItem(name, defaultTemplate);
    }

    /**
     * Retrieves an item from the cache, according to the loading policy
     * implemented.
     *
     * @param name the name of the item to retrieve
     * @param type the type of item to be retrieved
     * @return the corresponding <code>Cacheable</code> object, or <code>null</code>
     * if not found or an error has occurred
     */
    public Cacheable getItem(String name, String type) {
        CacheElement element;
        Cacheable item;
        long lastModified;

        element = cache.get(name);

        try {
            if (element == null) {
                // Check the connection first
                if( ! connectionOk() ) {
                    return null;
                }

                HashMap<String, CacheElement> newCache = (HashMap<String, CacheElement>)cache.clone();
                lastModified = retriever.lastModified( name );
                item = retriever.loadData(name, type, this);
                newCache.put(name, new CacheElement(name, type, item, lastModified));
                eventHandler.fireElementUpdated( this, name, lastModified );
                synchronized(this) {
                    cache = newCache;
                }
            } else if(( type != null ) && ( ! type.equals( element.getType() ))) {
                // Check the connection first
                if( ! connectionOk() ) {
                    return null;
                }

                lastModified = retriever.lastModified( name );
                item = retriever.loadData(name, type, this);
                synchronized(this) {
                    cache.put(name, new CacheElement(name, type, item, lastModified));
                }
                eventHandler.fireElementUpdated( this, name, lastModified );
            } else {
                // Return the item we already have.
                item = element.getObject();
            }
        } catch( TemplateException e ) {
            eventHandler.fireElementUpdateFailed( this, name, e );
            return null;
        }
        return item;
    }

    /**
     * Retrieves a list of objects currently in the cache.
     *
     * @return an iterator that can recurse over the cached objects. May return
     * <code>null</code> if there are no items in the cache, or the strategy
     * does not implement a cache.
     */
    public Iterator<CacheElement> listCachedFiles() {
        return Collections.unmodifiableCollection(cache.values()).iterator();
    }

    /**
     * Asks for a "blank" update. It is up to the
     * implementation to determine what has to be updated.
     */
    public void update() throws InterruptedException {
        if( connectionOk()) {
            updateCachedItems();
        }
    }

    /**
     * Asks for the named object to be updated.
     *
     * @param name the name of the object to update
     */
    public void update(String name) {
        // Do nothing
    }

    /**
     * Asks for the named object to be updated.
     *
     * @param name the name of the object to update
     * @param type the type of the object to update
     */
    public void update(String name, String type) {
        // Do nothing
    }

    /**
     * Stops automatically updating the cache.
     */
    public synchronized void stopAutoUpdate() {
        if (timer != null) {
            timer.stopTiming();
            timer = null;
        }
    }

    /**
     * Begins automatic updates of the cache.
     */
    public synchronized void startAutoUpdate() {
        stopAutoUpdate();

        try {
            update();
            timer = new UpdateTimer(this, delay);
            timer.startTiming(1);
        } catch( InterruptedException e ) {
            // Propagate the interrupted flag to the caller
            Thread.currentThread().interrupt();
        }
    }

    /**
     * Removes from the cache objects that correspond to deleted items.
     */
    private void updateCachedItems() throws InterruptedException {
        HashMap<String, CacheElement> newCache = (HashMap<String, CacheElement>)cache.clone();
        Set<String> keySet = newCache.keySet();
        Iterator<String> keyIterator = keySet.iterator();

        // Iterate through existing items, and check whether they're
        // still in the main data store
        while (keyIterator.hasNext()) {
            String name = keyIterator.next();
            CacheElement element = newCache.get(name);
            element.age();

            if ((!retriever.exists(name)) || (element.isExpired(maximumAge))) {
                 keyIterator.remove();
                 eventHandler.fireElementRemoved( this, name );
            } else {
                long lastModified;

                try {
                    lastModified = retriever.lastModified(name);
                } catch( TemplateException e ) {
                    eventHandler.fireElementUpdateFailed( this, name, e );
                    continue;
                }

                if( lastModified > element.lastModified()) {
                    String type = element.getType();
                    Cacheable item;
                    try {
                        item = retriever.loadData(name, type, this);
                    } catch( TemplateException e ) {
                        eventHandler.fireElementUpdateFailed( this, name, e );
                        item = null;
                    }
                    if( item != null ) {
                        newCache.put(name, new CacheElement(name, type, item, lastModified));
                        eventHandler.fireElementUpdated( this, name, lastModified );
                    }
                }
            }

            // Check whether the thread has been interrupted in the mean time.
            // If so, exit the loop immediately.
            if( Thread.interrupted() ) {
                throw new InterruptedException();
            }
        }

        // Finally, update the cache with the new contents
        synchronized(this) {
            cache = newCache;
        }
    }

    /**
     * Clears all the elements in the cache.
     */
    public void clearCache() {
        cache = new HashMap<String, CacheElement>();
    }

    /**
     * Returns a string representation of the object.
     *
     * @return a <code>String</code> representation of the object
     */
    public String toString() {
        StringBuilder buffer = new StringBuilder();
        buffer.append( "LoDWithRefreshCachingStrategy, " );
        buffer.append( cache.size() );
        buffer.append( " cached items. " );
        if( retriever != null ) {
            buffer.append( retriever.toString() );
        }
        return buffer.toString();
    }
}
