/*
 * FreeMarker: a tool that allows Java programs to generate HTML
 * output using templates.
 * Copyright (C) 1998-2011 Benjamin Geer
 * Email: beroul@users.sourceforge.net
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA  02111-1307, USA.
 */

package freemarker.template.cache;

/**
 * Parameter class for creating new <tt>CachingStrategy</tt> objects.
 * This is a convenience class for passing parameters to a caching strategy
 * to configure the cache. Any additional parameters can be added
 * to this parameter class without having to alter the method signature of
 * the cache itself.
 * <p>
 * This is an immutable class. Calls to any of the <code>withXXX()</code>
 * methods return a new object.
 * </p>
 *
 * @author Nicholas Cull
 * @version $Id: StrategyParameters.java 1112 2013-04-12 02:59:25Z run2000 $
 * @since 1.9
 */
public final class StrategyParameters implements Cloneable {
    /** The retriever from which cacheable items will be retrieved */
    private final CacheRetriever cRetriever;
    /** The event handler for cache events */
    private final CacheEventAdapter cEventHandler;
    /** The default template type when no type is specified by the caller */
    private final String aDefaultTemplateType;
    /** The time in seconds between cache refreshes, if appropriate */
    private final long nDelay;
    /** The maximum age before items are expired from the cache, if appropriate */
    private final int nMaximumAge;

    /**
     * Create a new StrategyParameters object with all the fields initialized.
     *
     * @param retriever the retriever from which cacheable items will be retrieved
     * @param eventHandler the event handler for cache events
     * @param defaultTemplateType the default template type when no type is specified
     * by the caller
     * @param delay the time in seconds between cache refreshes, if appropriate
     * @param maximumAge the maximum age before items are expired from the cache,
     * if appropriate
     */
    private StrategyParameters(CacheRetriever retriever, CacheEventAdapter eventHandler, String defaultTemplateType,
                              long delay, int maximumAge) {
        this.cRetriever = retriever;
        this.cEventHandler = eventHandler;
        this.aDefaultTemplateType = defaultTemplateType;
        this.nDelay = delay;
        this.nMaximumAge = maximumAge;
    }

    /**
     * Create a new instances of the StrategyParameters class. This includes
     * a default file retriever and cache event adapter.
     *
     * @return a new StrategyParameters object with a default file retriever
     * and cache event adapter
     */
    public static StrategyParameters newInstance() {
        return new StrategyParameters(
                new FileRetriever(),
                new CacheEventAdapter(),
                "template",
                5L,
                0
        );
    }

    /**
     * Create a new instances of the StrategyParameters class with the given
     * cache retriever. This includes a default cache event adapter.
     *
     * @param retriever the cache retriever for this strategy parameter object
     * @return a new StrategyParameters object with the given cache retriever
     * and a default cache event adapter
     */
    public static StrategyParameters newInstance(CacheRetriever retriever) {
        return new StrategyParameters(
                retriever,
                new CacheEventAdapter(),
                "template",
                5L,
                0
        );
    }

    /**
     * Create a new instances of the StrategyParameters class with the given
     * cache retriever and delay between cache refreshes.
     * This includes a default cache event adapter.
     *
     * @param retriever the cache retriever for this strategy parameter object
     * @param delay the time in seconds between cache refreshes
     * @return a new StrategyParameters object with the given cache retriever
     * and a default cache event adapter
     */
    public static StrategyParameters newInstance(CacheRetriever retriever, long delay) {
        return new StrategyParameters(
                retriever,
                new CacheEventAdapter(),
                "template",
                delay,
                0
        );
    }

    /**
     * Return a new StrategyParameters object with the given cache retriever
     * in place of the existing cache retriever. This ensures that every
     * StrategyParameters object is immutable.
     *
     * @param retriever the cache retriever to use in place of the existing
     * cache retriever
     * @return a new StrategyParameters object with the new cache retriever
     */
    public StrategyParameters withCacheRetriever(CacheRetriever retriever) {
        return new StrategyParameters(
                retriever,
                cEventHandler,
                aDefaultTemplateType,
                nDelay,
                nMaximumAge
        );
    }

    /**
     * Return a new StrategyParameters object with the given event handler
     * in place of the existing event handler. This ensures that every
     * StrategyParameters object is immutable.
     *
     * @param eventHandler the event handler to use in place of the existing
     * event handler
     * @return a new StrategyParameters object with the new event handler
     */
    public StrategyParameters withEventHandler(CacheEventAdapter eventHandler) {
        return new StrategyParameters(
                cRetriever,
                eventHandler,
                aDefaultTemplateType,
                nDelay,
                nMaximumAge
        );
    }

    /**
     * Return a new StrategyParameters object with the given default template
     * type in place of the existing default template type. This ensures that
     * every StrategyParameters object is immutable.
     *
     * @param defaultTemplateType the default template type to use in place of
     * the existing default template type
     * @return a new StrategyParameters object with the new default template type
     */
    public StrategyParameters withDefaultTemplateType(String defaultTemplateType) {
        return new StrategyParameters(
                cRetriever,
                cEventHandler,
                defaultTemplateType,
                nDelay,
                nMaximumAge
        );
    }

    /**
     * Return a new StrategyParameters object with the given delay
     * in place of the existing delay. This ensures that every
     * StrategyParameters object is immutable.
     *
     * @param delay the delay in seconds to use in place of the existing delay
     * @return a new StrategyParameters object with the new delay
     */
    public StrategyParameters withDelay(long delay) {
        return new StrategyParameters(
                cRetriever,
                cEventHandler,
                aDefaultTemplateType,
                delay,
                nMaximumAge
        );
    }

    /**
     * Return a new StrategyParameters object with the given maximum age
     * in place of the existing maximum age. This ensures that every
     * StrategyParameters object is immutable.
     *
     * @param maximumAge the maximum age in multiples of <code>delay</code>
     * seconds to use in place of the existing maximum age
     * @return a new StrategyParameters object with the new maximum age
     */
    public StrategyParameters withMaximumAge(int maximumAge) {
        return new StrategyParameters(
                cRetriever,
                cEventHandler,
                aDefaultTemplateType,
                nDelay,
                maximumAge
        );
    }

    /**
     * Get the cache retriever for this strategy parameter object.
     *
     * @return the cache retriever
     */
    public CacheRetriever getRetriever() {
        return cRetriever;
    }

    /**
     * Get the cache event adapter for this strategy parameter object.
     *
     * @return the cache event adapter
     */
    public CacheEventAdapter getEventHandler() {
        return cEventHandler;
    }

    /**
     * Get the default template type for this strategy parameter object.
     *
     * @return the default template type
     */
    public String getDefaultTemplateType() {
        return aDefaultTemplateType;
    }

    /**
     * Get the delay for this strategy parameter object.
     *
     * @return the delay in seconds
     */
    public long getDelay() {
        return nDelay;
    }

    /**
     * Get the maximum age for this strategy parameter object.
     *
     * @return the maximum age as a multiple of <code>delay</code> seconds
     */
    public int getMaximumAge() {
        return nMaximumAge;
    }

    /**
     * Clone this object.
     *
     * @return a shallow clone of this parameter object
     */
    @Override
    protected Object clone() {
        try {
            return super.clone();
        } catch (CloneNotSupportedException e) {
            throw new Error("Cloneable class not supported");
        }
    }
}
