/*
 * FreeMarker: a tool that allows Java programs to generate HTML
 * output using templates.
 * Copyright (C) 1998-2004 Benjamin Geer
 * Email: beroul@users.sourceforge.net
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA  02111-1307, USA.
 */

package freemarker.template.cache;

import java.util.Map;
import java.util.HashMap;

import freemarker.template.*;

/**
 * <p>Stores a register of prototype templates, which can be retrieved by the
 * template cache whenever it needs to compile a template. This allows the
 * cache to store and retrieve different template implementations as
 * required.</p>
 *
 * <p>This is an implementation of the Parameterized Factory Method
 * pattern.</p>
 *
 * @author Nicholas Cull
 * @version $Id: TemplateRegistry.java 1153 2013-04-15 10:59:37Z run2000 $
 */
public class TemplateRegistry implements Cloneable {
    /** A map of template types that can be instantiated by this object. */
    protected Map<String, CompileableFactory> m_cTemplates = new HashMap<String, CompileableFactory>();

    /** Creates new <code>TemplateRegistry</code>. */
    public TemplateRegistry() {
        registerDefaultTemplates();
    }

    /**
     * Creates a new <code>TemplateRegistry</code> as a clone of an existing
     * one.
     *
     * @param cOriginal the original <code>TemplateRegistry</code> to be
     * cloned
     */
    public TemplateRegistry( TemplateRegistry cOriginal ) {
        m_cTemplates = (Map<String, CompileableFactory>)
                ((HashMap<String, CompileableFactory>)cOriginal.m_cTemplates).clone();
    }

    /**
     * Registers the templates that will be held in this template registry.
     * This method may be subclassed to provide alternate template sets.
     */
    protected void registerDefaultTemplates() {
        // Register the templates in the map
        m_cTemplates.put( TemplateFactory.INSTANCE.getName(), TemplateFactory.INSTANCE );
        m_cTemplates.put( UnparsedTemplateFactory.INSTANCE.getName(), UnparsedTemplateFactory.INSTANCE );
        m_cTemplates.put( BinaryDataFactory.INSTANCE.getName(), BinaryDataFactory.INSTANCE );
    }

    /**
     * Retrieve a template factory from the registry. The returned factory
     * can then be used to create a compiled template from a given source.
     *
     * @param aKey the type of template factory to retrieve
     */
    public CompileableFactory getFactory( String aKey ) {
        return m_cTemplates.get( aKey );
    }

    /**
     * Adds a template factory to the registry. The factory can then be
     * retrieved from the registry on demand to create new template instances.
     *
     * @param cFactory the template factory
     */
    public void addFactory( CompileableFactory cFactory ) {
        Map<String, CompileableFactory> cNewTemplates = (Map<String, CompileableFactory>)((HashMap<String, CompileableFactory>)m_cTemplates).clone();

        String aName = cFactory.getName();
        // For thread-safety, we clone the old map before adding the
        // new template.
        cNewTemplates.put( aName, cFactory );
        m_cTemplates = cNewTemplates;
    }

    /**
     * Clones the current registry, and returns the clone back to the caller.
     * The clone operation performed here is a shallow clone.
     *
     * @return a clone of the current <code>TemplateRegistry</code>
     */
    public Object clone() throws CloneNotSupportedException {
        return new TemplateRegistry(this);
    }

    /**
     * Returns a string representation of the object.
     *
     * @return a <code>String</code> representation of the object
     */
    public String toString() {
        StringBuilder buffer = new StringBuilder( 50 );

        buffer.append( "TemplateRegistry, " );
        buffer.append( m_cTemplates.size() );
        buffer.append( " registered template factories." );
        return buffer.toString();
    }
}
