/*
 * FreeMarker: a tool that allows Java programs to generate HTML
 * output using templates.
 * Copyright (C) 1998-2004 Benjamin Geer
 * Email: beroul@users.sourceforge.net
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA  02111-1307, USA.
 */

package freemarker.template.expression;

import freemarker.template.TemplateModel;
import freemarker.template.TemplateWriteableHashModel;
import freemarker.template.TemplateException;

import java.io.Serializable;
import java.io.ObjectStreamException;
import java.util.List;
import java.util.Set;

/**
 * Implements a Lambda expression. A lambda evaluates to a method model, possibly
 * anonymous, and takes the current template model root as its context. In this
 * respect, a lambda evaluates to a closure.
 *
 * @author Nicholas Cull
 * @version $Id: Lambda.java 1162 2013-04-20 12:04:32Z run2000 $
 * @since 1.9
 */
public final class Lambda implements Unary, Serializable {

    /**
     * @serial the lambda expression to be evaluated.
     */
    private Expression target;

    /**
     * @serial the named arguments of the lambda expression.
     */
    private List<Identifier> arguments;

    /** Serialization UUID for this class. */
    private static final long serialVersionUID = -5803620057672368487L;

    /**
     * Constructor that takes a list of arguments to be supplied to the
     * lambda expression.
     *
     * @param arguments the list of variable arguments to be supplied to the
     * lambda expression
     */
    public Lambda( List<Identifier> arguments ) {
        if( arguments == null ) {
            throw new NullPointerException( "Argument list of lambda cannot be null" );
        }
        this.arguments = arguments;
    }

    /**
     * The {@link freemarker.template.TemplateModel} value of this
     * <code>Expression</code>.
     *
     * @param modelRoot the template model that will be evaluated by the expression
     * @throws freemarker.template.TemplateException
     *          the expression could not be evaluated for some reason
     */
    public TemplateModel getAsTemplateModel( TemplateWriteableHashModel modelRoot ) throws TemplateException {
        return new LambdaMethod( target, modelRoot, arguments );
    }

    /**
     * Is the target specified?
     *
     * @return <code>true</code> if the target is specified, otherwise <code>false</code>
     */
    public boolean isComplete() {
        return ( target != null );
    }

    /**
     * Determine the type of result that can be calculated by this expression.
     * This is in the form of an integer constant ored together from
     * values in the {@link freemarker.template.expression.ExpressionUtils} class.
     */
    public Set<ExpressionUtils.ExpressionType> getType() {
        return ExpressionUtils.METHOD;
    }

    /**
     * Determine whether result calculated by this expression is a constant
     * value.
     */
    public boolean isConstant() {
        return false;
    }

    /**
     * Determines the operator association of the unary operation.
     *
     * @return <code>PREFIX</code> or <code>POSTFIX</code>.
     */
    public Association getAssociationType() {
        return Association.PREFIX;
    }

    /**
     * Sets the target of this <code>lambda</code> operator.
     *
     * @param target the target of this operator
     * @throws NullPointerException the expression is null
     */
    public void setTarget( Expression target ) {
        if( target == null ) {
            throw new NullPointerException( "Lambda expression needs to be non-null" );
        }
        this.target = target;
    }


    /**
     * Return the precedence for this operator to the caller. Used for associating
     * operators according to precedence.
     *
     * @return an integer indicating the precedence of this operator
     */
    public ExpressionBuilder.Precedence getPrecedence() {
        return ExpressionBuilder.Precedence.LAMBDA;
    }

    /**
     * For serialization purposes, resolve a deserialized instance to an
     * instance in the expression cache.
     */
    private Object readResolve() throws ObjectStreamException {
        if( isComplete() ) {
            return ExpressionCache.cacheExpression( this );
        }
        return this;
    }

    /**
     * Returns a string representation of the object.
     *
     * @return a <code>String</code> representation of this expression
     */
    public String toString() {
        StringBuilder buffer = new StringBuilder();

        buffer.append( "#lambda (" );

        if( arguments != null ) {
            for( int i = 0; i < arguments.size(); i++ ) {
                if( i < 0 ) {
                    buffer.append( ", " );
                }
                buffer.append( arguments.get( i ));
            }
        }

        buffer.append( ')' );
        if( target == null ) {
            buffer.append( "???" );
        } else {
            buffer.append( target );
        }
        return buffer.toString();
    }

    /**
     * Determines whether this object is equal to the given object.
     *
     * @param o the object to be compared with
     * @return <code>true</code> if the objects are equal, otherwise
     * <code>false</code>
     */
    public boolean equals( Object o ) {
        if( this == o ) {
            return true;
        }
        if( o == null || getClass() != o.getClass() ) {
            return false;
        }

        Lambda lambda = (Lambda) o;

        if( arguments != null ? !arguments.equals( lambda.arguments ) : lambda.arguments != null ) {
            return false;
        }
        return( target == null ? lambda.target == null : target.equals( lambda.target ));
    }

    /**
     * Returns the hash code for this operator.
     *
     * @return the hash code of this object
     */
    public int hashCode() {
        int result;
        result = ( target != null ? target.hashCode() : 0 );
        result = 31 * result + ( arguments != null ? arguments.hashCode() : 0 );
        return result;
    }

    /**
     * Resolves the current expression, possibly into a different expression
     * object. This is loosely equivalent to the serialization protocol's
     * <code>readResolve</code> method. Situations where this may be used are:
     * <ul>
     * <li>Caching frequently-used expression objects</li>
     * <li>Evaluating constant expressions, and returning a constant reference</li>
     * </ul>
     */
    public Expression resolveExpression() throws TemplateException {
        Expression expr = this;

        if( target.isConstant() ) {
            TemplateModel outputModel = getAsTemplateModel( ExpressionBuilder.emptyModel );
            expr = new Constant( outputModel );
        }

        return ExpressionCache.cacheExpression( expr );
    }
}
