/*
 * FreeMarker: a tool that allows Java programs to generate HTML
 * output using templates.
 * Copyright (C) 1998-2004 Benjamin Geer
 * Email: beroul@users.sourceforge.net
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA  02111-1307, USA.
 */

package freemarker.template.expression;

import freemarker.template.TemplateMethodModel2;
import freemarker.template.TemplateModelException;
import freemarker.template.TemplateModel;
import freemarker.template.TemplateWriteableHashModel;
import freemarker.template.LocalModelWrapper;
import freemarker.template.TemplateException;

import java.io.Serializable;
import java.util.List;
import java.util.Map;
import java.util.HashMap;

/**
 * Implements an instance of a lambda expression.
 *
 * @author Nicholas Cull
 * @version $Id: LambdaMethod.java 1162 2013-04-20 12:04:32Z run2000 $
 * @since 1.9
 */
public final class LambdaMethod implements TemplateMethodModel2, Serializable {
    /**
     * @serial the lambda expression to be evaluated.
     */
    private final Expression expression;
    /**
     * @serial the template model for resolving free variables
     */
    private final TemplateWriteableHashModel capturedModel;
    /**
     * @serial the named arguments of the lambda expression.
     */
    private final List<Identifier> paramNames;

    /** Serialization UUID for this class. */
    private static final long serialVersionUID = -509257609923523900L;

    /**
     * Construct a new Lambda method given an expression to evaluate,
     * a template model for context evaluation, and a list of parameters
     * to be resolved when this method is called.
     *
     * @param expression the expression to be evaluated
     * @param model the template model to be used as context for free variables
     * @param paramNames names of parameters to be passed into the method
     */
    public LambdaMethod( Expression expression, TemplateWriteableHashModel model,
            List<Identifier> paramNames ) {
        this.expression = expression;
        this.capturedModel = model;
        this.paramNames = paramNames;
    }

    /**
     * Is the object empty?
     *
     * @return <code>true</code> if this object is empty, otherwise <code>false</code>
     */
    public boolean isEmpty() throws TemplateModelException {
        return expression == null;
    }

    /**
     * Executes a lambda method call. Arguments are passed as a <code>List</code> of
     * {@link freemarker.template.TemplateModel} objects.
     *
     * @param arguments a <code>List</code> of <code>TemplateModel</code> objects
     *                  containing the values of the arguments passed to the method.
     * @return the <code>TemplateModel</code> produced by the method, or <code>null</code>.
     */
    public TemplateModel exec( List<TemplateModel> arguments ) throws TemplateModelException {
        Map<String, TemplateModel> localModel = new HashMap<String, TemplateModel>();
        int paramSize = paramNames.size();

        if( paramSize != arguments.size() ) {
            throw new TemplateModelException( "Number of parameters to a lambda must be equal" );
        }

        for( int i = 0; i < paramSize; i++ ) {
            Identifier identifier = paramNames.get( i );
            localModel.put( identifier.getName(), arguments.get( i ));
        }
        TemplateWriteableHashModel thisModel = new LocalModelWrapper( capturedModel, localModel );

        try {
            return expression.getAsTemplateModel( thisModel );
        } catch( TemplateException e ) {
            throw new TemplateModelException( e.getMessage(), e );
        }
    }
}
