/*
 * FreeMarker: a tool that allows Java programs to generate HTML
 * output using templates.
 * Copyright (C) 1998-2004 Benjamin Geer
 * Email: beroul@users.sourceforge.net
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA  02111-1307, USA.
 */
package freemarker.template.expression;

import freemarker.template.TemplateModel;
import freemarker.template.TemplateWriteableHashModel;
import freemarker.template.TemplateException;

import java.util.Set;
import java.io.Serializable;
import java.io.ObjectStreamException;

/**
 * An expression that represents a conditional ternary operator. Most useful
 * in conjunction with the lambda operator.
 *
 * @author Nicholas Cull
 * @version $Id: Ternary.java 1153 2013-04-15 10:59:37Z run2000 $
 * @since 1.9
 */
public final class Ternary implements Operator, Serializable {
    /**
     * @serial the test expression to determine which result to evaluate
     */
    private Expression test;
    /**
     * @serial the expression to be evaluated if the test result is true
     */
    private Expression ifResult;
    /**
     * @serial the expression to be evaluated if the test result is false
     */
    private Expression elseResult;

    /** Serialization UUID for this class. */
    private static final long serialVersionUID = -7566309842912589863L;

    /**
     * Create a new Ternary expression.
     */
    public Ternary() {
    }

    /**
     * Sets the test expression for this ternary operator.
     *
     * @param expr the expression to be tested to determine which result
     * should be evaluated
     */
    public void setTest( Expression expr ) {
        this.test = expr;
    }

    /**
     * Sets the expression to be evaluated if the test evaluates to true.
     *
     * @param expr the expression to be evaluated if the test expression
     * evaluates true
     */
    public void setIfExpression( Expression expr ) {
        this.ifResult = expr;
    }

    /**
     * Sets the expression to be evaluates if the test evaluates to false.
     *
     * @param expr the expressoin to be evaluated if the test expression
     * evaluates false
     */
    public void setElseExpression( Expression expr ) {
        this.elseResult = expr;
    }

    /**
     * Performs ternary evaluation.
     *
     * @param modelRoot the template model that will be evaluated by the expression
     * @return the result of the ternary evaluation
     * @throws freemarker.template.TemplateException the expression could not be evaluated for some reason
     */
    public TemplateModel getAsTemplateModel( TemplateWriteableHashModel modelRoot)
            throws TemplateException {
        TemplateModel testModel = test.getAsTemplateModel( modelRoot );

        if( ExpressionUtils.isTrue( testModel )) {
            return ifResult.getAsTemplateModel( modelRoot );
        } else {
            return elseResult.getAsTemplateModel( modelRoot );
        }
    }

    /**
     * Determine the type of result that can be calculated by this expression.
     * This is in the form of an integer constant ored together from
     * values in the {@link freemarker.template.expression.ExpressionUtils} class.
     */
    public Set<ExpressionUtils.ExpressionType> getType() {
        return ExpressionUtils.VARIABLE;
    }

    /**
     * Is the <code>Expression</code> complete?
     *
     * @return <code>true</code> if this <code>Expression</code> is complete,
     *         otherwise <code>false</code>
     */
    public boolean isComplete() {
        return (test != null) && (ifResult != null) && (elseResult != null);
    }

    /**
     * Determine whether result calculated by this expression is a constant
     * value.
     */
    public boolean isConstant() {
        return false;
    }

    /**
     * Return the precedence for this operator to the caller. Used for associating
     * operators according to precedence.
     *
     * @return an integer indicating the precedence of this operator
     */
    public ExpressionBuilder.Precedence getPrecedence() {
        return ExpressionBuilder.Precedence.TERNARY;
    }

    /**
     * Resolves the current expression, possibly into a different expression
     * object. Situations where this may be used are:
     * <ul>
     * <li>Caching frequently-used expression objects</li>
     * <li>Evaluating constant expressions, and returning a constant reference</li>
     * </ul>
     */
    public Expression resolveExpression() throws TemplateException {
        Expression expr = this;

        if( test.isConstant() ) {
            TemplateModel testModel = test.getAsTemplateModel( ExpressionBuilder.emptyModel );
            if( ExpressionUtils.isTrue( testModel )) {
                expr = ifResult;
            } else {
                expr = elseResult;
            }
        }

        return ExpressionCache.cacheExpression( expr );
    }

    /**
     * For serialization purposes, resolve a deserialized instance to an
     * instance in the expression cache.
     */
    private Object readResolve() throws ObjectStreamException {
        if( isComplete() ) {
            return ExpressionCache.cacheExpression( this );
        }
        return this;
    }

    /**
     * Returns a string representation of the object.
     *
     * @return a <code>String</code> representation of this expression
     */
    public String toString() {
        StringBuilder buffer = new StringBuilder();

        buffer.append( test );
        buffer.append( " ? ");
        buffer.append( ifResult );
        buffer.append(" : ");
        buffer.append( elseResult );

        return buffer.toString();
    }

    /**
     * Determines whether this object is equal to the given object.
     *
     * @param o the object to be compared with
     * @return <code>true</code> if the objects are equal, otherwise
     * <code>false</code>
     */
    public boolean equals( Object o ) {
        if( this == o ) {
            return true;
        }
        if( o == null || getClass() != o.getClass() ) {
            return false;
        }

        Ternary ternary = (Ternary) o;

        if( elseResult != null ? !elseResult.equals( ternary.elseResult ) : ternary.elseResult != null ) {
            return false;
        }
        if( ifResult != null ? !ifResult.equals( ternary.ifResult ) : ternary.ifResult != null ) {
            return false;
        }
        return( test == null ? ternary.test == null : test.equals( ternary.test ));
    }

    /**
     * Returns the hash code for this operator.
     *
     * @return the hash code of this object
     */
    public int hashCode() {
        int result;
        result = ( test != null ? test.hashCode() : 0 );
        result = 31 * result + ( ifResult != null ? ifResult.hashCode() : 0 );
        result = 31 * result + ( elseResult != null ? elseResult.hashCode() : 0 );
        return result;
    }
}