/*
 * FreeMarker: a tool that allows Java programs to generate HTML
 * output using templates.
 * Copyright (C) 1998-2004 Benjamin Geer
 * Email: beroul@users.sourceforge.net
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA  02111-1307, USA.
 */

package freemarker.template.instruction;

import freemarker.template.*;
import freemarker.template.expression.Identifier;
import freemarker.template.compiler.TemplateBuilder;
import freemarker.template.compiler.ParseException;

import java.io.*;
import java.util.*;

/**
 * <p>An instruction representing a function definition. Function calls are
 * represented by the {@link CallInstruction} class.</p>
 *
 * <p>Unlike other instructions, <code>FunctionInstruction</code> is not
 * added directly to the {@link freemarker.template.TemplateProcessor} chain.
 * Instead, instances are added directly to the template using the
 * {@link freemarker.template.FunctionTemplateProcessor#addFunction} method.
 * At run time, these functions can be added to the data model for speedy
 * lookup.</p>
 *
 * <p>At compile time, instances of <code>FunctionInstruction</code> are
 * replaced by {@link NOOPInstruction}, so that the
 * {@link freemarker.template.TemplateProcessor} tree doesn't have to deal
 * with null values.</p>
 *
 * @version $Id: FunctionInstruction.java 1162 2013-04-20 12:04:32Z run2000 $
 */
public final class FunctionInstruction extends GenericStartInstruction
        implements FunctionTemplateProcessor, Serializable {

    private Identifier name;
    private List<Identifier> argumentNames;

    /** A mapping of inner function models within this function. */
    private Map<String, TemplateFunctionModel> functions;

    /** Serialization UUID for this class. */
    private static final long serialVersionUID = 8648224287266752240L;

    /**
     * Serialized form is an Identifier containing the function name,
     * an array of zero or more Identifiers containing the argument list,
     * an array of function names and inner function objects.
     * This is primarily for type correctness and avoids having to serialize a
     * List object.
     *
     * @serialField name Identifier the name of the function
     * @serialField arguments Identifier[] a list of arguments to be supplied
     * to the function
     * @serialField functionNames String[] an array of function names
     * @serialField functions TemplateFunctionModel[] an array of function
     * models associated with the names
     */
    private static final ObjectStreamField[] serialPersistentFields
        = { new ObjectStreamField("name", Identifier.class),
            new ObjectStreamField("arguments", Identifier[].class),
            new ObjectStreamField("functionNames", String[].class),
            new ObjectStreamField("functions", TemplateFunctionModel[].class) };

    /**
     * Constructor for a function definition.
     *
     * @param name the name of the function
     * @param arguments a List of Identifiers containing the names of each
     * of the arguments
     * @throws NullPointerException the name or argument list is null
     */
    public FunctionInstruction(Identifier name, List<Identifier> arguments) {
        if( name == null ) {
            throw new NullPointerException( "Function name cannot be null" );
        }
        if( arguments == null ) {
            throw new NullPointerException( "Argument list of function cannot be null" );
        }
        this.name = name;
        this.argumentNames = arguments;
    }

    /**
     * Retrieve the function name.
     *
     * @return the name of the function as a String
     */
    public String getName() {
        return name.getName();
    }

    /**
     * Retrieve a list of argument names.
     *
     * @return a List of Identifiers representing the argument names
     */
    public List<Identifier> getArgumentNames() {
        return argumentNames;
    }

    /**
     * Is this the right kind of instruction for the given {@link EndInstruction}?
     *
     * @param endInstruction the end instruction we're testing
     * @return <code>true</code> if the <code>EndInstruction</code> is a
     * function end instruction, otherwise <code>false</code>
     */
    public boolean testEndInstruction(Instruction endInstruction) {
        return (endInstruction.getEndType() == EndType.FUNCTION_END);
    }

    /**
     * Call the {@link freemarker.template.compiler.TemplateBuilder} with
     * this function instruction.
     *
     * @param builder the <code>TemplateBuilder</code> to be called back
     */
    public TemplateProcessor callBuilder(TemplateBuilder builder)
            throws ParseException {
        return builder.buildStatement(this);
    }

    /**
     * Retrieves an inner function from this function.  Called by
     * <code>CallInstruction</code>s at run-time.
     *
     * @param name the name of the function to be retrieved
     */
    public TemplateFunctionModel getFunction(String name) {
        if(functions ==  null) {
            return null;
        }
        return functions.get(name);
    }

    /**
     * Retrieve a <code>Set</code> of inner function names for this function.
     *
     * @return a <code>Set</code> of inner function names (<code>String</code> objects)
     * that have been defined for this template.
     */
    public Set<String> getFunctionNames() {
        if(functions == null) {
            return Collections.emptySet();
        }
        return Collections.unmodifiableSet( functions.keySet() );
    }

    /**
     * Adds an inner function to this function. Called by the
     * <code>TemplateBuilder</code> at compile-time.
     *
     * @param name the name of the function to be stored
     * @param function the inner function to be stored by this function
     * @throws IllegalStateException the method has been called after the
     * template has been compiled
     */
    public void addFunction(String name, TemplateFunctionModel function)
            throws IllegalStateException {

        if (functions == null) {
            functions = new HashMap<String, TemplateFunctionModel>();
        }
        functions.put(name, function);
    }

    /**
     * Processes the contents of this <code>&lt;function ... &gt;</code> and
     * outputs the resulting text to a <code>Writer</code>.
     *
     * @param p template processing parameters
     */
    public ExitStatus process(TemplateProcessorParameters p) throws IOException {
        return body.process(p);
    }

    /**
     * Returns a string representation of the object.
     *
     * @return a <code>String</code> representing this instruction subtree
     */
    public String toString() {
        StringBuilder buffer = new StringBuilder();
        buffer.append( "function " );
        buffer.append( name );
        buffer.append( ' ' );
        buffer.append( argumentNames );
        buffer.append( ' ' );
        buffer.append( body );
        return buffer.toString();
    }

    /**
     * For serialization, write this object as an Identifier containing the
     * function name, an array of Identifiers containing the argument list,
     * and a boolean indicating the scope of any variables.
     *
     * @param stream the output stream to write this object to
     */
    private void writeObject(ObjectOutputStream stream) throws IOException {
        ObjectOutputStream.PutField fields = stream.putFields();
        Identifier[] identifierArray = new Identifier[argumentNames.size()];

        argumentNames.toArray(identifierArray);

        int size = (functions == null) ? 0 : functions.size();
        TemplateFunctionModel[] functionArray = new TemplateFunctionModel[size];
        String[] nameArray = new String[size];

        if( functions != null ) {
            int i = 0;

            for (Map.Entry<String, TemplateFunctionModel> entry : functions.entrySet()) {
                nameArray[i] = entry.getKey();
                functionArray[i] = entry.getValue();
                i++;
            }
        }

        fields.put("name", name);
        fields.put("arguments", identifierArray);
        fields.put("functionNames", nameArray);
        fields.put("functions", functionArray);
        stream.writeFields();
    }

    /**
     * For serialization, read this object as an Identifier containing the
     * function name, an array of Identifiers containing the argument list,
     * and a boolean indicating the scope of any variables.
     * Check whether either the name or the argument list is null.
     *
     * @param stream the input stream to read serialized objects from
     */
    private void readObject(ObjectInputStream stream) throws IOException,
            ClassNotFoundException {

        ObjectInputStream.GetField fields = stream.readFields();

        name = (Identifier)fields.get("name", null);
        if( name == null ) {
            throw new InvalidObjectException( "Cannot create a FunctionInstruction with a null name" );
        }

        Identifier[] argumentArray = (Identifier[])fields.get("arguments", null);
        if( argumentArray == null ) {
            throw new InvalidObjectException( "Cannot create a FunctionInstruction with a null argument list" );
        }
        argumentNames = new ArrayList<Identifier>( Arrays.asList( argumentArray ));

        String[] functionNames = (String[])fields.get("functionNames", null);
        TemplateFunctionModel[] functionValues = (TemplateFunctionModel[])fields.get("functions", null);

        if(functionValues == null) {
            throw new InvalidObjectException( "Cannot create a FunctionInstruction with a null function list" );
        }
        functions = new HashMap<String, TemplateFunctionModel>();

        int size = functionValues.length;
        for(int i=0; i<size; i++) {
            String name = functionNames[i];
            TemplateFunctionModel functionValue = functionValues[i];
            functions.put( name, functionValue );
        }
    }
}
