/*
 * FreeMarker: a tool that allows Java programs to generate HTML
 * output using templates.
 * Copyright (C) 1998-2013 Benjamin Geer
 * Email: beroul@users.sourceforge.net
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA  02111-1307, USA.
 */

package freemarker.template.instruction;

import freemarker.template.*;
import freemarker.template.compiler.ParseException;
import freemarker.template.compiler.TemplateBuilder;
import freemarker.template.expression.Identifier;

import java.io.*;
import java.util.*;

/**
 * An instruction representing a local instruction.
 *
 * @author Nicholas Cull
 * @version $Id: LocalInstruction.java 1162 2013-04-20 12:04:32Z run2000 $
 * @since 1.9
 */
public final class LocalInstruction extends GenericStartInstruction
        implements FunctionTemplateProcessor, Serializable {

    /**
     * The variable names that will be local to this instruction.
     * If empty or null, all variables will become local.
     */
    private List<Identifier> localVariables;

    /** A mapping of inner function models within this local scope. */
    private Map<String,TemplateFunctionModel> functions;

    /** Serialization UUID for this class. */
    private static final long serialVersionUID = 160187086149474834L;

    /**
     * Serialized form is an array of zero or more Identifiers containing the
     * local variables, an array of function names and inner function objects.
     *
     * @serialField localVariables Identifier[] a list of local variables
     */
    private static final ObjectStreamField[] serialPersistentFields
        = { new ObjectStreamField("localVariables", Identifier[].class),
            new ObjectStreamField("functionNames", String[].class),
            new ObjectStreamField("functions", TemplateFunctionModel[].class) };

    /**
     * Constructor that takes a list of local variable identifiers to
     * indicate which variables should be interpreted as local to this scope.
     *
     * @param localVariables the list of local variables, or empty to indicate
     * all variables are local
     */
    public LocalInstruction(List<Identifier> localVariables) {
        this.localVariables = localVariables;
    }

    /**
     * Is this the right kind of instruction for the given {@link EndInstruction}?
     *
     * @param endInstruction the end instruction we're testing
     * @return <code>true</code> if the <code>EndInstruction</code> is a
     * local end instruction, otherwise <code>false</code>
     */
    public boolean testEndInstruction(Instruction endInstruction) {
        return (endInstruction.getEndType() == EndType.LOCAL_END);
    }

    /**
     * Evaluate this <code>&lt;local&gt;</code> instruction.
     *
     * @param p template processing parameters
     */
    public ExitStatus process(TemplateProcessorParameters p) throws IOException {
        TemplateWriteableHashModel localModelRoot;

        if(localVariables == null || localVariables.isEmpty()) {
            localModelRoot = new RootModelWrapper(p.getModelRoot());

            // Add the inner functions
            Set<String> functionNames = getFunctionNames();
            for (String functionName : functionNames) {
                try {
                    localModelRoot.put(functionName, getFunction(functionName));
                } catch (TemplateModelException e) {
                    // Do nothing and continue
                }
            }
        } else {
            Map<String,TemplateModel> localModel = new HashMap<String,TemplateModel>();

            for(int i = 0; i < localVariables.size(); i++) {
                Identifier identifier = localVariables.get(i);
                localModel.put(identifier.getName(), null);
            }

            // Add the inner functions
            Set<String> functionNames = getFunctionNames();
            for (String functionName : functionNames) {
                localModel.put(functionName, getFunction(functionName));
            }

            localModelRoot = new LocalModelWrapper(p.getModelRoot(), localModel);
        }

        // Process the local content
        TemplateProcessorParameters localParams = p.withModel(localModelRoot);
        ExitStatus result = body.process(localParams);

        return result;
    }

    /**
     * Call the {@link freemarker.template.compiler.TemplateBuilder} with
     * this local instruction.
     *
     * @param builder the <code>TemplateBuilder</code> to be called back
     */
    public TemplateProcessor callBuilder(TemplateBuilder builder)
            throws ParseException {
        return builder.buildStatement(this);
    }

    /**
     * Retrieves an inner function from this function.  Called by
     * <code>CallInstruction</code>s at run-time.
     *
     * @param name the name of the function to be retrieved
     */
    public TemplateFunctionModel getFunction(String name) {
        if(functions ==  null) {
            return null;
        }
        return functions.get(name);
    }

    /**
     * Retrieve a <code>Set</code> of inner function names for this function.
     *
     * @return a <code>Set</code> of inner function names (<code>String</code> objects)
     * that have been defined for this template.
     */
    public Set<String> getFunctionNames() {
        if(functions == null) {
            return Collections.emptySet();
        }
        return Collections.unmodifiableSet(functions.keySet());
    }

    /**
     * Adds an inner function to this function. Called by the
     * <code>TemplateBuilder</code> at compile-time.
     *
     * @param name the name of the function to be stored
     * @param function the inner function to be stored by this function
     * @throws IllegalStateException the method has been called after the
     * template has been compiled
     */
    public void addFunction(String name, TemplateFunctionModel function)
            throws IllegalStateException {

        if (functions == null) {
            functions = new HashMap<String,TemplateFunctionModel>();
        }
        functions.put(name, function);
    }

    /**
     * Returns a string representation of the object.
     *
     * @return a <code>String</code> representing this instruction subtree
     */
    public String toString() {
        return "local " + body;
    }

    /**
     * For serialization, write this object as array of Identifiers containing
     * the local variables, an array of function names and inner function objects.
     *
     * @param stream the output stream to write this object to
     */
    private void writeObject(ObjectOutputStream stream) throws IOException {
        ObjectOutputStream.PutField fields = stream.putFields();
        Identifier[] identifierArray = new Identifier[localVariables.size()];

        int size = (functions == null) ? 0 : functions.size();
        TemplateFunctionModel[] functionArray = new TemplateFunctionModel[size];
        String[] nameArray = new String[size];

        if( functions != null ) {
            Iterator iEntries = functions.entrySet().iterator();
            int i = 0;

            while(iEntries.hasNext()) {
                Map.Entry entry = (Map.Entry)iEntries.next();
                nameArray[i] = (String)entry.getKey();
                functionArray[i] = (TemplateFunctionModel)entry.getValue();
                i++;
            }
        }

        localVariables.toArray(identifierArray);

        fields.put("localVariables", identifierArray);
        fields.put("functionNames", nameArray);
        fields.put("functions", functionArray);
        stream.writeFields();
    }

    /**
     * For serialization, read this object as array of Identifiers containing
     * the local variables, an array of function names and inner function objects.
     *
     * @param stream the input stream to read serialized objects from
     */
    private void readObject(ObjectInputStream stream) throws IOException,
            ClassNotFoundException {

        ObjectInputStream.GetField fields = stream.readFields();
        Identifier[] argumentArray = (Identifier[])fields.get("localVariables", null);

        localVariables = new ArrayList<Identifier>( Arrays.asList( argumentArray ));

        String[] functionNames = (String[])fields.get("functionNames", null);
        TemplateFunctionModel[] functionValues = (TemplateFunctionModel[])fields.get("functions", null);

        if(functionValues == null) {
            throw new InvalidObjectException( "Cannot create a LocalInstruction with a null function list" );
        }
        functions = new HashMap<String, TemplateFunctionModel>();

        int size = functionValues.length;
        for(int i=0; i<size; i++) {
            String name = functionNames[i];
            TemplateFunctionModel functionValue = functionValues[i];
            functions.put( name, functionValue );
        }
    }
}
