/*
 * FreeMarker: a tool that allows Java programs to generate HTML
 * output using templates.
 * Copyright (C) 1998-2004 Benjamin Geer
 * Email: beroul@users.sourceforge.net
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA  02111-1307, USA.
 */

package freemarker.template.instruction;

import freemarker.template.*;
import freemarker.template.compiler.TemplateBuilder;
import freemarker.template.compiler.ParseException;
import freemarker.template.expression.Expression;
import freemarker.template.expression.ExpressionUtils;

import java.io.*;
import java.util.List;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Set;

/**
 * An instruction representing a switch-case structure.
 *
 * @version $Id: SwitchInstruction.java 1162 2013-04-20 12:04:32Z run2000 $
 */
public final class SwitchInstruction implements Instruction, TemplateProcessor,
        Serializable {

    private Expression testExpression;
    private List<DefaultCaseInstruction> caseInstructions;

    /** Serialization UUID for this class. */
    private static final long serialVersionUID = -1924465112006963670L;

    /**
     * Serialized form is a switch expression, an array of zero or more
     * case instructions, and an optional default instruction.
     *
     * @serialField switchExpression Expression the switch expression to
     * test case instructions against
     * @serialField caseInstructions CaseInstruction[] an array of case
     * instructions to test against the switch expression
     * @serialField defaultInstruction DefaultCaseInstruction an optional
     * default instruction
     */
    private static final ObjectStreamField[] serialPersistentFields
        = { new ObjectStreamField("switchExpression", Expression.class),
            new ObjectStreamField("caseInstructions", CaseInstruction[].class),
            new ObjectStreamField("defaultInstruction", DefaultCaseInstruction.class) };

    /**
     * Constructor that takes an expression to be evaluated when testing
     * each of the <code>case</code> statements.
     *
     * @param testExpression the expression to be tested.
     * @throws NullPointerException the test expression was null
     * @throws IllegalArgumentException the testExpression was not a scalar or number
     */
    public SwitchInstruction(Expression testExpression) {
        Set<ExpressionUtils.ExpressionType> expressionType = testExpression.getType();

        if(! ( expressionType.contains( ExpressionUtils.ExpressionType.STRING ) ||
                expressionType.contains( ExpressionUtils.ExpressionType.NUMBER ))) {
            throw new IllegalArgumentException( "Expression for switch instruction must be numeric or scalar" );
        }
        this.testExpression = testExpression;
        this.caseInstructions = new ArrayList<DefaultCaseInstruction>();
    }

    /**
     * Adds a case instruction to the switch.
     *
     * @param caseInstruction a <code>CaseInstruction</code> to be evaluated.
     */
    public void addCase(DefaultCaseInstruction caseInstruction) {
        caseInstructions.add(caseInstruction);
    }

    /**
     * Is this an end instruction?
     *
     * @return <code>false</code>, indicating that this is not an end instruction
     */
    public boolean isEndInstruction() {
        return false;
    }

    /**
     * Determine what type of end instruction this is, if any.
     *
     * @return <code>NONE</code>, indicating that this is not an end instruction
     */
    public EndType getEndType() {
        return Instruction.EndType.NONE;
    }

    /**
     * A {@link freemarker.template.compiler.TemplateBuilder} can call
     * this method to have an <code>Instruction</code> call it back to be built.
     * This implementation calls the
     * {@link freemarker.template.compiler.TemplateBuilder#buildStatement(SwitchInstruction)}
     * method, passing back a reference to itself. This approach is intended
     * to make type-checking of {@link Instruction} objects unnecessary.
     *
     * @param builder the builder to be called back by this method
     */
    public TemplateProcessor callBuilder(TemplateBuilder builder)
            throws ParseException {
        return builder.buildStatement(this);
    }

    /**
     * Process this <code>&lt;switch ... &gt;</code> instruction.
     *
     * @param p template processing parameters
     */
    public ExitStatus process(TemplateProcessorParameters p) throws IOException {
        TemplateModel switchModel;

        try {
            switchModel = testExpression.getAsTemplateModel(p.getModelRoot());
        } catch( TemplateException e ) {
            p.getEventHandler().fireExceptionThrown(this, e, p.getWriter(),
                    "freemarker.template.instruction.SwitchInstruction.process",
                    TemplateRuntimeHandler.Severity.WARNING);
            return TemplateProcessor.ExitStatus.OK;
        }

        boolean processedCase = false;
        for(DefaultCaseInstruction caseInstruction : caseInstructions) {
            boolean processCase = false;

            // Fall through if a previous case tested true.
            if ((processedCase) || (caseInstruction.isDefault())) {
                processCase = true;
            } else {
                // Otherwise, if this case isn't the default, test it.
                Expression caseExpression = caseInstruction.getExpression();
                try {
                    processCase = ExpressionUtils.isEqual( switchModel,
                            caseExpression.getAsTemplateModel(p.getModelRoot()));
                } catch (TemplateException e) {
                    p.getEventHandler().fireExceptionThrown(this, e, p.getWriter(),
                            "freemarker.template.instruction.SwitchInstruction.process",
                            TemplateRuntimeHandler.Severity.WARNING);
                }
            }

            if (processCase) {
                ExitStatus result;
                result = caseInstruction.process(p);
                processedCase = true;
                if (result == TemplateProcessor.ExitStatus.BREAK) {
                    // If a break is encountered, return immediately
                    return TemplateProcessor.ExitStatus.OK;
                } else if (result != TemplateProcessor.ExitStatus.OK) {
                    return result;
                }
            }
        }
        return TemplateProcessor.ExitStatus.OK;
    }

    /**
     * Returns a string representation of the object.
     *
     * @return a <code>String</code> representing this instruction subtree
     */
    public String toString() {
        StringBuilder buffer = new StringBuilder();
        buffer.append( "switch " );
        buffer.append( testExpression );

        if( caseInstructions != null ) {
            buffer.append( ' ' );
            buffer.append( caseInstructions );
        }
        return buffer.toString();
    }

    /**
     * For serialization, write this object as an Expression, an array of
     * case objects, and an optional default object.
     *
     * @param stream the output stream to write this object to
     */
    private void writeObject(ObjectOutputStream stream) throws IOException {
        ObjectOutputStream.PutField fields = stream.putFields();
        List<DefaultCaseInstruction> caseList = new ArrayList<DefaultCaseInstruction>( caseInstructions );
        DefaultCaseInstruction defaultInstruction = null;

        if( ! caseList.isEmpty() ) {
            int lastIndex = caseList.size() - 1;
            defaultInstruction = caseList.get( lastIndex );
            if( defaultInstruction.isDefault() ) {
                caseList.remove( lastIndex );
            } else {
                defaultInstruction = null;
            }
        }

        CaseInstruction[] cases = new CaseInstruction[caseList.size()];
        caseList.toArray(cases);

        fields.put("switchExpression", testExpression);
        fields.put("caseInstructions", cases);
        fields.put("defaultInstruction", defaultInstruction);
        stream.writeFields();
    }

    /**
     * For serialization, read this object as an Expression, an array of
     * case objects, and an optional default object.
     *
     * @param stream the input stream to read serialized objects from
     */
    private void readObject(ObjectInputStream stream)
            throws IOException, ClassNotFoundException {

        ObjectInputStream.GetField fields = stream.readFields();
        CaseInstruction[] cases;
        DefaultCaseInstruction defaultInstruction;

        testExpression = (Expression)fields.get("switchExpression", null);
        cases = (CaseInstruction[])fields.get("caseInstructions", null);
        defaultInstruction = (DefaultCaseInstruction)fields.get("defaultInstruction", null);

        if( testExpression == null ) {
            throw new InvalidObjectException( "Cannot create a SwitchInstruction with a null test" );
        }
        if( cases == null ) {
            caseInstructions = new ArrayList<DefaultCaseInstruction>( 1 );
        } else {
            caseInstructions = new ArrayList<DefaultCaseInstruction>( Arrays.asList( cases ));
        }
        if( defaultInstruction != null ) {
            caseInstructions.add( defaultInstruction );
        }
    }
}
