/*
 * FreeMarker: a tool that allows Java programs to generate HTML
 * output using templates.
 * Copyright (C) 1998-2004 Benjamin Geer
 * Email: beroul@users.sourceforge.net
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA  02111-1307, USA.
 */

package freemarker.template.instruction;

import freemarker.template.*;
import freemarker.template.expression.Variable;

import java.io.*;

/**
 * An instruction that processes a TemplateTransformModel.
 *
 * @version $Id: TransformInstruction.java 1162 2013-04-20 12:04:32Z run2000 $
 */
public final class TransformInstruction extends GenericStartInstruction
        implements Serializable {

    /** @serial the name of the variable containing the transformation */
    private final Variable transformVariable;

    /** Serialization UUID for this class. */
    private static final long serialVersionUID = 6158167624358601586L;

    /**
     * Creates new <code>TransformInstruction</code>, with a given
     * transformation variable.
     *
     * @param variable the variable representing the transformation to use
     * @throws NullPointerException the variable was null
     */
    public TransformInstruction( Variable variable ) {
        if( variable == null ) {
            throw new NullPointerException( "Variable for transform instruction cannot be null" );
        }
        transformVariable = variable;
    }

    /**
     * <p>Performs a <code>&lt;transform ... &gt;</code> operation on any
     * child instructions. In the event that a transformation cannot be
     * found or is empty, processing of the given block proceeds without
     * transformation.</p>
     *
     * <p><b>Note:</b><br />
     * This implementation is time-efficient rather than space efficient.
     * For a space-efficient tradeoff, use <code>PipeReader</code> and
     * <code>PipeWriter</code> instead. <code>PipeWriter</code> would need
     * to be in a separate thread.</p>
     *
     * @param p template processing parameters
     */
    public ExitStatus process(TemplateProcessorParameters p) throws IOException {

        // If the body is null, skip all processing here
        if (body == null) {
            return TemplateProcessor.ExitStatus.OK;
        }

        String transformName;
        try {
            transformName = transformVariable.getName(p.getModelRoot());
        } catch (TemplateException te) {
            p.getEventHandler().fireExceptionThrown(this, new TemplateException(
                    "Couldn't resolve name of transform variable", te),
                    p.getWriter(),
                    "freemarker.template.instruction.TransformInstruction.process",
                    TemplateRuntimeHandler.Severity.WARNING);
            return TemplateProcessor.ExitStatus.OK;
        }

        TemplateModel baseModel;
        try {
            baseModel = transformVariable.getAsTemplateModel(p.getModelRoot());
        } catch (TemplateException te) {
            p.getEventHandler().fireExceptionThrown(this, new TemplateException(
                    "Couldn't get referent of " + transformName, te),
                    p.getWriter(),
                    "freemarker.template.instruction.TransformInstruction.process",
                    TemplateRuntimeHandler.Severity.WARNING);
            return TemplateProcessor.ExitStatus.OK;
        }

        // If we can't find a transformer, process directly
        try {
            if ((baseModel == null) || (baseModel.isEmpty())) {
                body.process(p);
                return TemplateProcessor.ExitStatus.OK;
            }
        } catch (TemplateModelException e) {
            p.getEventHandler().fireExceptionThrown(this, new TemplateException(
                    "Couldn't perform transformation " +
                            transformName, e),
                    p.getWriter(),
                    "freemarker.template.instruction.TransformInstruction.process",
                    TemplateRuntimeHandler.Severity.WARNING);
        }

        // Process the body into a string buffer.
        StringWriter sw = new StringWriter();
        TemplateProcessorParameters localParams = p.withWriter(sw);

        // Get the content we have so far
        ExitStatus result;
        result = body.process(localParams);

        if (baseModel instanceof TemplateTransformModel2) {
            // Now read from the string buffer
            Reader in = new StringReader( sw.toString() );

            try {
                ((TemplateTransformModel2)baseModel).transform( in, p.getWriter());

            } catch (TemplateModelException e) {
                p.getEventHandler().fireExceptionThrown(this, new TemplateException(
                        "Couldn't perform transformation " +
                                transformName, e),
                        p.getWriter(),
                        "freemarker.template.instruction.TransformInstruction.process",
                        TemplateRuntimeHandler.Severity.WARNING);
            }
        } else if (baseModel instanceof TemplateTransformModel) {
            // Now read from the string buffer
            Reader inData = new StringReader( sw.toString() );
            PrintWriter outData = new PrintWriter(p.getWriter());

            try {
                ((TemplateTransformModel)baseModel).transform( inData, outData );

            } catch (TemplateModelException e) {
                p.getEventHandler().fireExceptionThrown(this, new TemplateException(
                        "Couldn't perform transformation " +
                                transformName, e),
                        p.getWriter(),
                        "freemarker.template.instruction.TransformInstruction.process",
                        TemplateRuntimeHandler.Severity.WARNING);
            }
        } else {
            p.getEventHandler().fireExceptionThrown(this, new TemplateException(
                    transformName + " is not a TemplateTransformModel or TemplateTransformModel2"
            ), p.getWriter(),
                    "freemarker.template.instruction.TransformInstruction.process",
                    TemplateRuntimeHandler.Severity.WARNING);
        }
        return result;
    }

    /**
     * Is this the right kind of instruction for the given {@link EndInstruction}?
     *
     * @param endInstruction the end instruction we're testing
     * @return <code>true</code> if the <code>EndInstruction</code> is a
     * transform end instruction, otherwise <code>false</code>
     */
    public boolean testEndInstruction(Instruction endInstruction) {
        return (endInstruction.getEndType() == EndType.TRANSFORM_END);
    }

    /**
     * Returns a string representation of the object.
     *
     * @return a <code>String</code> representing this instruction subtree
     */
    public String toString() {
        StringBuilder buffer = new StringBuilder();
        buffer.append( "transform " );
        buffer.append( transformVariable );
        buffer.append( ' ' );
        buffer.append( body );
        return buffer.toString();
    }

    /**
     * For serialization, read this object normally, then check whether the
     * transform variable is null.
     *
     * @param stream the input stream to read serialized objects from
     */
    private void readObject(ObjectInputStream stream) throws IOException,
            ClassNotFoundException {
        stream.defaultReadObject();
        if( transformVariable == null ) {
            throw new InvalidObjectException( "Cannot create a TransformInstruction with a null variable" );
        }
    }
}
