/*
 * FreeMarker: a tool that allows Java programs to generate HTML
 * output using templates.
 * Copyright (C) 1998-2004 Benjamin Geer
 * Email: beroul@users.sourceforge.net
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA  02111-1307, USA.
 */

package freemarker.template.instruction;

import freemarker.template.*;
import freemarker.template.expression.*;

import java.io.*;
import java.util.Set;

/**
 * An instruction that outputs the value of a {@link freemarker.template.expression.Variable}.
 *
 * @version $Id: VariableInstruction.java 1162 2013-04-20 12:04:32Z run2000 $
 */
public final class VariableInstruction extends EmptyInstruction
        implements Serializable {

    /**
     * @serial the expression to be evaluated and written to the output stream
     */
    private final Expression expression;

    /**
     * @serial whether the variable should be written using the default
     * escape transform, or left unescaped
     */
    private final boolean escaped;

    /** Serialization UUID for this class. */
    private static final long serialVersionUID = 7720592044181771388L;

    /**
     * Constructor that takes a {@link freemarker.template.expression.Expression}
     * to be output.
     *
     * @param expression the expression to be output
     * @throws NullPointerException the expression was null
     * @throws IllegalArgumentException the expression was not a scalar nor a number
     */
    public VariableInstruction(Expression expression, boolean escaped) {
        Set<ExpressionUtils.ExpressionType> expressionType = expression.getType();

        if(! (expressionType.contains(ExpressionUtils.ExpressionType.STRING) ||
                expressionType.contains(ExpressionUtils.ExpressionType.NUMBER))) {
            throw new IllegalArgumentException( "Variable instruction must be scalar or numeric" );
        }
        this.expression = expression;
        this.escaped = escaped;
    }

    /**
     * Process this <code>${ ... }</code> instruction.
     *
     * @param p template processing parameters
     */
    public ExitStatus process(TemplateProcessorParameters p) throws IOException {

        String variableValue;
        TemplateModel variableModel;

        try {
            variableModel = expression.getAsTemplateModel(p.getModelRoot());
            variableValue = ExpressionUtils.getAsStringOrEmpty(variableModel);
        } catch (TemplateException e) {
            String variableName;
            try {
                if(expression instanceof Variable) {
                    variableName = ((Variable)expression).getName(p.getModelRoot());
                } else {
                    variableName = "expression";
                }
            } catch (TemplateException te) {
                p.getEventHandler().fireExceptionThrown(this, new TemplateException(
                        "Couldn't get value of expression, " +
                                "or resolve its name", te),
                        p.getWriter(),
                        "freemarker.template.instruction.VariableInstruction.process",
                        TemplateRuntimeHandler.Severity.WARNING);
                return TemplateProcessor.ExitStatus.OK;
            }
            p.getEventHandler().fireExceptionThrown(this, new TemplateException(
                    "Couldn't get value of expression " +
                            variableName, e),
                    p.getWriter(),
                    "freemarker.template.instruction.VariableInstruction.process",
                    TemplateRuntimeHandler.Severity.WARNING);
            return TemplateProcessor.ExitStatus.OK;
        }

        TemplateModel escapeModel = p.getEscape();
        if(escaped && (escapeModel != null)) {
            if (escapeModel instanceof TemplateTransformModel2) {
                // Now read from the string buffer
                Reader in = new StringReader( variableValue );

                try {
                    ((TemplateTransformModel2)escapeModel).transform(in, p.getWriter());

                } catch (TemplateModelException e) {
                    p.getEventHandler().fireExceptionThrown(this, new TemplateException(
                            "Couldn't perform variable escaping", e),
                            p.getWriter(),
                            "freemarker.template.instruction.VariableInstruction.process",
                            TemplateRuntimeHandler.Severity.WARNING);
                }
            } else if (escapeModel instanceof TemplateTransformModel) {
                // Now read from the string buffer
                Reader inData = new StringReader( variableValue );
                PrintWriter outData = new PrintWriter(p.getWriter());

                try {
                    ((TemplateTransformModel)escapeModel).transform( inData, outData );

                } catch (TemplateModelException e) {
                    p.getEventHandler().fireExceptionThrown(this, new TemplateException(
                            "Couldn't perform variable escaping", e),
                            p.getWriter(),
                            "freemarker.template.instruction.VariableInstruction.process",
                            TemplateRuntimeHandler.Severity.WARNING);
                }
            } else {
                p.getEventHandler().fireExceptionThrown(this, new TemplateException(
                        "Variable escape is not a TemplateTransformModel or TemplateTransformModel2"
                ), p.getWriter(),
                        "freemarker.template.instruction.VariableInstruction.process",
                        TemplateRuntimeHandler.Severity.WARNING);
            }
        } else {
            p.getWriter().write(variableValue);
        }
        return TemplateProcessor.ExitStatus.OK;
    }

    /**
     * Returns a string representation of the object.
     *
     * @return a <code>String</code> representing this instruction subtree
     */
    public String toString() {
        StringBuilder buffer = new StringBuilder();
        buffer.append( "output " );
        if( expression != null ) {
            buffer.append( expression );
        }
        return buffer.toString();
    }
}
