/*
 * FreeMarker: a tool that allows Java programs to generate HTML
 * output using templates.
 * Copyright (C) 1998-2004 Benjamin Geer
 * Email: beroul@users.sourceforge.net
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA  02111-1307, USA.
 */

package freemarker.testcase;

import junit.framework.TestCase;
import junit.framework.Test;
import junit.framework.TestSuite;

import java.util.*;

import freemarker.template.*;
import freemarker.ext.beans2.*;
import freemarker.testcase.beans.*;

/**
 * A simple test case for testing most aspects of the
 * <code>freemarker.ext.beans</code> package.
 *
 * @author Nicholas Cull
 * @version $Id: TestBeanModels2.java 1072 2011-06-19 11:36:23Z run2000 $
 */
public final class TestBeanModels2 extends TestCase {

    /**
     * Constructor that passes the name of the test up to the JUnit
     * constructor.
     *
     * @param aTestname the name of the test
     */
    public TestBeanModels2( String aTestname ) {
        super( aTestname );
    }

    /**
     * Performs the bean model test of the List model.
     */
    public void testListModel() throws Exception {

        // Step 1: create an object model to wrap
        SimpleHash modelRoot = new SimpleHash();

        List<String> items = new ArrayList<String>();
        items.add( "test1" );
        items.add( "test2" );

        modelRoot.put( "beans", Wrapper.wrap( items ));

        // Step 2: assert that the model is correct
        TemplateHashModel hashModel = (TemplateHashModel)modelRoot.get( "beans" );
        //System.out.println( hashModel.getClass().toString() );
        assertTrue( "Expecting a ListWrapper", hashModel instanceof ListWrapper );

        List<TemplateModel> arguments = new ArrayList<TemplateModel>();

        TemplateMethodModel2 methodModel = (TemplateMethodModel2)hashModel.get( "size" );
        assertTrue( "Expecting a MethodModel", methodModel instanceof MethodWrapper );
        //System.out.println( methodModel.getClass().toString() );
        TemplateNumberModel numberModel = (TemplateNumberModel)methodModel.exec( arguments );

        assertEquals( numberModel.getAsNumber(), 2 );
        //System.out.println( numberModel.getAsNumber() );

        methodModel = (TemplateMethodModel2)hashModel.get( "get" );
        assertTrue( "Expecting a MethodModel", methodModel instanceof MethodWrapper );
        arguments.clear();
        arguments.add( new FastNumber( 0 ));
        TemplateScalarModel scalarModel = (TemplateScalarModel)methodModel.exec( arguments );
        assertEquals( "test1", scalarModel.getAsString() );
        //System.out.println( scalarModel.getAsString() );

        arguments.clear();
        arguments.add( new FastNumber( 1 ));

        scalarModel = (TemplateScalarModel)methodModel.exec( arguments );
        assertEquals( "test2", scalarModel.getAsString() );
        //System.out.println( scalarModel.getAsString() );

        // These two tests check the indexable representation of the list
        scalarModel = (TemplateScalarModel)((TemplateIndexedModel)hashModel).getAtIndex( 0 );
        assertEquals( "test1", scalarModel.getAsString() );
        //System.out.println( scalarModel.getAsString() );

        scalarModel = (TemplateScalarModel)((TemplateIndexedModel)hashModel).getAtIndex( 1 );
        assertEquals( "test2", scalarModel.getAsString() );
        //System.out.println( scalarModel.getAsString() );

        methodModel = (TemplateMethodModel2)hashModel.get( "iterator" );
        assertTrue( "Expecting a MethodModel", methodModel instanceof MethodWrapper );
        arguments.clear();
        TemplateListModel2 listModel = (TemplateListModel2)methodModel.exec( arguments );
        assertTrue( "Expecting an IteratorMethodWrapper", listModel instanceof IteratorMethodWrapper );
        //System.out.println( listModel.getClass().toString() );

        TemplateIteratorModel iteratorModel = listModel.templateIterator();
        //System.out.println( iteratorModel.getClass().toString() );

        TemplateModel item;
        int index = 0;

        while( iteratorModel.hasNext() ) {
            item = iteratorModel.next();
            index++;
            assertTrue( "Expecting a FastScalar", item instanceof FastScalar );
            //System.out.println( item.getClass().toString() );
            String value = ((TemplateScalarModel)item).getAsString();
            assertEquals( "Item should match index", "test" + index, value );

        }
        assertEquals( "Index should be 2, but is " + index, 2, index );
        listModel.releaseIterator( iteratorModel );

        // Again, directly from the ListWrapper this time
        listModel = (TemplateListModel2)hashModel;
        iteratorModel = listModel.templateIterator();
        assertTrue( "Iterator should be an instance of ListIteratorWrapper", iteratorModel instanceof ListIteratorWrapper );
        //System.out.println( iteratorModel.getClass().toString() );

        index = 0;

        while( iteratorModel.hasNext() ) {
            item = iteratorModel.next();
            index++;
            assertTrue( "Expecting a FastScalar", item instanceof FastScalar );
            //System.out.println( item.getClass().toString() );
            String value = ((TemplateScalarModel)item).getAsString();
            assertEquals( "Item should match index", "test" + index, value );

            ((TemplateWriteableIteratorModel)iteratorModel).set( new FastScalar( "new" + index ));

        }
        assertEquals( "Index should be 2, but is " + index, 2, index );
        listModel.releaseIterator( iteratorModel );

        Iterator iterator = items.iterator();
        index = 0;
        while( iterator.hasNext() ) {
            String value = (String)iterator.next();
            index++;
            //System.out.println( item.getClass().toString() );
            assertEquals( "Item should match index", "new" + index, value );
        }

        ((TemplateWriteableIndexedModel)hashModel).putAtIndex( 1, new FastScalar( "test3" ));
        assertEquals( "Item at index 1 should be test3", items.get( 1 ), "test3" );
    }

    /**
     * Performs the bean model test of the Collection model.
     */
    public void testCollectionModel() throws Exception {
        // Step 1: create an object model to wrap
        SimpleHash modelRoot = new SimpleHash();

        Collection<String> items = new TreeSet<String>();
        items.add( "test1" );
        items.add( "test2" );
        items.add( "test3" );

        modelRoot.put( "beans", Wrapper.wrap( items ));

        // Step 2: assert that the model is correct
        TemplateHashModel hashModel = (TemplateHashModel)modelRoot.get( "beans" );
        //System.out.println( hashModel.getClass().toString() );
        assertTrue( "Expecting a CollectionWrapper", hashModel instanceof CollectionWrapper );

        List<TemplateModel> arguments = new ArrayList<TemplateModel>();

        TemplateMethodModel2 methodModel = (TemplateMethodModel2)hashModel.get( "size" );
        assertTrue( "Expecting a MethodWrapper", methodModel instanceof MethodWrapper );
        //System.out.println( methodModel.getClass().toString() );
        TemplateNumberModel numberModel = (TemplateNumberModel)methodModel.exec( arguments );

        assertEquals( numberModel.getAsNumber(), 3 );

        methodModel = (TemplateMethodModel2)hashModel.get( "iterator" );
        assertTrue( "Expecting a MethodModel", methodModel instanceof MethodWrapper );
        arguments.clear();
        TemplateListModel2 listModel = (TemplateListModel2)methodModel.exec( arguments );

        assertTrue( "Expecting an IteratorMethodWrapper", listModel instanceof IteratorMethodWrapper );
        //System.out.println( listModel.getClass().toString() );

        TemplateIteratorModel iteratorModel = listModel.templateIterator();
        //System.out.println( iteratorModel.getClass().toString() );

        TemplateModel item;
        int index = 0;

        while( iteratorModel.hasNext() ) {
            item = iteratorModel.next();
            index++;
            assertTrue( "Expecting a FastScalar", item instanceof FastScalar );
            //System.out.println( item.getClass().toString() );
            String value = ((TemplateScalarModel)item).getAsString();
            assertEquals( "Item should match index", "test" + index, value );

        }
        assertEquals( "Index should be 3, but is " + index, 3, index );
        listModel.releaseIterator( iteratorModel );

        // Again, directly from the CollectionWrapper this time
        listModel = (TemplateListModel2)hashModel;
        iteratorModel = listModel.templateIterator();
        //System.out.println( iteratorModel.getClass().toString() );

        index = 0;

        while( iteratorModel.hasNext() ) {
            item = iteratorModel.next();
            index++;
            assertTrue( "Expecting a FastScalar", item instanceof FastScalar );
            //System.out.println( item.getClass().toString() );
            String value = ((TemplateScalarModel)item).getAsString();
            assertEquals( "Item should match index", "test" + index, value );

        }
        assertEquals( "Index should be 3, but is " + index, 3, index );
        listModel.releaseIterator( iteratorModel );

        methodModel = (TemplateMethodModel2)hashModel;
        arguments.clear();
        arguments.add( new FastScalar( "test4" ));
        arguments.add( FastBoolean.TRUE );
        methodModel.exec( arguments );
        assertTrue( "Set should contain test4", items.contains( "test4" ));

        arguments.clear();
        arguments.add( new FastScalar( "test2" ));
        arguments.add( FastBoolean.FALSE );
        methodModel.exec( arguments );
        assertTrue( "Set should not contain test2", ! items.contains( "test2" ));

        arguments.clear();
        arguments.add( new FastScalar( "test2" ));
        TemplateModel result = methodModel.exec( arguments );
        assertTrue( "Result should be false", result == FastBoolean.FALSE );
    }

    /**
     * Performs the bean model test of the Array model.
     */
    public void testArrayModel() throws Exception {
        // Step 1: create an object model to wrap
        SimpleHash modelRoot = new SimpleHash();

        String items[] = new String[ 3 ];
        items[ 0 ] = "test1";
        items[ 1 ] = "test2";
        items[ 2 ] = "test3";

        modelRoot.put( "beans", Wrapper.wrap( items ));

        // Step 2: assert that the model is correct
        TemplateHashModel hashModel = (TemplateHashModel)modelRoot.get( "beans" );
        //System.out.println( hashModel.getClass().toString() );
        assertTrue( "Expecting an ArrayWrapper", hashModel instanceof ArrayWrapper );

        TemplateNumberModel numberModel = (TemplateNumberModel)hashModel.get( "length" );
        assertTrue( "Expecting a FastNumber", numberModel instanceof FastNumber );
        //System.out.println( numberModel.getClass().toString() );

        assertEquals( numberModel.getAsNumber(), 3 );

        TemplateIteratorModel iteratorModel = ((TemplateListModel2)hashModel).templateIterator();
        assertTrue( "Expecting a FastIndexedWriteableIterator", iteratorModel instanceof FastIndexedWriteableIterator );
        //System.out.println( iteratorModel.getClass().toString() );

        TemplateModel item;
        int index = 0;

        while( iteratorModel.hasNext() ) {
            item = iteratorModel.next();
            index++;
            assertTrue( "Expecting a FastScalar", item instanceof FastScalar );
            //System.out.println( item.getClass().toString() );
            String value = ((TemplateScalarModel)item).getAsString();
            assertEquals( "Item should match index", "test" + index, value );

        }
        assertEquals( "Index should be 3, but is " + index, 3, index );
        ((TemplateListModel2)hashModel).releaseIterator( iteratorModel );


        item = ((TemplateIndexedModel)hashModel).getAtIndex( 2 );
        assertTrue( "Expecting a FastScalar", item instanceof FastScalar );
        String value = ((TemplateScalarModel)item).getAsString();
        assertEquals( "Item should match index", "test3", value );

        ((TemplateWriteableIndexedModel)hashModel).putAtIndex( 2, new FastScalar( "test4" ));
        assertEquals( "Item should match newly assigned value", "test4", items[ 2 ]);

        TemplateScalarModel scalar = (TemplateScalarModel)hashModel;
        assertTrue( "Scalar value should be non-null", scalar.getAsString() != null );
    }

    /**
     * Performs the bean model test of the Map model.
     */
    public void testMapModel() throws Exception {

        // Step 1: create an object model to wrap
        SimpleHash modelRoot = new SimpleHash();

        Map<String,Number> items = new TreeMap<String,Number>();
        items.put( "test1", new Long( 1 ));
        items.put( "test2", new Long( 2 ));
        items.put( "test3", new Long( 3 ));

        modelRoot.put( "beans", Wrapper.wrap( items ));

        // Step 2: assert that the model is correct
        TemplateHashModel hashModel = (TemplateHashModel)modelRoot.get( "beans" );
        //System.out.println( hashModel.getClass().toString() );
        assertTrue( "Expecting a BeanWrapper", hashModel instanceof BeanWrapper );

        List<TemplateModel> arguments = new ArrayList<TemplateModel>();

        TemplateMethodModel2 methodModel = (TemplateMethodModel2)hashModel.get( "size" );
        assertTrue( "Expecting a MethodWrapper", methodModel instanceof MethodWrapper );
        //System.out.println( methodModel.getClass().toString() );
        TemplateNumberModel numberModel = (TemplateNumberModel)methodModel.exec( arguments );

        assertEquals( numberModel.getAsNumber(), 3 );

        arguments.clear();
        methodModel = (TemplateMethodModel2)hashModel.get( "keySet" );
        assertTrue( "Expecting a MethodWrapper", methodModel instanceof MethodWrapper );
        //System.out.println( methodModel.getClass().toString() );
        TemplateHashModel hashModel2 = (TemplateHashModel)methodModel.exec( arguments );
        //System.out.println( listModel.getClass().toString() );

        methodModel = (TemplateMethodModel2)hashModel2.get( "iterator" );
        assertTrue( "Expecting a MethodModel", methodModel instanceof MethodWrapper );
        arguments.clear();
        TemplateListModel2 listModel = (TemplateListModel2)methodModel.exec( arguments );
        assertTrue( "Expecting an IteratorMethodWrapper", listModel instanceof IteratorMethodWrapper );

        TemplateIteratorModel iteratorModel = listModel.templateIterator();
        //System.out.println( iteratorModel.getClass().toString() );

        TemplateModel item;
        int index = 0;

        while( iteratorModel.hasNext() ) {
            item = iteratorModel.next();
            index++;
            assertTrue( "Expecting a FastScalar", item instanceof FastScalar );
            //System.out.println( item.getClass().toString() );
            String value = ((TemplateScalarModel)item).getAsString();
            assertEquals( "Item should match index", "test" + index, value );

        }
        assertEquals( "Index should be 3, but is " + index, 3, index );
        listModel.releaseIterator( iteratorModel );

        // Now try and get the values from the map
        iteratorModel = listModel.templateIterator();
        //System.out.println( iteratorModel.getClass().toString() );

        index = 0;
        methodModel = (TemplateMethodModel2)hashModel.get( "get" );

        while( iteratorModel.hasNext() ) {
            item = iteratorModel.next();
            index++;
            assertTrue( "Expecting a FastScalar", item instanceof FastScalar );
            arguments.clear();
            arguments.add( item );
            TemplateNumberModel result = (TemplateNumberModel)methodModel.exec( arguments );
            assertTrue( "Expecting a NumberWrapper", result instanceof NumberWrapper );
            //System.out.println( result.getClass().toString() );
            assertEquals( "Item should match index", index, result.getAsNumber() );

        }
        listModel.releaseIterator( iteratorModel );
        assertEquals( "Index should be 3, but is " + index, 3, index );

        // Second time, directly using the MapWrapper as a method
        index = 0;
        methodModel = (TemplateMethodModel2)hashModel;
        iteratorModel = listModel.templateIterator();

        while( iteratorModel.hasNext() ) {
            item = iteratorModel.next();
            index++;
            assertTrue( "Expecting a FastScalar", item instanceof FastScalar );
            arguments.clear();
            arguments.add( item );
            TemplateNumberModel result = (TemplateNumberModel)methodModel.exec( arguments );
            assertTrue( "Expecting a NumberWrapper", result instanceof NumberWrapper );
            //System.out.println( result.getClass().toString() );
            assertEquals( "Item should match index", index, result.getAsNumber() );

        }
        listModel.releaseIterator( iteratorModel );
        assertEquals( "Index should be 3, but is " + index, 3, index );


        // Again, directly from the MapWrapper this time
        listModel = (TemplateListModel2)hashModel;
        iteratorModel = listModel.templateIterator();

        index = 0;
        while( iteratorModel.hasNext() ) {
            item = iteratorModel.next();
            index++;
            assertTrue( "Expecting a BeanWrapper", item instanceof BeanWrapper );
            TemplateScalarModel scalar = (TemplateScalarModel)((TemplateHashModel)item).get( "key" );
            assertTrue( "Expecting a FastScalar", scalar instanceof FastScalar );
            TemplateNumberModel value = (TemplateNumberModel)((TemplateHashModel)item).get( "value" );
            assertTrue( "Expecting a NumberWrapper", value instanceof NumberWrapper );
            //System.out.println( result.getClass().toString() );
            assertEquals( "Item should match index", index, value.getAsNumber() );

        }

        listModel.releaseIterator(iteratorModel);
        arguments.clear();
        arguments.add( new FastScalar( "test2" ));
        arguments.add( new FastNumber( 5 ));

        methodModel.exec( arguments );
        assertEquals( "Item 2 should be an Integer value 5", new Integer( 5 ), items.get( "test2" ));

    }

    /**
     * Performs the bean model test of the Scalar model.
     */
    public void testScalarModel() throws Exception {
        SimpleBean bean = new SimpleBean();

        // Step 1: create an object model to wrap
        SimpleHash modelRoot = new SimpleHash();

        modelRoot.put( "bean", Wrapper.wrap( bean ));

        // Step 2: assert that the model is correct
        TemplateHashModel hashModel = (TemplateHashModel)modelRoot.get( "bean" );
        assertTrue( "Expecting an BeanWrapper", hashModel instanceof BeanWrapper );
        //System.out.println( hashModel.getClass().toString() );

        TemplateNumberModel numberModel = (TemplateNumberModel)hashModel.get( "intValue" );
        assertNotNull( numberModel );
        assertTrue( "Expecting a NumberWrapper", numberModel instanceof NumberWrapper );
        //System.out.println( scalarModel.getClass().toString() );
        assertEquals( numberModel.getAsNumber(), 5 );

        numberModel = (TemplateNumberModel)hashModel.get( "byteValue" );
        assertNotNull( numberModel );
        assertTrue( "Expecting a NumberWrapper", numberModel instanceof NumberWrapper );
        //System.out.println( scalarModel.getClass().toString() );
        assertEquals( numberModel.getAsNumber(), 127 );

        numberModel = (TemplateNumberModel)hashModel.get( "longValue" );
        assertNotNull( numberModel );
        assertTrue( "Expecting a NumberWrapper", numberModel instanceof NumberWrapper );
        //System.out.println( scalarModel.getClass().toString() );
        assertEquals( numberModel.getAsNumber(), -10 );

        numberModel = (TemplateNumberModel)hashModel.get( "shortValue" );
        assertNotNull( numberModel );
        assertTrue( "Expecting a NumberModel", numberModel instanceof NumberWrapper );
        //System.out.println( scalarModel.getClass().toString() );
        assertEquals( numberModel.getAsNumber(), 42 );

        TemplateScalarModel scalarModel = (TemplateScalarModel)hashModel.get( "message" );
        assertNotNull( scalarModel );
        assertTrue( "Expecting a FastScalar", scalarModel instanceof FastScalar );
        //System.out.println( scalarModel.getClass().toString() );
        assertEquals( scalarModel.getAsString(), "Hello, world!" );

        numberModel = (TemplateNumberModel)hashModel.get( "floatObject" );
        assertNotNull( numberModel );
        assertTrue( "Expecting a NumberWrapper", numberModel instanceof NumberWrapper );
        //System.out.println( scalarModel.getClass().toString() );
        assertEquals( numberModel.getAsNumber(), 3 );
        assertEquals( ((TemplateScalarModel)numberModel).getAsString(), "3.5" );

        numberModel = (TemplateNumberModel)hashModel.get( "doubleObject" );
        assertNotNull( numberModel );
        assertTrue( "Expecting a NumberModel", numberModel instanceof NumberWrapper );
        //System.out.println( scalarModel.getClass().toString() );
        assertEquals( numberModel.getAsNumber(), 3 );
        assertEquals( ((TemplateScalarModel)numberModel).getAsString(), "3.14" );

        TemplateMethodModel2 dummyModel = (TemplateMethodModel2) hashModel.get( "setShortValue" );
        FastNumber argument1 = new FastNumber( 7 );
        List<TemplateModel> arguments = new ArrayList<TemplateModel>();
        arguments.add( argument1 );
        dummyModel.exec( arguments );
        assertEquals( "Expecting short value assigned to 7", bean.getShortValue(), (short)7 );

        ((TemplateWriteableHashModel)hashModel).put( "shortValue", numberModel );
        assertEquals( "Expecting short value assigned to 3", bean.getShortValue(), (short)3 );

        ((TemplateWriteableHashModel)hashModel).put( "bigInt", numberModel );
        ((TemplateWriteableHashModel)hashModel).put( "bigDecimal", numberModel );

        TemplateNumberModel numberWrapper = new NumberWrapper( new Float( 5.7F ));
        ((TemplateWriteableHashModel)hashModel).put( "double", numberWrapper );
        assertEquals( "Expecting double value assigned to 5.7", new Double(
                new Float( 5.7F ).doubleValue() ), bean.getDoubleObject() );

        numberWrapper = new FastNumber( 9 );
        ((TemplateWriteableHashModel)hashModel).put( "float", numberWrapper );
        assertEquals( "Expecting float value assigned to 9", new Float( 9 ), bean.getFloatObject() );
    }

    /**
     * Performs the bean model test of container models.
     */
    public void testContainerModels() throws Exception {
        SimpleBean bean = new SimpleBean();

        // Step 1: create an object model to wrap
        SimpleHash modelRoot = new SimpleHash();

        modelRoot.put( "bean", Wrapper.wrap( bean ));

        // Step 2: assert that the model is correct
        TemplateHashModel hashModel = (TemplateHashModel)modelRoot.get( "bean" );
        assertTrue( "Expecting an BeanWrapper", hashModel instanceof BeanWrapper );
        //System.out.println( hashModel.getClass().toString() );

        List<TemplateModel> list1 = new ArrayList<TemplateModel>();
        list1.add( FastBoolean.TRUE );
        list1.add( FastBoolean.TRUE );
        list1.add( FastBoolean.FALSE );

        FastList<?> listModel = new FastList<TemplateModel>( list1 );

        ((TemplateWriteableHashModel)hashModel).put( "array1", listModel );

        list1.clear();
        list1.add( new FastNumber( 1 ));
        list1.add( new FastNumber( 2 ));
        list1.add( new FastNumber( 3 ));

        ((TemplateWriteableHashModel)hashModel).put( "array2", listModel );
        ((TemplateWriteableHashModel)hashModel).put( "array4", listModel );
        ((TemplateWriteableHashModel)hashModel).put( "array5", listModel );
        ((TemplateWriteableHashModel)hashModel).put( "linkedList", listModel );
        ((TemplateWriteableHashModel)hashModel).put( "vector", listModel );
        ((TemplateWriteableHashModel)hashModel).put( "list", listModel );

        list1.clear();
        list1.add( new FastScalar( "1" ));
        list1.add( new FastScalar( "2" ));
        list1.add( new FastScalar( "3" ));

        ((TemplateWriteableHashModel)hashModel).put( "array3", listModel );
        ((TemplateWriteableHashModel)hashModel).put( "vector", listModel );
        ((TemplateWriteableHashModel)hashModel).put( "list", listModel );
        ((TemplateWriteableHashModel)hashModel).put( "list2", listModel );
        ((TemplateWriteableHashModel)hashModel).put( "set", listModel );
        ((TemplateWriteableHashModel)hashModel).put( "sortedSet", listModel );

        list1.clear();
        List<TemplateModel> list2 = new ArrayList<TemplateModel>();
        list2.add( new FastScalar( "1" ));
        list2.add( new FastScalar( "2" ));
        list2.add( new FastScalar( "3" ));

        list1.add( Wrapper.wrap( list2 ));

        list2 = new ArrayList<TemplateModel>();
        list2.add( new FastNumber( 1 ));
        list2.add( new FastNumber( 2 ));
        list2.add( new FastNumber( 3 ));

        list1.add( Wrapper.wrap( list2 ));

        ((TemplateWriteableHashModel)hashModel).put( "array6", listModel );

        String[] args = new String[] { "list1", "list2", "list3" };
        TemplateModel hashModel2 = Wrapper.wrap( args );

        TemplateMethodModel2 method = (TemplateMethodModel2) hashModel.get( "checkList" );
        assertTrue( "Method checkList should be a MethodWrapper" , method instanceof MethodWrapper );

        List<TemplateModel> arguments = new ArrayList<TemplateModel>();
        arguments.add( hashModel2 );
        method.exec( arguments );
    }

    public void testIndexedProperties() throws Exception {
        IndexedPropertyBean bean = new IndexedPropertyBean();

        // Step 1: create an object model to wrap
        SimpleHash modelRoot = new SimpleHash();

        modelRoot.put( "bean", Wrapper.wrap( bean ));

        // Step 2: assert that the model is correct
        TemplateHashModel hashModel = (TemplateHashModel)modelRoot.get( "bean" );
        assertTrue( "Expecting an BeanWrapper", hashModel instanceof BeanWrapper );
        //System.out.println( hashModel.getClass().toString() );

        TemplateHashModel propertyModel = (TemplateHashModel)hashModel.get( "name" );
        assertNotNull( propertyModel );
        assertTrue( "Expecting an IndexedPropertyWrapper", propertyModel instanceof IndexedPropertyWrapper );
        //System.out.println( scalarModel.getClass().toString() );

        TemplateNumberModel length = (TemplateNumberModel) propertyModel.get( "length" );
        assertTrue( "Length result should be a FastNumber", length instanceof FastNumber );
        assertEquals( "Length should be 5", length.getAsNumber(), 5 );

        // Again, directly from the MapWrapper this time
        TemplateListModel2 listModel = (TemplateListModel2)propertyModel;
        TemplateIteratorModel iteratorModel = listModel.templateIterator();

        int index = 0;
        while( iteratorModel.hasNext() ) {
            TemplateModel item = iteratorModel.next();
            index++;
            assertTrue( "Expecting a FastScalar", item instanceof FastScalar );
            assertEquals( "Item should match index", "name" + index, ((TemplateScalarModel)item).getAsString() );
        }

        listModel.releaseIterator(iteratorModel);

        ((TemplateWriteableIndexedModel)propertyModel).putAtIndex( 2, new FastScalar( "name7" ));
        TemplateScalarModel item = (TemplateScalarModel) ((TemplateIndexedModel)propertyModel).getAtIndex( 2 );
        assertTrue( "Item should be a FastScalar", item instanceof FastScalar );
        assertEquals( "Item should contain value name7", item.getAsString(), "name7" );

        Object array = Unwrapper.defaultUnwrap( propertyModel );
        assertTrue( "Item should be an Array", array.getClass().isArray() );
        assertEquals( "Item should be an Array of Strings", array.getClass().getComponentType(), String.class );

        String[] result = (String[])array;
        assertEquals( "Third item should be name7", result[2], "name7" );
    }

    /**
     * Performs the bean model test of the Scalar model.
     */
    public void testStaticModel() throws Exception {
        // Step 1: create an object model to wrap
        SimpleHash modelRoot = new SimpleHash();

        modelRoot.put( "system", Wrapper.wrap( System.class ));
        modelRoot.put( "static", Wrapper.wrap( StaticClass.class ));

        // Step 2: assert that the model is correct
        TemplateHashModel hashModel = (TemplateHashModel)modelRoot.get( "system" );
        assertTrue( "Expecting a StaticClassWrapper", hashModel instanceof StaticClassWrapper );
        //System.out.println( hashModel.getClass().toString() );

        TemplateMethodModel2 method = (TemplateMethodModel2)hashModel.get( "currentTimeMillis" );
        assertTrue( "Expecting a StaticMethodWrapper", method instanceof StaticMethodWrapper );
        //System.out.println( method.getClass().toString() );

        List<TemplateModel> arguments = new ArrayList<TemplateModel>();

        TemplateModel output = method.exec( arguments );
        //System.out.println( output.getClass().toString() );
        assertTrue( "Expecting a NumberWrapper", output instanceof NumberWrapper );
        //System.out.println( output.getAsNumber() );

        arguments.clear();
        arguments.add( new FastScalar( "line.separator" ));
        method = (TemplateMethodModel2) hashModel.get( "getProperty" );
        assertTrue( "Expecting a StaticMethodWrapper", method instanceof StaticMethodWrapper );

        output = method.exec( arguments );
        assertTrue( "Expecting a FastScalar", output instanceof FastScalar );

        TemplateScalarModel model = (TemplateScalarModel) hashModel;
        assertEquals( "Expecting a java.lang.System class", "class java.lang.System", model.getAsString() );

        // Now test fields
        hashModel = (TemplateHashModel)modelRoot.get( "static" );
        assertTrue( "Expecting a StaticClassWrapper", hashModel instanceof StaticClassWrapper );

        output = hashModel.get( "number" );
        assertTrue( "Number should be a NumberWrapper", output instanceof NumberWrapper );
        assertEquals( "Number should be 5", 5, ((TemplateNumberModel)output).getAsNumber() );

        output = hashModel.get( "connected" );
        assertTrue( "Connected should be a FastBoolean", output instanceof FastBoolean );
        assertTrue( "Connected should be true", FastBoolean.getBoolean( output ));

        ((TemplateWriteableHashModel)hashModel).put( "message", new FastScalar( "Goodbye, cruel world!" ));
        output = hashModel.get( "message" );
        assertTrue( "Connected should be a FastScalar", output instanceof FastScalar );
        assertEquals( "Message should be goodbye...", "Goodbye, cruel world!", ((TemplateScalarModel)output).getAsString() );

    }

    public void testEnumModel() throws Exception {
        // Step 1: create an object model to wrap
        SimpleHash modelRoot = new SimpleHash();

        modelRoot.put( "enum", Wrapper.wrap( TestEnum.class ));

        // Step 2: assert that the model is correct
        TemplateHashModel hashModel = (TemplateHashModel)modelRoot.get( "enum" );
        assertTrue( "Expecting a StaticEnumWrapper", hashModel instanceof StaticEnumWrapper );
        //System.out.println( hashModel.getClass().toString() );

        // Directly from the StaticEnumWrapper
        TemplateListModel2 listModel = (TemplateListModel2)hashModel;
        TemplateIteratorModel iteratorModel = listModel.templateIterator();

        int index = 64;
        while( iteratorModel.hasNext() ) {
            TemplateModel item = iteratorModel.next();
            index++;
            String name = ((TemplateScalarModel)item).getAsString();
            char c = name.charAt(0);
            assertEquals( "Expecting an enumeration value", (int)c, index );
            //System.out.println( (int)c );

            TemplateScalarModel nameResult = (TemplateScalarModel)((TemplateHashModel)item).get("name");
            assertEquals( "Expecting enumeration name to equal toString", nameResult.getAsString(), name );
        }

        listModel.releaseIterator(iteratorModel);

        TemplateModel instance = hashModel.get("AARDVARK");
        assertEquals( "Expect enum of AARDVARK", ((TemplateScalarModel)instance).getAsString(), "AARDVARK" );
        TemplateMethodModel2 methodModel = (TemplateMethodModel2)((TemplateHashModel)instance).get( "ordinal" );
        List<TemplateModel> args = new ArrayList<TemplateModel>();
        TemplateModel result = methodModel.exec( args );
        assertEquals( "Expecting ordinal value of 0", ((TemplateNumberModel)result).getAsNumber(), 0 );

        instance = (TemplateScalarModel)hashModel.get("BUMBLEBEE");
        assertEquals( "Expect enum of BUMBLEBEE", ((TemplateScalarModel)instance).getAsString(), "BUMBLEBEE" );
        methodModel = (TemplateMethodModel2)((TemplateHashModel)instance).get( "ordinal" );
        result = methodModel.exec( args );
        assertEquals( "Expecting ordinal value of 1", ((TemplateNumberModel)result).getAsNumber(), 1 );

        instance = (TemplateScalarModel)hashModel.get("CAT");
        assertEquals( "Expect enum of CAT", ((TemplateScalarModel)instance).getAsString(), "CAT" );
        methodModel = (TemplateMethodModel2)((TemplateHashModel)instance).get( "ordinal" );
        result = methodModel.exec( args );
        assertEquals( "Expecting ordinal value of 2", ((TemplateNumberModel)result).getAsNumber(), 2 );

        instance = (TemplateScalarModel)hashModel.get("DOG");
        assertEquals( "Expect enum of DOG", ((TemplateScalarModel)instance).getAsString(), "DOG" );
        methodModel = (TemplateMethodModel2)((TemplateHashModel)instance).get( "ordinal" );
        result = methodModel.exec( args );
        assertEquals( "Expecting ordinal value of 3", ((TemplateNumberModel)result).getAsNumber(), 3 );

        instance = (TemplateScalarModel)hashModel.get("EAGLE");
        assertEquals( "Expect enum of EAGLE", ((TemplateScalarModel)instance).getAsString(), "EAGLE" );
        methodModel = (TemplateMethodModel2)((TemplateHashModel)instance).get( "ordinal" );
        result = methodModel.exec( args );
        assertEquals( "Expecting ordinal value of 4", ((TemplateNumberModel)result).getAsNumber(), 4 );

        // Directly from the dynamic EnumWrapper object this time
        listModel = (TemplateListModel2)instance;
        iteratorModel = listModel.templateIterator();

        index = 64;
        while( iteratorModel.hasNext() ) {
            TemplateModel item = iteratorModel.next();
            index++;
            String name = ((TemplateScalarModel)item).getAsString();
            char c = name.charAt(0);
            assertEquals( "Expecting an enumeration value", (int)c, index );
            //System.out.println( name );

            TemplateScalarModel nameResult = (TemplateScalarModel)((TemplateHashModel)item).get("name");
            assertEquals( "Expecting enumeration name to equal toString", nameResult.getAsString(), name );
        }

        listModel.releaseIterator(iteratorModel);
    }

    /**
     * Performs tests on coersion of Simple and Fast models into forms usable
     * by the <code>freemarker.ext.beans</code> package.
     */
    public void testMethodCalls() throws Exception {
        SimpleBean bean = new SimpleBean();

        // Step 1: create an object model to wrap
        SimpleHash modelRoot = new SimpleHash();

        modelRoot.put( "bean", Wrapper.wrap( bean ));

        // Step 2: assert that the model is correct
        TemplateHashModel hashModel = (TemplateHashModel)modelRoot.get( "bean" );
        assertTrue( "Expecting an BeanWrapper", hashModel instanceof BeanWrapper );
        //System.out.println( hashModel.getClass().toString() );

        TemplateMethodModel2 method = (TemplateMethodModel2)hashModel.get( "checkList" );
        assertTrue( "Expecting a MethodWrapper", method instanceof MethodWrapper );
        //System.out.println( method.getClass().toString() );

        // Step 3: call the method model
        SimpleList list = new SimpleList();
        list.add( "test1" );
        list.add( "test2" );
        list.add( "test3" );

        List<FastScalar> objectList = new ArrayList<FastScalar>();
        objectList.add( new FastScalar( "test1" ));
        objectList.add( new FastScalar( "test2" ));
        objectList.add( new FastScalar( "test3" ));

        FastList list2 = new FastList<FastScalar>( objectList );

        List<TemplateModel> argumentList = new ArrayList<TemplateModel>();
        argumentList.add( list );

        TemplateNumberModel result = (TemplateNumberModel)method.exec( argumentList );
        assertTrue( "Expecting a NumberWrapper", result instanceof NumberWrapper );
        //System.out.println( result.getClass().toString() );

        assertEquals( 3, result.getAsNumber() );

        // Again, with FastList
        argumentList.clear();
        argumentList.add( list2 );

        result = (TemplateNumberModel)method.exec( argumentList );
        assertTrue( "Expecting a NumberWrapper", result instanceof NumberWrapper );
        //System.out.println( result.getClass().toString() );

        assertEquals( 3, result.getAsNumber() );

        // Now with the hash models
        method = (TemplateMethodModel2)hashModel.get( "checkMap" );
        assertTrue( "Expecting a MethodWrapper", method instanceof MethodWrapper );
        //System.out.println( method.getClass().toString() );

        SimpleHash hash = new SimpleHash();
        hash.put( "test1", "test1" );
        hash.put( "test2", "test2" );
        hash.put( "test3", "test3" );

        Map<String,TemplateModel> objectMap = new HashMap<String,TemplateModel>();
        FastScalar fastScalar = new FastScalar( "test1" );
        objectMap.put( "test1", fastScalar );
        fastScalar = new FastScalar( "test2" );
        objectMap.put( "test2", fastScalar );
        fastScalar = new FastScalar( "test3" );
        objectMap.put( "test3", fastScalar );

        FastHash hash2 = new FastHash<TemplateModel>( objectMap );

        // Now call the method
        argumentList.clear();
        argumentList.add( hash );

        result = (TemplateNumberModel)method.exec( argumentList );
        assertTrue( "Expecting a NumberWrapper", result instanceof NumberWrapper );
        //System.out.println( result.getClass().toString() );

        assertEquals( 3, result.getAsNumber() );

        // Again, with FastHash
        argumentList.clear();
        argumentList.add( hash2 );

        result = (TemplateNumberModel)method.exec( argumentList );
        assertTrue( "Expecting a NumberWrapper", result instanceof NumberWrapper );
        //System.out.println( result.getClass().toString() );

        assertEquals( 3, result.getAsNumber() );
    }

    /**
     * Return a suite of all the tests in this class by reflection.
     *
     * @return a <code>TestSuite</code> of all the methods starting with the
     * word "test"
     */
    public static Test suite() {
        return new TestSuite( TestBeanModels2.class );
    }

    /**
     * Bootstrap for the self-test code.
     */
    public static void main( String[] argc ) throws Exception {

        for( int i = 0 ; i < 100; i++ ) {
            junit.textui.TestRunner.run( suite() );
        }
    }
}
