/*
 * FreeMarker: a tool that allows Java programs to generate HTML
 * output using templates.
 * Copyright (C) 1998-2004 Benjamin Geer
 * Email: beroul@users.sourceforge.net
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA  02111-1307, USA.
 */

package freemarker.testcase;

import freemarker.template.*;
import java.io.IOException;
import java.io.InputStream;
import java.util.Properties;

/**
 * Simple class (JUnit agnostic) to test cache implementation. This
 * implementation creates multiple threads, each with a different template
 * to retrieve, and counts how many templates were retrieved in a given
 * period. In addition, it also collects statistics of the various events
 * that were fired during the test, and outputs them at the end.
 *
 * @author  <a href="mailto:run2000@users.sourceforge.net">Nicholas Cull</a>
 * @version $Id: TestCacheThreaded.java 1101 2013-04-01 04:17:32Z run2000 $
 * @see CacheThread
 * @see CacheThread2
 * @see CacheThread3
 * @see CacheThread4
 * @see CacheThread5
 */
public final class TestCacheThreaded implements CacheListener {
    private int nRemoved = 0;
    private int nFailed = 0;
    private int nUnavailable = 0;
    private int nUpdated = 0;

    /** Creates new TestCacheThreaded */
    public TestCacheThreaded() {
    }

    /**
     * Performs the test. The test results are sent to <code>System.out</code>.
     */
    public void test() throws Exception {
        FileTemplateCache cCache = new FileTemplateCache(
            getTestcasePath() );

        // Use the default implementation
        cCache.setLoadingPolicy( FileTemplateCache.Strategy.LOAD_ON_DEMAND_WITH_REFRESH );
        cCache.addCacheListener(this);

        Thread cThread1 = new Thread( new CacheThread( 11000, cCache ));
        Thread cThread2 = new Thread( new CacheThread2( 11000, cCache ));
        Thread cThread3 = new Thread( new CacheThread4( 11000, cCache ));
        Thread cThread4 = new Thread( new CacheThread3( 11000, cCache ));
        Thread cThread5 = new Thread( new CacheThread5( 11000, cCache ));
        cThread1.start();
        cThread2.start();
        cThread3.start();
        cThread4.start();
        cThread5.start();

        Thread.sleep( 6000 );

        //cCache.setLoadingPolicy( FileTemplateCache.PRELOAD );

        Thread.sleep( 6000 );
        System.out.println( "Stopping auto-update." );
        cCache.stopAutoUpdate();

        System.out.println( "Elements updated: " + nUpdated );
        System.out.println( "Cache failures: " + nFailed );
        System.out.println( "Cache unavailable: " + nUnavailable );
        System.out.println( "Elements removed: " + nRemoved );
    }

    /**
     * Called when an element is removed from the cache.
     *
     * @param e the {@link freemarker.template.CacheEvent} fired.
     */
    public void elementRemoved(CacheEvent e) {
        nRemoved++;
    }

    /**
     * Called when a cache element fails to update.  The event fired
     * contains the <code>Exception</code> object.
     *
     * @param e the {@link freemarker.template.CacheEvent} fired.
     */
    public void elementUpdateFailed(CacheEvent e) {
        nFailed++;
    }

    /**
     * Called when a cache cannot be updated.  The event fired
     * contains the <code>Exception</code> object.
     *
     * @param e the {@link freemarker.template.CacheEvent} fired.
     */
    public void cacheUnavailable(CacheEvent e) {
        nUnavailable++;
    }

    /**
     * Called when a cache element updates successfully.
     *
     * @param e the {@link freemarker.template.CacheEvent} fired.
     */
    public void elementUpdated(CacheEvent e) {
        nUpdated++;
    }

    /**
     * Retrieve the root path of the FM-Classic distribution from a
     * properties file. This is used to set the root of the FM-Classic
     * cache, which in turn is used to find the example source code.
     *
     * @return the path to the root of the FM-Classic distribution.
     */
    private String getTestcasePath() throws IOException {
        InputStream input = getClass().getResourceAsStream( "testcase.properties" );
        Properties properties = new Properties();

        properties.load( input );
        input.close();

        return (String)properties.get( "testcasePath" );
    }

    /**
     * Main code for running the test.
     */
    public static void main(String args[]) throws Exception {
        TestCacheThreaded cTester = new TestCacheThreaded();

        cTester.test();
    }
}
