/*
 * FreeMarker: a tool that allows Java programs to generate HTML
 * output using templates.
 * Copyright (C) 1998-2004 Benjamin Geer
 * Email: beroul@users.sourceforge.net
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA  02111-1307, USA.
 */

package freemarker.testcase;

import freemarker.template.*;
import freemarker.ext.misc.JavaLoggingExceptionListener;

import java.io.*;
import java.util.logging.*;

import junit.framework.*;

/**
 * Performs a simple test on the {@link freemarker.ext.misc.JavaLoggingExceptionListener}
 * class to see whether the Java logger works correctly with it.
 *
 * @version $Id: TestJavaLoggingExceptionHandler.java 1101 2013-04-01 04:17:32Z run2000 $
 */
public final class TestJavaLoggingExceptionHandler extends AbstractTestCase {
    private FastHash modelRoot;
    private TemplateRuntimeHandler eventAdapter;

    /**
     * Constructor that passes the name of the test up to the JUnit
     * constructor.
     *
     * @param name the name of the test
     */
    public TestJavaLoggingExceptionHandler( String name ) {
        super( name );
    }

    /**
     * Set up the test case prior to running.
     */
    public void setUp() {
        eventAdapter = new SimpleEventAdapter(new JavaLoggingExceptionListener() );

        setUpFiles( "test-logging.html" );

        // Make a template data model.
        modelRoot = new FastHash();
        modelRoot.put( "message", new FastScalar( "Hello, world!" ));

    }

    /**
     * Performs the test. A fail result will generate information about
     * the failure and pass it up to JUnit.
     */
    public void runTest() throws Exception {
        StringWriter cWriter = new StringWriter();
        ByteArrayOutputStream cLogWriter = new ByteArrayOutputStream();
        Template cTemplate;

        // Configure the logger
        Logger logger = Logger.getLogger("");
        SimpleFormatter simpleFormatter = new SimpleFormatter();
        StreamHandler consoleHandler = new StreamHandler(cLogWriter, simpleFormatter);
        logger.addHandler(consoleHandler);

        cTemplate = TemplateFactory.INSTANCE.createFromString(m_aTemplateText);
        //System.out.println( cTemplate.toString() );

        // Process the template.
        TemplateProcessorParameters p = TemplateProcessorParameters
                .newInstance(cWriter)
                .withModel(modelRoot)
                .withEventHandler(eventAdapter);
        cTemplate.process(p);

        consoleHandler.flush();
        System.out.println( "Output:" );
        System.out.println( cLogWriter.toString() );

    }

    /**
     * Main method for running the test case.
     *
     * @param args the command line arguments
     */
    public static void main( String args[] ) throws Exception {
        AbstractTestCase cTest = new TestJavaLoggingExceptionHandler( "test-logging.html" );
        TestResult cResults = new TestResult();

        cTest.run( cResults );
    }

}
