/*
 * FreeMarker: a tool that allows Java programs to generate HTML
 * output using templates.
 * Copyright (C) 1998-2004 Benjamin Geer
 * Email: beroul@users.sourceforge.net
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA  02111-1307, USA.
 */

package freemarker.testcase.models;

import freemarker.template.*;
import java.io.*;

/**
 * A TemplateTransformModel that includes properties. These properties can be
 * set at model construction time, or, for the purposes of this demonstration,
 * can be passed in from a wrapper {@link TemplateMethodModel}.
 *
 * @version $Id: TransformModel2.java 1101 2013-04-01 04:17:32Z run2000 $
 */
public final class TransformModel2 implements TemplateTransformModel2 {

    private boolean m_bAmpersands = false;
    private boolean m_bQuotes = false;
    private boolean m_bTags = false;
    private String  m_aComment = "";

    private static final int READER_BUFFER_SIZE = 4096;

    /** Creates new TransformModel1 */
    public TransformModel2 () {
    }

    /**
     * Indicates whether we escape ampersands. This property can be set either
     * while the model is being constructed, or via a property passed in through
     * a {@link TemplateMethodModel}.
     *
     * @param bAmpersands should ampersand characters be escaped?
     */
    public void setAmpersands( boolean bAmpersands ) {
        m_bAmpersands = bAmpersands;
    }

    /**
     * Indicates whether we escape quotes. This property can be set either
     * while the model is being constructed, or via a property passed in through
     * a {@link TemplateMethodModel}.
     *
     * @param bQuotes should quote characters be escaped?
     */
    public void setQuotes( boolean bQuotes ) {
        m_bQuotes = bQuotes;
    }

    /**
     * Indicates whether we escape tags. This property can be set either
     * while the model is being constructed, or via a property passed in through
     * a {@link TemplateMethodModel}.
     *
     * @param bTags should tags be escaped?
     */
    public void setTags( boolean bTags ) {
        m_bTags = bTags;
    }

    /**
     * Sets a comment for this transformation. This property can be set either
     * while the model is being constructed, or via a property passed in through
     * a {@link TemplateMethodModel}.
     *
     * @param aComment the text to be used in place of '*' characters
     */
    public void setComment( String aComment ) {
        m_aComment = aComment;
    }

    /**
     * Performs a transformation/filter on FM-Classic output.
     *
     * @param source the input to be transformed
     * @param output the destination of the transformation
     */
    public void transform( Reader source, Writer output )
            throws IOException, TemplateModelException {

        // Output the source, converting unsafe characters to their
        // equivalent entities.
        boolean bCommentSent = false;

        char[]  aBuffer = new char[ READER_BUFFER_SIZE ];
        int buffSize = source.read( aBuffer );
        while( buffSize >= 0 ) {
            int currOffset = 0;
            int prevOffset = 0;
            while( currOffset < buffSize ) {
                char c = aBuffer[ currOffset ];
                switch (c) {
                    case '&':
                        if( m_bAmpersands ) {
                            if( currOffset > prevOffset ) {
                                output.write( aBuffer, prevOffset, currOffset - prevOffset );
                            }
                            output.write( "&amp;" );
                            prevOffset = currOffset + 1;
                        }
                        break;
                    case '<':
                        if( m_bTags ) {
                            if( currOffset > prevOffset ) {
                                output.write( aBuffer, prevOffset, currOffset - prevOffset );
                            }
                            output.write( "&lt;" );
                            prevOffset = currOffset + 1;
                        }
                        break;
                    case '>':
                        if( m_bTags ) {
                            if( currOffset > prevOffset ) {
                                output.write( aBuffer, prevOffset, currOffset - prevOffset );
                            }
                            output.write( "&gt;" );
                            prevOffset = currOffset + 1;
                        }
                        break;
                    case '"':
                        if( m_bQuotes ) {
                            if( currOffset > prevOffset ) {
                                output.write( aBuffer, prevOffset, currOffset - prevOffset );
                            }
                            output.write( "&quot;" );
                            prevOffset = currOffset + 1;
                        }
                        break;
                    case '\'':
                        if( m_bQuotes ) {
                            if( currOffset > prevOffset ) {
                                output.write( aBuffer, prevOffset, currOffset - prevOffset );
                            }
                            output.write( "&apos;" );
                            prevOffset = currOffset + 1;
                        }
                        break;
                    case '*':
                        if( ! bCommentSent ) {
                            if( currOffset > prevOffset ) {
                                output.write( aBuffer, prevOffset, currOffset - prevOffset );
                            }
                            output.write( m_aComment );
                            bCommentSent = true;
                            prevOffset = currOffset + 1;
                        }
                        break;
                }
                currOffset++;
            }
            if( currOffset > prevOffset ) {
                output.write( aBuffer, prevOffset, currOffset - prevOffset );
            }
            buffSize = source.read( aBuffer );
        }
    }

    /**
     * Is the model empty?
     *
     * @return <code>false</code> to indicate the model is not empty
     */
    public boolean isEmpty() throws TemplateModelException {
        return false;
    }

}
